/**
 * 'Activity' instances keep track of what we're currently doing.
 */

#include <math.h>

#include "activities.h"
#include "angles.h"
#include "qp.h"
#include "exception.h"
#include "astrotime.h"
#include "kalman.h"

#include "buffered-serial.h"
extern BufferedSerial bs;

#ifndef DEBUG
// DEBUG=0 turns off debugging, DEBUG>1 turns on extra debugging, requiring stdio.h
// XXX
#define DEBUG 0
#endif

#if PROVARINO
#include <iostream>
using std::cout;
using std::endl;
#define COUT cout << __FILE__ << ':' << __LINE__ << ": "
#define logprintf BufferedSerial::testenv_logprintf

#else
#define logprintf bs.logprintf

#endif

// the maximum speed we drive the drives (avoid under-voltage faults)
// FIXME: I think the that changes to the fault handling (handling of
// fault_flags in qp.ino, in revision 82f238) mean that we can drive
// these at full speed.
#define MAX_SPEED 0.5


/**
 * Construct an activity base-class.
 */
Activity::Activity(TelescopeController* tc, State* st, ActivityTypes type)
    : type_(type), next_(NULL), state(st), controller(tc), activity_status(0)
{
    if (st == NULL) {
        fatal("An Activity must have a non-NULL state");
    }
    if (tc == NULL) {
        fatal("An Activity must have a non-NULL controller");
    }
}

void Activity::cancel(void)
{
    mark_as_complete();
    state->step(millis());
    // constructor guarantees that controller is non-NULL
    controller->stop();
}

void Activity::az_click(ms_t t, double angle)
{
    // empty
}

void Activity::alt_click(ms_t t, double angle)
{
    // empty
}

Calibration::Calibration(TelescopeController* tc, State* state)
    : Activity(tc, state, Activity::CalibrationActivity)
{
    // XXX This calibration step is now I think almost completely
    // redundant, since the Kalman filters on each axis do this much
    // more accurately.  It might be worth replacing this by a
    // 'warm up' activity, which drives the telescope to a few points
    // on the sky, to get the Kalman filters working.

    // The calibration process moves through a sequence of phases:
    //
    // ph_position:
    //   move down and east to end-stop (ends in at_endstop)
    // ph_az_forward:
    //   move east for CAL_CLICKS_WEST clicks (ends in az_click)
    // ph_az_backward:
    //   move west to end-stop (ends in at_endstop)
    // ph_alt_upward:
    //   move up for CAL_CLICKS_UP clicks (ends in alt_click)
    // ph_alt_downward:
    //   move down to endstop (ends at_endstop)

    // The private variables don't need to be initialised -- they're
    // set during the calibration process -- but initialise them to
    // these sentinel values to make errors spottable
    reftime = time_az1 = time_az2 = time_alt1 = time_alt2 = TIME_NEVER;
    cal_count = 0;
}

// better values are perhaps 100 and 40 (ie, 200 and 80 degrees), but this is faster for testing
#define CAL_CLICKS_WEST 40
#define CAL_CLICKS_UP   40
#define CAL_CLICKS_OFFSET 5

ms_t Calibration::start(ms_t timenow)
{
    logprintf(PSTR("#calib: start"));
    phase = ph_position;
    controller->up(-MAX_SPEED, true);
    controller->east(+MAX_SPEED, false);
    return TIME_NEVER;
}

void Calibration::az_click(ms_t t, double azimuth)
{
    Activity::az_click(t, azimuth);
    switch(phase) {
      case ph_az_forward:
        cal_count++;
        if (cal_count == 0) {
            // start counting/timing here
            reftime = t;
        } else if (cal_count == CAL_CLICKS_WEST) {
            logprintf(PSTR("#calib: east..."));
            time_az1 = t - reftime;
            reftime = t;
            phase = ph_az_backward;
            controller->east(+MAX_SPEED, true);
        }
        break;

      case ph_az_backward:
        cal_count--;
        if (cal_count == 0) {
            time_az2 = t - reftime;
        }
        break;

      case ph_position:
        break;

      default:
        logprintf(PSTR("#calib: odd az_click in phase %d"), phase);
        break;
    }
}

void Calibration::alt_click(ms_t t, double altitude)
{
    Activity::alt_click(t, altitude);
    switch (phase) {
      case ph_alt_upward:
        cal_count++;
        if (cal_count == 0) {
            // start counting/timing here
            reftime = t;
        } else if (cal_count == CAL_CLICKS_UP) {
            logprintf("#calib: down...");
            time_alt1 = t - reftime;
            reftime = t;
            phase = ph_alt_downward;
            controller->up(-MAX_SPEED, true);
        }
        break;

      case ph_alt_downward:
        cal_count--;
        if (cal_count == 0) {
            time_alt2 = t - reftime;
        }
        break;

      case ph_position:
        break;

      default:
        logprintf(PSTR("#calibration: odd alt_click in phase %d"), phase);
        break;
    }
}
void Calibration::at_endstop(ms_t t, int az_click_count, int alt_click_count)
{
    // logprintf(PSTR("#calib: stop at " MS_FMT " (%d,%d)"),
    //           t, az_click_count, alt_click_count);
    switch (phase) {
      case ph_position:
        logprintf(PSTR("#calib: west..."));
        phase = ph_az_forward;
        cal_count = -CAL_CLICKS_OFFSET;
        // calibrate in azimuth: head west
        controller->east(-MAX_SPEED, true);
        break;

      case ph_az_forward:
        logprintf(PSTR("#calib: odd -- shouldn't be at endstop in ph_az_forward phase"));
        break;

      case ph_az_backward:
        logprintf(PSTR("#calib: up..."));
        cal_count = -CAL_CLICKS_OFFSET;
        // now calibrate in altitude: head up
        phase = ph_alt_upward;
        controller->up(+MAX_SPEED, true);
        break;

      case ph_alt_upward:
        logprintf(PSTR("#calib: odd -- at endstop in ph_alt_upward"));
        break;

      case ph_alt_downward:
        {
            Kalman* Kaz = state->kalman_phi();
            Kalman* Kalt = state->kalman_theta();

            // we're finished calibrating, and are back at the home position
            mark_as_complete();

            logprintf(PSTR("#calib az: %d az: " MS_FMT "/" MS_FMT
                           " => %g/%g rad/s (%g)"),
                      CAL_CLICKS_WEST,  time_az1, time_az2,
                      Kaz->fullspeed_pos(), Kaz->fullspeed_neg(), clicks_per_sec_az());
            logprintf(PSTR("#calib: %d alt: " MS_FMT "/" MS_FMT
                           " => %g/%g rad/s (%g)"),
                      CAL_CLICKS_UP, time_alt1, time_alt2,
                      Kalt->fullspeed_pos(), Kalt->fullspeed_neg(), clicks_per_sec_alt());
        }
        break;

      default:
        logprintf(PSTR("#calib: endstop in odd phase %d"), phase);
        break;
    }

    //logprintf(PSTR("#calib: in phase %d"), phase);
}

/**
 * Return the relationship between applied motor power and angular
 * speed, for the azimuthal motor.  The figure returned here
 * corresponds to the actual angular speed, in clicks per second, when the motor is commanded
 * with {@link TelescopeController::east}(1.0).  The relationship between this
 * argument and the motor's angular speed is assumed to be otherwise linear.
 */
double Calibration::clicks_per_sec_az(void)
{
    return 2.0*CAL_CLICKS_WEST / (time_az1 + time_az2) * 1000;
}

/**
 * Return the relationship between applied motor power and angular
 * speed, for the azimuthal motor.
 * @see rad_per_sec_az
 */
double Calibration::clicks_per_sec_alt(void)
{
    return 2.0*CAL_CLICKS_UP / (time_alt1 + time_alt2) * 1000;
}

ScheduleFollowing::ScheduleFollowing(TelescopeController* tc, State* st, ActivityTypes act)
    : Activity(tc, st, act)
{
    // empty
}

void ScheduleFollowing::cancel(void)
{
    Activity::cancel();
}

ms_t ScheduleFollowing::start(ms_t timenow)
{
    return poke(timenow);
}

ms_t ScheduleFollowing::poke(ms_t t)
{
    ms_t stepsize = state->step(t);
    ms_t next_poke_time;

    if (stepsize == 0) {
        cancel();
#if DEBUG
        logprintf(PSTR("#SF done"));
#endif
        next_poke_time = TIME_NEVER;
    } else {
        double theta1 = state->suggested_thetadot_frac();
        double phi1 = state->suggested_phidot_frac();

        double m = fmax(fabs(theta1), fabs(phi1));

#if DEBUG
        logprintf(PSTR("#SF> %g %g"), phi1, theta1);
#endif

        if (m > 1.0) {
            theta1 /= m;
            phi1 /= m;
#if DEBUG
            logprintf(PSTR("#SF>m=%g"), m);
            // Or, much more prolix:
            // logprintf(PSTR("#FS overspeed by %g: scaled from "
            //                "(%g, %g) to (%g,%g)"),
            //           m,
            //           state->suggested_phidot_frac(),
            //           state->suggested_thetadot_frac(),
            //           phi1, theta1);
            // Other items of possible debugging interest are:
            //   * Position pos = state->T().coords(Position::AzAlt);, then pos.lon(), pos.lat(),
            //   * controller->curr_az_speed(),
            //   * controller->curr_alt_speed(),
#endif
        }
        controller->up(theta1, false);
        controller->east(phi1, false);

        next_poke_time = t + stepsize;
    }

    return next_poke_time;
}

// void ScheduleFollowing::az_click(ms_t t, double azimuth)
// {
// #if DEBUG
//     logprintf(PSTR("#SF az! %g"), azimuth);
// #endif
//     Activity::az_click(t, azimuth);
// }
// void ScheduleFollowing::alt_click(ms_t t, double altitude)
// {
// #if DEBUG
//     logprintf(PSTR("#SF alt! %g"), altitude);
// #endif
//     Activity::alt_click(t, altitude);
// }
void ScheduleFollowing::at_endstop(ms_t t, int az_click_count, int alt_click_count)
{
#if DEBUG
    logprintf(PSTR("#SF end (%d,%d)clicks @ " MS_FMT),
              az_click_count, alt_click_count, t);
#endif
    cancel();
}

/**
 * Provide the target to drive to.
 * Note that the coordinates are given in AzAlt.
 *
 * @param tc a pointer to the telescope-control object
 * @param state the current state of the telescope pointing
 * @param az the target azimuth, in radians east from north
 * @param alt the target altitude, in radians towards the zenith (must be in [-pi/2,+pi/2])
 */
DriveAzAlt::DriveAzAlt(TelescopeController* tc, State* state, double az, double alt)
    : ScheduleFollowing(tc, state, Activity::DriveAzAltActivity),
      schedule(SlewSchedule(Position(az, alt, Position::AzAlt)))
{
    if (! schedule.schedule_ok()) {
        logprintf(PSTR("!DriveAzAlt: target (%g, %g) is in the empty sector"),
                  az, alt);
        mark_as_invalid();
    }
}

ms_t DriveAzAlt::start(ms_t t)
{
    // we mustn't set the schedule before this, in case this activity is queued up behind another
    state->set_schedule(&schedule);

#if DEBUG
#if PROVARINO
    COUT << "DriveAzAlt: (" << az << "," << alt << "), schedule=" << schedule
         << ", and in state " << *state << endl;
#else
    logprintf(PSTR("#DriveAzAlt: target (%g, %g)azalt"), az, alt);
#endif
#endif

    return ScheduleFollowing::start(t);
}

/**
 * Create a tracking activity.
 * This requires observatory location, and the speed at which the telescope is to track.
 * By default, it tracks at one sidereal day per solar day.
 *
 * @param tc a pointer to the telescope control object
 * @param state the current state of the telescope
 * @param obs the observatory position, as a `Position::Geog` position
 * @param speed the rotation speed clockwise about the celestial pole
 *   (ie, this will be negative for sidereal tracking); if this is
 *   given as zero, then we track at one sidereal day per solar day
 */
TrackSidereal::TrackSidereal(TelescopeController*tc,
                             State* state,
                             Position obs,
                             double speed)
    : ScheduleFollowing(tc, state, Activity::TrackSiderealActivity),
      // at 15'/min = 7.2e-5 rad/s = 7.2e-8 rad/ms
      schedule(TrackingSchedule(Quaternion(Quaternion::K), // rotate around k
                                (speed == 0.0
                                 ? -Angles::dms2r(0, 15)/(60 * STperUT)
                                 : speed),
                                obs))
{
    Quaternion actual = state->T();
    schedule.limit_tracking(actual, tc);

#if DEBUG
#if PROVARINO
    COUT << "TrackSidereal: schedule=" << schedule << endl;
#else
    logprintf(PSTR("#TrackSidereal: speed=%g rad/s"), schedule.speed());
#endif
#endif
}

ms_t TrackSidereal::start(ms_t t)
{
    // we mustn't set the schedule before this, in case this activity is queued up behind another
    state->set_schedule(&schedule);
    return ScheduleFollowing::start(t);
}

/**
 * Create a new activity which tracks across the sky in azimuth.
 * @param tc a pointer to the telescope control object
 * @param state the current state of the telescope
 * @param speed the required tracking speed, in rad/s
 */
TrackAzimuth::TrackAzimuth(TelescopeController* tc, State* state, double speed)
    : ScheduleFollowing(tc, state, Activity::TrackAzimuthActivity),
      schedule(TrackingSchedule(Quaternion(Quaternion::K), -speed, Position::get_null()))
{
    Quaternion actual = state->T();
    schedule.limit_tracking(actual, tc);

#if DEBUG
#if PROVARINO
    COUT << "TrackAzimuth: schedule=" << schedule << endl;
#else
    logprintf(PSTR("#TrackAzimuth: speed=%g rad/s"), schedule.speed());
#endif
#endif
}

ms_t TrackAzimuth::start(ms_t t)
{
    // we mustn't set the schedule before this, in case this activity is queued up behind another
    state->set_schedule(&schedule);
    return ScheduleFollowing::start(t);
}


#if 0
Activity::~Activity() { COUT << "Activity destructor" << endl; }
Calibration::~Calibration() { COUT << "Calibration destructor" << endl; }
ScheduleFollowing::~ScheduleFollowing() { COUT << "ScheduleFollowing destructor" << endl; }
DriveAzAlt::~DriveAzAlt() { COUT << "DriveAzAlt destructor" << endl; }
TrackAzimuth::~TrackAzimuth() { COUT << "TrackAzimuth destructor" << endl; }
#else
Activity::~Activity() {}
Calibration::~Calibration() {}
ScheduleFollowing::~ScheduleFollowing() {}
DriveAzAlt::~DriveAzAlt() {}
TrackSidereal::~TrackSidereal() {}
TrackAzimuth::~TrackAzimuth() { }
#endif


static byte allocations = 0;

#define ACTIVITY_BUFSIZE 160
#define N_ACTIVITIES 3
static char objbuf[ACTIVITY_BUFSIZE * N_ACTIVITIES];

/**
 * Return the maximum number of activities it's possible to have
 * allocated at one time.  This is a compilation constant.
 */
int Activity::max_activities(void)
{
    return N_ACTIVITIES;
}

/**
 * Return the number of currently allocated activities.  This will
 * generally be the same as the number of activities in the queue, but
 * this counts the allocations rather than the queue size.
 */
int Activity::n_activities(void)
{
    int n = 0;
    for (byte a = allocations; a != 0; a>>=1) {
        if (a & 1) n++;
    }
    return n;
}

/**
 * Allocate space for an Activity.
 *
 * If there is no space for a new activity, then we throw a
 * recoverable error, which takes us back to the qp main loop.
 */
void* Activity::operator new(size_t size)
{
    /*
     * Override the `new` operator, so
     * that we can create activities using `new` without worrying about
     * general free store issues on the Arduino.  We do this by allocating
     * the objects from a _very small_ pool.  This is in preference
     * to doing so using a factory method, which prevents us having
     * significantly different constructors, or requires object-mutation
     * functions, which mean that the objects have to be non-constant.
     *
     * We call this recoverable_error method, rather than throwing a
     * `bad_alloc` exception, since exceptions are disabled on the Arduino).
     */
    if (size > ACTIVITY_BUFSIZE) {
        fatal(PSTR("Activity::new: size %ld > %d!\n"), size, ACTIVITY_BUFSIZE);
    }


#define BIT_SET_P(mask, n) (mask & (1<<n))
#define BIT_SET(mask, n) (mask |= (1<<n))
#define BIT_CLEAR(mask, n) (mask &= ~(1<<n))

    void* rval = NULL;
    byte allocnum;

    for (allocnum = 0; allocnum < N_ACTIVITIES; allocnum++) {
        if (! BIT_SET_P(allocations, allocnum)) {
            BIT_SET(allocations, allocnum);
            rval = &objbuf[allocnum * ACTIVITY_BUFSIZE];
#if DEBUG > 1
            printf("Allocate slot %d, mask=%x\n", allocnum, allocations);
#endif
            break;
        }
    }

    if (allocnum == N_ACTIVITIES) {
        fatal(PSTR("#Activity::new: new called but no space available"));
    }

#if DEBUG
    logprintf(PSTR("#alloc new %d mask=%d"), ((char*)rval - &objbuf[0])/ACTIVITY_BUFSIZE, allocations);
#if DEBUG > 1
    printf("Activity::new: %ld <= %d, returning %p\n", size, ACTIVITY_BUFSIZE, rval);
#endif
#endif

    return rval;
}

/**
 * Deallocate an Activity
 */
void Activity::operator delete(void *p)
{
    byte allocnum;

#if DEBUG
    logprintf(PSTR("#alloc delete %d mask=%d"), ((char*)p - &objbuf[0])/ACTIVITY_BUFSIZE, allocations);
#endif

    for (allocnum = 0; allocnum < N_ACTIVITIES; allocnum++) {
        if (p == &objbuf[allocnum * ACTIVITY_BUFSIZE]) {
            if (BIT_SET_P(allocations, allocnum)) {
                BIT_CLEAR(allocations, allocnum);
#if DEBUG > 1
                printf("Free slot %d, mask=%x\n", allocnum, allocations);
#endif
                break;
            } else {
                // XXX this should be a fatal error, no?
                fatal(PSTR("Activity::delete: double free"));

                // failing that...
                // logprintf(PSTR("!Activity::delete: double free n=%d mask=%x"), allocnum, allocations);
                // BIT_CLEAR(allocations, allocnum);
                break;
            }
        }
    }

    if (allocnum == N_ACTIVITIES) {
        fatal(PSTR("Activity::delete: called on unallocated pointer"));
    }
#if DEBUG > 1
    printf("Activity::delete: releasing %p\n", p);
#endif
}

/**
 * Delete all activity allocations. This doesn't do any cancelling or
 * other tidyup.  It's primarily intended to support unit tests, but
 * might be of use in, for example, some catch-all exception handler.
 */
void Activity::delete_all_allocations(void)
{
#if DEBUG
    logprintf(PSTR("#alloc delete-all"));
#endif
    allocations = 0;
}

/**
 * Maintain a queue of upcoming activities.  Activities should be
 * created using `new XXXActivity()`, and either appended to the list
 * here, or replace the contents of the queue.  Activities are not
 * started by being created or added to this queue.  Once activities
 * have been added to this queue, the queue takes ownership of them,
 * so they should not be otherwise deleted.
 */
ActivityQueue::ActivityQueue()
    : head_(NULL)
{
    // empty
}

/**
 * Return the current activity.
 * This does not change the queue.
 * @return a pointer to the activity, or NULL if the queue is empty
 */
Activity* ActivityQueue::current() const
{
    return head_;
}

/**
 * Return the next activity.  The 'current activity' --
 * the head of the queue -- is dequeued, cancelled (possibly
 * redundantly) and deallocated.  Has no effect, returning only NULL, if the queue is empty.
 *
 * @return a pointer to the activity now at the head of the queue;
 *    or NULL if the queue is now empty
 */
Activity* ActivityQueue::next()
{
    if (head_ != NULL) {
        Activity* n = head_->next_;
        head_->cancel();
        delete head_;
        head_ = n;
    }
    return head_;
}

/**
 * Cancel all the activities in the queue.
 * Each Activity has its {@link Activity::cancel} method called, and
 * is then removed from the queue and deallocated.  The queue ends up empty.
 */
void ActivityQueue::cancel()
{
    while (head_ != NULL) {
        next();                 // also calls cancel()
    }
    // It is tempting to call Activity::delete_all_allocations() here
    // -- belt and braces! -- but that would additionally deallocate
    // any activities which are about to be added to the queue
}

/**
 * Replace any activities in the queue.  Equivalent to {@link cancel}
 * followed by {@link append}.
 */
void ActivityQueue::replace(Activity* a)
{
    cancel();
    head_ = a;
}

/**
 * Appends the given activity to the end of the queue.
 * Like the {@link replace} method, this takes ownership of the object.
 */
void ActivityQueue::append(Activity* a)
{
    a->next_ = NULL;            // should be true anyway

    if (head_ == NULL) {
        head_ = a;
    } else {
        Activity* p = head_;
        while (p->next_ != NULL) {
            p = p->next_;
        }
        p->next_ = a;
    }
}
