#ifndef ACTIVITIES_H_LOADED
#define ACTIVITIES_H_LOADED 1

#include "types.h"
#include "state.h"
#include "qp.h"

/**
 * The Activity abstract class encapsulates the state for extended
 * projects, such as calibrations.  The various callbacks below may
 * return the current object, or a pointer to a new Activity subclass.
 *
 * The method {@link is_complete} indicates that an Activity has
 * completed, and that any state or results within it may be retrieved.
 */
class Activity {
    friend class ActivityQueue;

 public:
    /**
     * The activity types here correspond to the this class and its
     * subclasses.  These are used as the values returned by {@link
     * type}, and are associated in the obvious ways with the
     * associated derived types of this class.
     */
    enum ActivityTypes { BaseActivity,
                         CalibrationActivity,
                         ScheduleFollowingActivity,
                         DriveAzAltActivity,
                         TrackSiderealActivity,
                         TrackAzimuthActivity, };

 private:
    const ActivityTypes type_;
    Activity* next_;

 protected:
    /** The state which we are to manipulate */
    State* const state;         // const pointer to non-const State

    /**
     * The controller object provides methods for physically moving
     * the telescope.
     */
    TelescopeController* const controller; // const pointer to non-const controller

    /**
     * The internal status of the object: complete / invalid.
     * @see is_complete
     * @see mark_as_complete
     * @see is_valid
     * @see mark_as_invalid
     */
    byte activity_status;
#define ACTIVITY_STATUS_COMPLETE 1
#define ACTIVITY_STATUS_INVALID 1<<1
    // /**
    //  * Set to true by a derived class when its work is done.
    //  * @see is_complete
    //  */
    // bool complete_p;

    /** Mark the activity as complete */
    void mark_as_complete(void) { activity_status |= ACTIVITY_STATUS_COMPLETE; }
    /** Mark the activity as invalid. */
    void mark_as_invalid(void) {
        activity_status |= (ACTIVITY_STATUS_COMPLETE | ACTIVITY_STATUS_INVALID);
    }

    Activity(TelescopeController*, State*, ActivityTypes);

 public:
    /**
     * The method called when the system detects an azimuthal click.
     * This is for the information of the activity; it should not
     * attempt to update a telescope position.
     * @param t the current time
     * @param azimuth the estimated angle, in radians east from south
     */
    virtual void az_click(ms_t t, double azimuth);

    /**
     * The method called when the system detects an altitude click.
     * This is for the information of the activity; it should not
     * attempt to update a telescope position.
     * @param t the current time
     * @param altitude the estimated angle, in radians north from horizontal
     */
    virtual void alt_click(ms_t t, double altitude);

    /**
     * The method called when the system detects a timeout, at the
     * limit of an axis's travel.
     * @param t the current time
     * @param az_click_count the azimuthal click-count at this position
     * @param alt_click_count the altitudinal click-count at this position
     */
    virtual void at_endstop(ms_t t, int az_click_count, int alt_click_count) = 0;

    /** Return true if the activity is complete. */
    bool is_complete(void) const { return activity_status & ACTIVITY_STATUS_COMPLETE; }

    /** Return true if the activity is valid. */
    bool is_valid(void) const { return ! (activity_status & ACTIVITY_STATUS_INVALID); }

    /** Cancel the activity.  Idempotent. */
    virtual void cancel(void);

    // The following is necessary because on the Arduino, there is no RTTI, so no dynamic_cast.
    /** Return the type of Activity */
    ActivityTypes type(void) { return type_; }

    /**
     * The activity will be started by the main loop calling the
     * `start` function.
     *
     * @returns the time at which the activity should next be poked
     * (return `TIME_NEVER` to avoid being poked again).
     * @see poke(ms_t)
     */
    virtual ms_t start(ms_t now) { return TIME_NEVER; }

    /**
     * From time to time, the main loop will 'poke' this activity, by
     * calling this function.  This allows the activity to undertake
     * some periodic activity.
     *
     * @param t the current time
     * @returns the time at which the main loop should next poke this activity
     */
    virtual ms_t poke(ms_t t) { return TIME_NEVER; };

    static void* operator new (size_t size);
    static void operator delete (void *p);
    // overriding delete requires (I think) that we define destructors
    virtual ~Activity();

    static int max_activities(void);
    static int n_activities(void);
    static void delete_all_allocations(void);
};

/**
 * Calibrating.
 */
class Calibration : public Activity {
 private:
    ms_t reftime;
    ms_t time_az1, time_az2, time_alt1, time_alt2;
    signed int cal_count;
    enum {ph_position,
          ph_az_forward, ph_az_backward,
          ph_alt_upward, ph_alt_downward} phase;


 public:
    /** Create a calibration instance */
    Calibration(TelescopeController*, State*);
    ~Calibration();

    ms_t start(ms_t);
    void az_click(ms_t t, double);
    void alt_click(ms_t t, double);
    void at_endstop(ms_t t, int az_click_count, int alt_click_count);
    double clicks_per_sec_az(void);
    double clicks_per_sec_alt(void);
};

/**
 * The base of a couple of classes where the activity consists of
 * following a schedle of rotations.
 */
class ScheduleFollowing : public Activity {
 protected:
    /** Create an instance of the activity */
    ScheduleFollowing(TelescopeController*, State*, ActivityTypes);
    virtual ~ScheduleFollowing() = 0;

 public:
    virtual ms_t start(ms_t);
    virtual void cancel(void);
    virtual ms_t poke(ms_t t);
    //virtual void az_click(ms_t t, double);
    //virtual void alt_click(ms_t t, double);
    virtual void at_endstop(ms_t t, int az_click_count, int alt_click_count);
};

/**
 * Activity: Driving the telescope to a given AzAlt location.
 */
class DriveAzAlt : public ScheduleFollowing {
 private:
    SlewSchedule schedule;

 public:
    DriveAzAlt(TelescopeController*, State*, double phi, double theta);
    ~DriveAzAlt();
    ms_t start(ms_t);
};

/**
 * Activity: track a point on the celestial sphere through a sidereal rotation
 */
class TrackSidereal : public ScheduleFollowing {
 private:
    TrackingSchedule schedule;

 public:
    TrackSidereal(TelescopeController*, State*, Position obs, double speed=0.0);
    ~TrackSidereal();
    ms_t start(ms_t);
};

/**
 * Track across the sky in azimuth.  This corresponds to a rotation
 * about the zenith (ie, in increasing azimuth/westwards).
 */
class TrackAzimuth : public ScheduleFollowing {
 private:
    TrackingSchedule schedule;

 public:
    TrackAzimuth(TelescopeController*, State*, double speed);
    ~TrackAzimuth();
    ms_t start(ms_t);
};

/**
 * Maintain a queue of Activity instances
 */
class ActivityQueue {
 private:
    Activity* head_;

 public:
    ActivityQueue();
    Activity* current() const;
    Activity* next();
    void append(Activity*);
    void replace(Activity*);
    void cancel(void);
};

#endif /* ACTIVITIES_H_LOADED */
