#include <math.h>

#include "exception.h"
#include "angles.h"

/**
 * Convert an angle in h:m:s to radians.  Any sign on the hour is
 * translated into an overall sign; thus
 *
 *    hms2r(-h, m, s) = -hms2r(h, m, s)
 *
 * The minutes and seconds must be positive.
 * The result is always normalised into [0, 2pi).
 */
double Angles::hms2r(int h, int m, double s)
{
    double pipi = 2*M_PI;

    if (m < 0) fatal(PSTR("Range exception in %s: %s=%g"), "hms2r", "m", m);
    if (s < 0) fatal(PSTR("Range exception in %s: %s=%g"), "hms2r", "s", s);

    double v;
    if (h < 0) {
        v = -(-h + m/60.0 + s/3600.0) * (M_PI/12.0);
    } else {
        v = (h + m/60.0 + s/3600.0) * (M_PI/12.0);
    }

    // normalise it
    while (v >= pipi)
        v -= pipi;
    while (v < 0.0)
        v += pipi;

    return v;
}

/**
 * Convert an angle in d:m:s to radians.  Any sign on the degree is
 * translated into an overall sign; thus
 *
 *    dms2r(-d, m, s) = -dms2r(d, m, s)
 *
 * The minutes and seconds must be positive.
 * The result is normalised into [0, 2pi) if `norm` is true.
 */
double Angles::dms2r(int d, int m, double s, bool norm)
{
    double pipi = 2*M_PI;

    if (m < 0) fatal(PSTR("Range exception in %s: %s=%g"), "dms2r", "m", m);
    if (s < 0) fatal(PSTR("Range exception in %s: %s=%g"), "dms2r", "s", s);

    double v;
    if (d < 0) {
        v = -(-d + m/60.0 + s/3600.0) * (M_PI/180.0);
    } else {
        v = (d + m/60.0 + s/3600.0) * (M_PI/180.0);
    }

    if (norm) {
        while (v >= pipi)
            v -= pipi;
        while (v < 0.0)
            v += pipi;
    }

    return v;
}

/** Convert radians to degrees */
double Angles::r2d(double r)
{
    return r * (180.0/M_PI);
}

/** Convert degrees to radians */
double Angles::d2r(double d)
{
    return d * (M_PI/180.0);
}

/**
 * Convert an angle in radians into hours, minutes and seconds.
 * @param buf a buffer into which the formatted angle is written
 * @param buflen the size of the buffer, which must be large enough to accommodate a terminating null
 * @param angle the angle
 * @return the argument `buf`
 */
char* Angles::r2hms_b(char* buf, size_t buflen, double angle)
{
    while (angle < 0.0)
        angle += 2*M_PI;

    double t = angle * 12/M_PI; // sidereal time in hours
    int H = floor(t);
    t = (t - H)*60;             // remainder in minutes
    int M = floor(t);
    double S = (t - M)*60;

    snprintf(buf, buflen, "%02d:%02d:%06.3f", H, M, S);
    return buf;
}

#if 1
/**
 * Convert an angle in radians into degrees and minutes.
 * There's no need for the accuracy of seconds, and it just clutters the output.
 * If `norm` is true, then normalise the angle to be in [0,2pi);
 * if not, then it may be negative.
 * @param buf a buffer into which the formatted angle is written
 * @param buflen the size of the buffer, which must be large enough to accommodate a terminating null
 * @param angle the angle
 * @param norm true if the angle is to be normalised to [0,2pi)
 * @return the argument `buf`
 */
char* Angles::r2dms_b(char* buf, size_t buflen, double angle, bool norm)
{
    char pfx[] = { ' ', '+', '-' };
    int pos;

    if (norm) {
        while (angle < 0.0)
            angle += 2*M_PI;
        while (angle >= 2*M_PI)
            angle -= 2*M_PI;
        pos = 0;
    } else {
        if (angle >= 0.0) {
            pos = 1;
        } else {
            pos = 2;
            angle *= -1;
        }
    }

    double d = r2d(angle);      // angle in degrees
    int D = floor(d);
    d = (d - D)*60;             // remainder in minutes
    int M = round(d);
    //double S = (d - M)*60;

    snprintf(buf, buflen, "%c%dd%02dm", pfx[pos], D, M);
    return buf;
}
#else
/**
 * Convert an angle in radians into degrees, minutes and seconds.
 * If `norm` is true, then normalise the angle to be in [0,2pi);
 * if not, then it may be negative.
 * @param buf a buffer into which the formatted angle is written
 * @param buflen the size of the buffer, which must be large enough to accommodate a terminating null
 * @param angle the angle
 * @param norm true if the angle is to be normalised to [0,2pi)
 * @return the argument `buf`
 */
char* Angles::r2dms_b(char* buf, size_t buflen, double angle, bool norm)
{
    char pfx[] = { ' ', '+', '-' };
    int pos;

    if (norm) {
        while (angle < 0.0)
            angle += 2*M_PI;
        while (angle >= 2*M_PI)
            angle -= 2*M_PI;
        pos = 0;
    } else {
        if (angle >= 0.0) {
            pos = 1;
        } else {
            pos = 2;
            angle *= -1;
        }
    }

    double d = r2d(angle);      // angle in degrees
    int D = floor(d);
    d = (d - D)*60;             // remainder in minutes
    int M = floor(d);
    double S = (d - M)*60;

    snprintf(buf, buflen, "%c%dd%02dm%05.2fs", pfx[pos], D, M, S);
    return buf;
}
#endif

#if PROVARINO
/**
 * Convert an angle in radians into hours, minutes and seconds.
 */
std::string Angles::r2hms(double angle)
{
    char b[13];
    return std::string(r2hms_b(b, 13, angle));
}

/**
 * Convert an angle in radians into degrees, minutes and seconds.
 * If `norm` is true, then normalise the angle to be in [0,2pi);
 * if not, then it may be negative.
 */
std::string Angles::r2dms(double angle, bool norm)
{
    char b[16];
    return std::string(r2dms_b(b, 16, angle, norm));
}
#endif
