// Routines to support a simple time object.
//
// It might be reasonable to replace parts of this with code from SOFA
// <http://www.iausofa.org>.  Not doing so means that we can only
// calculate GMST/LMST, and not GAST/LAST, since the difference
// involves the equation of the equinoxes, the calculation of which is
// beyond the simple intentions of this module.  The errors are of
// order 1s of time = 15as = 72 urad.

#include <math.h>

#include "astrotime.h"
#include "angles.h"
#include "exception.h"

#if PROVARINO
#include <iostream>
#include <string>
#endif

static int calculate_mjd(const int year,
                         const int month,
                         const int day);

/**
 * The default constructor produces a valid, but largely useless
 * structure, representing midnight on 1858 November 17 (MJD=0) at the meridian.
 */
AstroTime::AstroTime()
    : ut24(0), obslong(0), mjd_ut0(0), gmst(-1)
{
    // empty
}

/**
 * Create a new time.  The time must be given in UT1.
 *
 * @param year must be after 1859 (ie, positive MJD)
 * @param month month in [1, 12]
 * @param day day in [1, 31]
 * @param hour UT1 hour, positive
 * @param minute UT1 minute, positive
 * @param second UT1 second, positive
 * @param obslong the observatory east-longitude, in radians
 */
AstroTime::AstroTime(const int year,
                     const int month,
                     const int day,
                     const int hour,
                     const int minute,
                     const double second,
                     const double obslong)
    : obslong(obslong)
{
    if (hour < 0)	fatal(PSTR("Range exception in %s: %s=%g"), "AstroTime", "hour", hour);
    if (minute < 0)	fatal(PSTR("Range exception in %s: %s=%g"), "AstroTime", "minute", minute);
    if (second < 0.0)	fatal(PSTR("Range exception in %s: %s=%g"), "AstroTime", "second", second);

    ut24 = hour + minute/60.0 + second/3600.0;
    mjd_ut0 = calculate_mjd(year, month, day);
    if (ut24 >= 24.0) {
        // Is this the most reliable way to do this normalisation?
        double n_days = ut24/24;
        int ndi = floor(n_days);
        ut24 = (n_days - ndi) * 24;
        mjd_ut0 += ndi;
    }

    gmst = -1;                  // negative sentinel value
}

/**
 * Return a new astrotime, with the time given by the tm struct.
 */
// AstroTime::AstroTime(const struct tm* t, const double obslong)
//     : obslong(obslong)
// {
//     ut24 = t->tm_hour + t->tm_min/60.0 + t->tm_sec/3600.0;
//     mjd_ut0 = calculate_mjd(1900 + t->tm_year, t->tm_mon + 1, t->tm_mday);
//     gmst = -1;
// }

/**
 * Returns a new time, offset from the reference time by a number of
 * UT1 seconds.
 */
AstroTime::AstroTime(const AstroTime& ref, const double extra_secs)
{
    obslong = ref.obslong;
    gmst = -1;

    mjd_ut0 = ref.mjd_ut0;
    ut24 = ref.ut24 + extra_secs/3600;

    if (ut24 >= 24.0 || ut24 < 0.0) {
        // Is this the most reliable way to do this normalisation?
        double n_days = ut24/24;
        int ndi = floor(n_days);
        ut24 = (n_days - ndi) * 24;
        mjd_ut0 += ndi;
    }
}

/**
 * Create a new time.  The time is given in LST on the supplied date.
 * LST must be supplied in radians in the range [0,2pi).
 */
#if 0
AstroTime::AstroTime(const int year,
                     const int month,
                     const int day,
                     const double lst_rad,
                     const double obslong)
    : obslong(obslong)
{
    if (lst_rad < 0 || lst_rad > 2*M_PI)
        fatal(PSTR("Range exception in %s: %s=%g"), "AstroTime", "lst_rad", lst_rad);

    mjd_ut0 = calculate_mjd(year, month, day);

    gmst = -1;                  // negative sentinel value
    hour = minute = 0;
    second = 0.0;
    double gmst0 = gmst_2pi();  // GMST at UT0, in radians
    gmst = lst_rad - obslong;

    // now convert this to h:m:s of UT, in hours
    double t = gmst / STperUT * 12/M_PI;
    hour = floor(t);
    t = (t - hour) * 60;
    minute = floor(t);
    second = (t - minute) * 60;
}
#endif

/**
 * Create a new time, defaulting to noon.
 */
AstroTime::AstroTime(const int year,
                     const int month,
                     const int day)
    : ut24(12), obslong(0.0)
{
    mjd_ut0 = calculate_mjd(year, month, day);
    gmst = -1;                  // negative sentinel value
}

/**
 * Return a new AstroTime representing the same UT as this one, but
 * for a different observatory longitude.
 *
 * @param new_obs_lon the observatory east-longitude, in radians
 */
AstroTime AstroTime::move(double new_obs_lon)
{
    AstroTime rval = AstroTime();

    rval.ut24 = ut24;
    rval.mjd_ut0 = mjd_ut0;
    rval.gmst = -1;
    rval.obslong = new_obs_lon;

    return rval;
}

/**
 * Calculate the MJD given Y, M, D.
 */
static int calculate_mjd(const int year,
                         const int month,
                         const int day)
{
    // MJD=0 was on 1858 November 17
    if (year < 1859)
        fatal(PSTR("Range exception in %s: %s=%g"), "calculate_mjd", "year", year);
    if (month < 1 || month > 12)
        fatal(PSTR("Range exception in %s: %s=%g"), "calculate_mjd", "month", month);
    if (day < 1 || day > 31)
        fatal(PSTR("Range exception in %s: %s=%g"), "calculate_mjd", "day", day);

    // The following calculation, without the 2400001, is from the
    // Explanatory Supplement, and calculates the Julian Day Number
    // for a given year/month/day.  That is it calculates the JD for
    // noon on the given day, which is a round number.  The MJD for
    // that instant is that number minus 2400000.5.  Thus the MJD for
    // 00:00 on that day is that number minus 0.5.  To obtain the MJD
    // for the day which includes this hour/minute/second, we must
    // therefore subtract 2400001.
    //
    // Thus 2000 January 1.5 is JD 2451545.0, which is MJD 51544.5,
    // and the MJD of 00:00 on that day is 51544.
    long int m1412 = (month - 14) / 12;
    return (1461L * (year + 4800 + m1412)) / 4
        + (367L * (month - 2 - 12 * m1412)) / 12
        - (3L * ((year + 4900 + m1412)/100)) / 4 + day - (32075 + 2400001);
}

/**
 * Return the MJD at time 00:00 on the indicated day.  This is a round number.
 */
const int AstroTime::mjd(void) const
{
    return mjd_ut0;
}

#if 0
/**
 * Returns the Earth Rotation Angle (similar to, but replacing, GMST)
 * of the instant, in radians, and forced into the interval [0, 2pi).
 *
 * The following algorithm is from IERS Convensions 2003 (IERS
 * Technical Note 32), Section 5.4.4:
 *
 * The Earth Rotation Angle, \theta, is obtained by the use of its conventional
 * relationship with UT1 as given by Capitaine et al. (2000),
 *
 * \theta(Tu) = 2\pi(0.7790572732640 + 1.00273781191135448 Tu),
 *
 * where Tu=(Julian UT1 date−2451545.0), and UT1=UTC+(UT1−UTC), or equivalently
 *
 * \theta(Tu) = 2\pi(UT1 Julian day number elapsed since 2451545.0
 *    + 0.7790572732640
 *    + 0.00273781191135448 Tu),
 *
 * the quantity UT1−UTC to be used (if not estimated from the
 * observations) being the IERS value.
 */
const double AstroTime::earth_rotation_angle_2pi(void) const
{
    if (era < 0.0) {
        // mjd() returns the number of complete days since 2400000.5,
        // so (days since JD2451545.0) is MJD-51544.5
        double Tu = mjd() - 51544.5 + (hour*3600.0 + minute*60.0 + second)/(24*60*60);

        double era_days = fmod(0.7790572732640 + 1.00273781191135448 * Tu, 1.0);

        if (era_days < 0.0)
            era_days += 1.0;
        era = 2 * M_PI * era_days;
    }
    return era;
}
#endif

/**
 * Returns the Local Mean Sidereal Time at the observatory.
 */
const double AstroTime::lmst(void) const
{
    double lst = gmst_2pi() + obslong;

    while (lst < 0.0) {
        lst += 2*M_PI;
    }
    while (lst > 2*M_PI){
        lst -= 2*M_PI;
    }
    return lst;
}

/**
 * Return UT1 as a number of hours.
 * The result is always in the range [0, 24).
 */
const double AstroTime::ut1(void) const
{
    return ut24;
}

/**
 * The following calculates GMST of date, in radians.
 *
 * This evaluates the expression for GMST in terms of UT, as given in
 * Explanatory Supplement Eq.2.24-1, but with the resulting time
 * normalised to be in [0,2pi)
 */
const double AstroTime::gmst_2pi(void) const
{
    if (gmst < 0.0) {
#if 0
        double Tu = (mjd() + 0.5 - 51545.0) / 36525.0;
        // The following would almost certainly require doubles to have 8-byte precision
        double gmst = ((-6.2e-6 * Tu + 0.093104) * Tu + 8640184.812866) * Tu + 24110.54841;
        // double gmst_at_0 = 24110.54841
        //     + 8640184.812866 * Tu
        //     + 0.093104 * Tu * Tu
        //     - 6.2e-6 * Tu*Tu*Tu;
#endif
        /*
         * This is the expression for GMST1 of 0h UT1, as
         * given in ES Eq.2.24-1, rewritten in terms of du rather than
         * Tu, and with the negligible third order term dropped (the
         * second order term is negligible, too, realistically, but
         * ... why not?).  Rewriting it in this way means that the
         * 1st-order coefficient is of roughly the same order as du,
         * improving accuracy even with single-precision floats.
         */
        double du = (mjd() + 0.5 - 51545.0);
        double c2 = 0.093104 / (36525.0*36525.0);
        double c1 = 8640184.812866 / 36525.0;
        double c0 = 24110.54841;
        double gmst_sec = (c2*du + c1) * du + c0;

        double secs_in_day = 24.0 * 60.0 * 60.0;

        //printf("gmst_sec(mjd=%d, du=%g) = %g", mjd(), du, gmst_sec);
        gmst_sec += STperUT * ut24*3600;
        gmst = fmod(gmst_sec, secs_in_day) / secs_in_day * 2*M_PI;
        // gmst now in [-pi,pi)
        if (gmst < 0.0)
            gmst += 2*M_PI;
        // gmst now in [0,2pi)
    }

    return gmst;
}

static int ut1_str_b(char* buf, size_t buflen, double ut24, bool iso_p)
{
    double t = ut24;
    int hour = floor(t);
    t = (t - hour) * 60;
    int minute = floor(t);
    double second = (t - minute) * 60;

    int len;
    if (iso_p) {
        len = snprintf(buf, buflen, "%02d:%02d:%06.3f", hour, minute, second);
    } else {
        len = snprintf(buf, buflen, "%2d:%02d:%06.3f", hour, minute, second);
    }
    return len;
}

static int ymd_str_b(char* buf, size_t buflen, int mjd_ut0, bool iso_p)
{
    // The following calculation is from the Explanatory Supplement
    int L = mjd_ut0 + 2400001 + 68569;// JD + 68569
    int N = (4 * L) / 146097;
    L -= (146097 * N + 3)/4;
    int I = (4000 * (L+1)) / 1461001;
    L -= (1461 * I)/4 - 31;
    int J = (80 * L) / 2447;
    int D = L - (2447 * J) / 80;
    L = J/11;
    int M = J + 2 - 12*L;
    int Y = 100 * (N - 49) + I + L;

    int len;
    if (iso_p) {
        len = snprintf(buf, buflen, "%d-%02d-%02d", Y, M, D);
    } else {
        const char* mname[] = { "",
                                "Jan", "Feb", "Mar", "Apr",
                                "May", "Jun", "Jul", "Aug",
                                "Sep", "Oct", "Nov", "Dec", };
        len = snprintf(buf, buflen, "%d %s %2d", Y, mname[M], D);
    }
    return len;
}

/**
 * Format the current time into the given buffer.
 * @param buf a buffer to receive the formatted string
 * @param buflen the size of this buffer;
 *   if the string would be longer than this it is truncated
 */
char* AstroTime::str(char* buf, size_t buflen)
{
    int len1 = ymd_str_b(buf, buflen, mjd_ut0, true);

    char* p = buf + len1;
    strcpy(p, "T");
    while (*p != '\0')
        p++;

    ut1_str_b(p, buflen-(p-buf), ut24, true);

    return buf;
}

#if PROVARINO
/**
 * Return the time as a UT1 string.
 */
std::string AstroTime::ut1_str(void) const
{
    char buf[16];
    ut1_str_b(buf, 16, ut24, false);
    return std::string(buf);
}

/**
 * Return the date (without time) as a string.
 */
std::string AstroTime::ymd_str(void) const
{
    char buf[12];
    ymd_str_b(buf, 12, mjd_ut0, false);
    return std::string(buf);
}

/**
 * Return the UT corresponding to the given date, as a string.
 */
std::string AstroTime::str(void) const
{
    return ymd_str() + ' ' + ut1_str();
}

/**
 * Write out the given time, as UT/LST
 */
std::ostream& operator<<(std::ostream &strm, const AstroTime& a)
{
    return strm << a.str() << "/" << Angles::r2hms(a.lmst());
}
#endif

