#include <math.h>
#include "buffered-serial.h"
#include "exception.h"
#include "qp.h"

#if PROVARINO
#include <iostream>
#endif

// Implementation: the goal of this class, as far as output is
// concerned, is to avoid being hung up in writing out a long(ish)
// string, and so miss interrupts.  We do this by simply avoiding
// writing to the Serial port if there is no space left in its
// associated buffer, and if the line being output starts with a `#`.
//
// An earlier version of this class maintained its own circular buffer.
// However this is unnecessary, since the HardwareSerial class
// maintains such a buffer as well, from which it outputs bytes on an
// interrupt-driven schedule (see HardwareSerial.cpp in
// <https://github.com/arduino/Arduino/tree/master/hardware/arduino/avr/cores/arduino>).

/**
 * Create a new BufferedSerial object, and initialise the underlying
 * serial port.
 */
BufferedSerial::BufferedSerial()
{
    n_discards = 0;
    line_type = unknown;
}

/**
 * Initialise the buffered-serial object, after the program is under way.
 * This can't be done at constructor time
 * (it's not completely clear why not, but it's plausibly because there's some
 * run-time initialisation that the Arduino runtime does before calling `setup()`).
 */
void BufferedSerial::init(void)
{
    Serial.begin(19200);
}

// Reading

// the limit MAXITEMS must be the largest number that state.n can hold
// (which is 5, for a 4-bit wide field)
#define MAXITEMS 15

// Give me that old-time bit-twiddling.
//
// bit 9 : the item buffer is full
//     8 : EOL found -- line can be read
//     7 : the number has been negated
//     6 : \
//     5 : number of decimal places -- in range [0..7]
//     4 : /
//     3 : \
//     2 : number of items in list -- in range [0..15]
//     1 :  |
//     0 : /
static unsigned int state = 0;
#define STATE_N			(state & 0xf)
#define STATE_N_INC		(state++)
#define STATE_DP		((state & (7<<4)) >> 4)
#define STATE_DP_INC		(state += 16)
#define STATE_DP_ZERO		(state &= ~(7<<4))
#define STATE_NEGATIVE		(state & (1<<7))
#define STATE_NEGATIVE_SET	(state |= (1<<7))
#define STATE_NEGATIVE_CLEAR	(state &= ~(1<<7))
#define STATE_LINE_READY_P	(state & (1<<8))
#define STATE_LINE_READY_P_SET	(state |= (1<<8))
#define STATE_FULL_P		(state & (1<<9))
#define STATE_FULL_P_SET	(state |= (1<<9))
static InputItem item[MAXITEMS+1] = { { ITEM_INVALID, {0} }};

/**
 * Retrieve items from the Serial input.
 * @return a pointer to the array of items, or NULL if a complete line
 * has not yet been read
 */
InputItem* BufferedSerial::read_items_get(int* n)
{
    if (STATE_LINE_READY_P) {
        *n = STATE_N;
        return item;
    } else {
        return NULL;
    }
}

/**
 * Reset the buffer of read items.
 * @see read_items_get
 */
void BufferedSerial::read_items_reset(void)
{
    state = 0;
    item[0].type = ITEM_INVALID;
}

static void finish_item()
{
    if (STATE_FULL_P) {
        return;
    }

    switch (item[STATE_N].type) {
      case ITEM_CH:
        break;
      case ITEM_INT:
        if (STATE_NEGATIVE)
            item[STATE_N].v.i *= -1;
        break;
      case ITEM_FLOAT:
        {
            float factor;
            if (STATE_NEGATIVE) {
                factor = -1.0;
            } else {
                factor = 1.0;
            }
            for (int dp = STATE_DP; dp>0; dp--) {
                factor *= 10;
            }
            item[STATE_N].v.f = item[STATE_N].v.i / factor;
        }
        break;
    }
    STATE_NEGATIVE_CLEAR;
    STATE_DP_ZERO;

#if DEBUG
    switch (item[STATE_N].type) {
      case ITEM_INVALID:
        printf(">> --\n");
        break;
      case ITEM_CH:
        printf(">> %c\n", item[STATE_N].v.c);
        break;
      case ITEM_INT:
        printf(">> %d\n", item[STATE_N].v.i);
        break;
      case ITEM_FLOAT:
        printf(">> %g\n", item[STATE_N].v.f);
        break;
    }
#endif

    // increment STATE_N
    if (STATE_N == MAXITEMS) {
        // ...or not
        STATE_FULL_P_SET;
    } else if (item[STATE_N].type == ITEM_INVALID) {
        // do nothing -- there's nothing here, and we shouldn't
        // increment state.n
    } else {
        // normal case
        STATE_N_INC;
        item[STATE_N].type = ITEM_INVALID;
    }
}

/**
 * Read a character from Serial, and return the number of bytes actually read
 * @see read_items_get
 */
int BufferedSerial::read_items_read_char(void)
{
    int ch;

    if (STATE_LINE_READY_P) {
        // do nothing
        logprintf(PSTR("read_items_read_char: state=%x"), state);
        return 0;
    }

    ch = Serial.read();

    if (ch < 0) {
        // we shouldn't have been called in this case, but return zero anyway
        return 0;
    }

#if DEBUG
    if (isprint(ch))
        printf("ch=%c", ch);
    else
        printf("ch=0x%x", ch);
    printf(": neg=%d  n=%d  dp=%d  ready?=%d  full?=%d --> item[%d] = %d, i=%d\n",
           STATE_NEGATIVE, STATE_N, STATE_DP, STATE_LINE_READY_P, STATE_FULL_P,
           STATE_N, item[STATE_N].type, item[STATE_N].v.i);
#endif

    int current_type = item[STATE_N].type;

    if (ch == '!') {
        // emergency stop -- discard everything, and jump to the fault handler in qp.ino
        state = 0;
#if PROVARINO
        // we're in a testing environment
        throw "found ! in input";
#else
        //fault_exception(0);     // non-fatal exception
        recoverable_error(PSTR("emergency stop"));
#endif
    } else if (ch == '\n' || ch == '\r') {
        // end of line
        // (ignore this if we're at the beginning of the line)
        if (STATE_N != 0) {
            finish_item();
            STATE_LINE_READY_P_SET;
        }

    } else if (STATE_FULL_P) {
        // we are about to add item MAXITEM+1,
        // but there's no more space -- do nothing

    } else if (isalpha(ch)) {
        finish_item();

        if (! STATE_FULL_P) {
            item[STATE_N].v.c = ch;
            item[STATE_N].type = ITEM_CH;
            finish_item();
        }

    } else if (ch == '-') {
        if (current_type != ITEM_INVALID) {
            // We're in some sort of item, but this can't be a part of
            // it, so end that one.
            // This actually accepts "1.0-1.0" as two numbers
            // (which we arguably oughtn't to accept, but let's be nice)
            finish_item();
        }
        item[STATE_N].type = ITEM_INT;
        item[STATE_N].v.i = 0;
        STATE_NEGATIVE_SET;

    } else if (ch == '.') {
        if (current_type == ITEM_INVALID) {
            item[STATE_N].v.i = 0;
        }
        item[STATE_N].type = ITEM_FLOAT;
        STATE_DP_ZERO;

    } else if (isdigit(ch)) {
        if (current_type == ITEM_INVALID) {
            item[STATE_N].type = ITEM_INT;
            item[STATE_N].v.i = 0;
        }
        item[STATE_N].v.i = 10 * item[STATE_N].v.i + (ch - '0');
        if (item[STATE_N].type == ITEM_FLOAT) {
            STATE_DP_INC;
        }

    } else {
        // some other character
        finish_item();
    }

    return 1;
}

/**
 * Return true if there is at least one byte available to be read.
 */
bool BufferedSerial::read_available(void)
{
    return Serial.available() > 0;
}

// Writing

/**
 * The logprintf function supports a simple printf-like syntax for
 * writing data to the Serial port.  The format string must be in
 * program memory -- that is, as an argument to the PSTR macro:
 * PSTR("format").
 *
 * The format specifiers are:
 *
 *  spec | explanation
 *  -----|------------
 *    %% | print out a percent sign
 *    %c | a character; non-printing characters are displayed as hex
 *    %l | a long int (such as the return from millis()
 *    %d | an ordinary int, printed in decimal
 *    %x | an int, printed in hex
 *    %s | a PSTR string
 *    %g | a float
 *    %G | a float, but with more decimal places printed
 *
 * For convenience, the function writes a newline to the port at the
 * end of the format string.  If you wish to suppress this, then
 * include the 'format specifier' `%|` at the end of the string.
 *
 * There is minimal error-checking on the format specifier.
 */
void BufferedSerial::logprintf(PGM_P format, ...)
{
    va_list ap;

    va_start(ap, format);
    logvprintf(format, ap);
    va_end(ap);
}

/**
 * The logvprintf is the varargs companion to the logprintf
 * function.  Once a varargs argument pointer is set up, it can be
 * passed to this function for handling.  The `logprintf` function is
 * equivalent to:
 *
 *     void Util::logprintf(PGM_P format, ...)
 *     {
 *         va_list ap;
 *
 *         va_start(ap, format);
 *         logvprintf(format, ap);
 *         va_end(ap);
 *     }
 *
 * @see logprintf
 */
void BufferedSerial::logvprintf(PGM_P format, va_list ap)
{
    char c;
    PGM_P s;

    bool print_trailing_newline_p = true;

    c = pgm_read_byte_near(format);
    if (line_type == unknown) {
        if (c == '#')
            line_type = comment;
        else
            line_type = non_comment;
    }

    for (s = format; (c = pgm_read_byte_near(s)) != '\0'; s++) {
        if (c == '%') {
            c = pgm_read_byte_near(++s);
            switch (c) {
              case '%':
                append_byte('%');
                break;
              case '|':
                print_trailing_newline_p = false;
                break;

              case 'c':         // character
                {
                      char ac = va_arg(ap, int); // int, because char arg is promoted
                      if (isprint(ac)) {
                          char cc = ac;
                          append_byte(cc);
                      } else if (ac < 0) {
                          print_string("\\0xff");
                      } else {
                          print_string("\\0x");
                          print_int(ac, 16);
                      }
                }
                break;
              case 'l':         // unsigned long int (ie, ms_t)
                {
                    unsigned long int lu = va_arg(ap, unsigned long);
                    print_int(lu, 10);
                }
                break;
              case 'g':
                {
                    float f = va_arg(ap, double);
                    print_float(f);
                }
                break;
              case 'G':
                {
                    float f = va_arg(ap, double);
                    print_float(f, 6);
                }
                break;
              case 's':
                {
                    PGM_P sp;
                    for (sp = va_arg(ap, PGM_P);
                         (c = pgm_read_byte_near(sp)) != '\0';
                         sp++) {
                        append_byte(c);
                    }
                }
                break;
              case 'x':
                {
                    unsigned int i = va_arg(ap, unsigned int);
                    print_string("0x");
                    print_int(i, 16);
                }
                break;
              case 'd':         // FALL THROUGH
              default:          // erm.... just take it to be an int and hope for the best
                {
                    int i = va_arg(ap, int);
                    print_int(i, 10);
                }
                break;
            }
        } else {
            append_byte(c);
        }
    }

    if (print_trailing_newline_p) {
        append_byte('\n');
        line_type = unknown;
    }
}

void BufferedSerial::print_simple_float(const float farg, const int dp)
{
    // This is not the most sophisticated print-float function in the
    // world, but it is sufficient for our present purposes.

    float f = 0.5;
    for (int i=0; i<dp; i++)
        f /= 10.0;
    f += farg;
    int fi = floor(f);
    print_int(fi, 10);
    float frac = f - fi;

    int trailing_zeros = 0;     // suppress trailing '0' characters
    bool shown_point = false;
    if (frac != 0.0) {
        for (int dpx=dp; dpx>0; dpx--) {
            frac *= 10.0;
            int fx = floor(frac);

            if (fx == 0) {
                trailing_zeros++;
            } else {
                if (! shown_point) {
                    append_byte('.');
                    shown_point = true;
                }
                for (int i=0; i<trailing_zeros; i++)
                    append_byte('0');
                trailing_zeros = 0;
                append_byte(fx + '0');
                frac -= fx;
            }
        }
    }
}

void BufferedSerial::print_float(const float f, const int dp)
{
    double fx = fabs(f);

    if (fx < 1e4 && fx >= 0.001) {
        if (signbit(f)) {
            append_byte('-');
        }
        print_simple_float(fx, dp);

    } else if (fx == 0.0) {
        print_string("0.0");

    } else {
        if (signbit(f))
            append_byte('-');
        float l = log10(fx);
        int exponent = floor(l);
        float mantissa = l - exponent;

        print_simple_float(pow(10, mantissa), dp);
        append_byte('e');
        print_simple_float(exponent, 2);
    }
}

void BufferedSerial::print_string(const char* s)
{
    while (*s != '\0') {
        append_byte(*s++);
    }
}

void BufferedSerial::print_int(const int iarg, const int base)
{
    if (iarg == 0) {
        append_byte('0');
    } else {
        int i = iarg;
        if (i < 0) {
            append_byte('-');
            i *= -1;
        }
        byte ib[10];                // 10^10 > 2^32
        int idx;
        for (idx=0; i>0; idx++) {
            ib[idx] = i%base;
            i /= base;
        }
        for (idx--; idx>=0; idx--) {
            if (ib[idx] > 9) {
                append_byte(ib[idx] + 'a' - 10);
            } else {
                append_byte(ib[idx] + '0');
            }
        }
    }
}

// We append a byte to the buffer which the HardwareSerial object (sc
// Serial) maintains.  This buffer is cleared by IRQs provided by that
// class.  The (apparently undocumented) method availableForWrite()
// (see HardwareSerial.cpp in
// <https://github.com/arduino/Arduino/tree/master/hardware/arduino/avr/cores/arduino>)
// indicates how much space is available in that buffer.  If there is
// none, and this is a discardable line, then discard this byte.  If
// this is not a discardable line, then the Serial.write will block
// until the buffer is emptied.
void BufferedSerial::append_byte(const byte b)
{
    if (line_type == comment && Serial.availableForWrite() == 0) {
        n_discards++;
    } else {
        Serial.write(b);
    }
}

/**
 * Append a single byte to the output.
 */
int BufferedSerial::putchar(const int c)
{
    const byte b = (const byte)c;
    append_byte(b);
    return c;
}

/**
 * Return the number of bytes discarded since the last call to this
 * method.  If {@link logprintf} fills up the buffer, extra bytes are
 * silently discarded.
 */
int BufferedSerial::count_discarded(void)
{
    int n = n_discards;
    n_discards = 0;
    return n;
}

#if PROVARINO
/**
 * Send any contents of the Serial buffer to stdout.
 * This is a convenience method, for testing, and depends on the
 * non-standard `get` method in the Serial implementation in
 * `simple_stream.h`.
 *
 * Having this available means that even the test methods don't
 * have to refer to, or declare, the Serial object explicitly.
 */
void BufferedSerial::display_serial_contents(void)
{
    if (Serial.available_out()) {
        std::cout << ">> " <<  Serial.get();
    }
}

// A replacement for logprintf for testing
// (thought it doesn't handle the %G format).
// Use as "#define logprintf BufferedSerial::testenv_logprintf"
void BufferedSerial::testenv_logprintf(PGM_P format, ...)
{
    va_list ap;
    va_start(ap, format);
    vprintf(format, ap);
    va_end(ap);
    fputc('\n', stdout);
}
#endif
