#ifndef BUFSERIAL_H_LOADED
#define BUFSERIAL_H_LOADED 1

#include <stdarg.h>

#include "types.h"

// The actual serial device
//extern HardwareSerial Serial;
//extern Serial_ Serial;
extern UARTClass Serial;


#define ITEM_CH 0
#define ITEM_INT 1
#define ITEM_FLOAT 2
#define ITEM_INVALID 3
/**
 * The items read from the serial port.
 * @see read_items_read_char
 * @see read_items_get
 * @see read_items_reset
 */
struct InputItem {
    /**
     * The type of the item, which is one of the types `ITEM_CH`,
     * `ITEM_INT`, `ITEM_FLOAT` or `ITEM_INVALID`.
     */
    unsigned char type;
    /** The types of data that may be in an input item */
    union item_data {
        char c;             ///< a character
        int i;              ///< an integer
        float f;            ///< a float
    };
    /** The value of the item. */
    union item_data v;
};

#define SERIALBUFSIZ 128

/**
 * The BufferedSerial class mediates access to an underlying serial
 * device.
 *
 * For input, it provides the following methods to support reading
 * items from the serial bus.
 *
 *     int read_items_read_char(void);
 *     InputItem* read_items_get(int* n);
 *     void read_items_reset(void);
 *
 * Whenever `read_available()` is true, call `read_items_read_char`:
 * this takes one byte off the bus, and cumulatively parses these bytes
 * into an array of arguments of type character, integer or float.
 * Before a complete line is available (terminated by `\n`), the
 * function `read_items_get` returns NULL.  After a complete line is available,
 * `read_items_get` will return non-NULL, providing a pointer to the
 * array of InputItem structs, and it will stop accumulating bytes
 * from the bus.  Subsequent calls to the function will return the
 * same non-NULL pointer.  Call `read_items_reset` to clear this, and
 * start accumulating bytes again.
 *
 * These functions will parse a sequence of single characters
 * `/[a-zA-Z]/`, integers `/-?[0-9]+/` and floats `/-?[0-9]*\.[0-9]/` from
 * the Serial port, separated by any other characters, which are
 * ignored.  Thus `aB123;4.    .5   \n` parses as ['a', 'B', 123, 4.0, 0.5]
 *
 * For output, there are the function
 *
 *     void logprintf(PGM_P format, ...);
 *
 * This formats a string to an internal circular buffer, which the
 * underlying Serial library sends to the output on an
 * interrupt-driven schedule.
 *
 * If the output buffer becomes full, extra writes to it are
 * discarded if the string starts with a `#` character.
 * The method {@link count_discarded} returns the number
 * of bytes thus discarded since the last call to the method.
 *
 * @see read_items_read_char
 * @see read_items_get
 * @see read_items_reset
 * @see logprintf
 */
class BufferedSerial {
private:
    enum { unknown, comment, non_comment } line_type;
    int n_discards;
    void print_simple_float(const float f, const int dp);
    void print_float(const float f, const int dp=3);
    void print_string(const char* s);
    void print_int(const int i, const int base);
    void append_byte(const byte);

 public:
    BufferedSerial();
    void init(void);
    void logprintf(PGM_P format, ...);
    void logvprintf(PGM_P format, va_list ap);
#undef putchar
    int putchar(const int);

    int read_items_read_char(void);
    InputItem* read_items_get(int* n);
    void read_items_reset(void);
    bool read_available(void);

    int count_discarded(void);

#if PROVARINO
    void display_serial_contents(void);
    static void testenv_logprintf(PGM_P format, ...);
#endif
};

#endif /* BUFSERIAL_H_LOADED */
