/**
 * Configuration constants for qp.ino.
 *
 * Specify approximate (at least, sane) values for the various
 * calibration and position parameters, so that things don't break if they're not set.
 */

#ifndef CONFIG_QP_H_LOADED
#define CONFIG_QP_H_LOADED 1

/**
 * Observatory position: Glasgow is 55d54m8sN, -4d18m26sE.
 * These are the 'actual' coordinates; they are distinguished from the
 * 'effective' coordinates, which take into account misalignments of
 * the telescope axis.
 *
 * See qp.ino:EnvironmentInformation::obs_lat and friends, and the 'O'
 * calibration command.
 */
#define CONFIG_OBS_LAT_ACTUAL 0.975677834
#define CONFIG_OBS_LON_ACTUAL -0.075175170
#define CONFIG_OBS_LAT_EFFECTIVE 0.975677834
#define CONFIG_OBS_LON_EFFECTIVE -0.075175170

/**
 * The initial values for the drive calibration, in rad/s.  These aren't
 * crucial, since it is the job of the Kalman filters to get these
 * correct, but these should be set to be approximately correct, so
 * that those filters converge on a good value rapidly.  An
 * underestimate is better than an overestimate.
 */
#define CONFIG_MAX_DRIVE_AZ 0.03
#define CONFIG_MAX_DRIVE_ALT 0.03

/**
 * The PWM drive is potentially somewhat non-linear at low values, so
 * set a lower limit on the commanded speed, as a fraction of full speed.
 */
#define CONFIG_MIN_SPEED 0.1

/**
 * We detect that the drive has reached its end-stop, when we get no
 * position clicks for ENDSTOP_TIMEOUT ms.  There's no problem
 * continuing to power the drive when it's at the end-stop, since the
 * switch there diconnects the drive when it arrives.
 *
 * This is a slightly unreliable way of detecting arrival at the
 * endstop, since it can be deceived if the drive is simply moving too
 * slowly.  So set this to be longer than the time the drive would
 * take to get from one click-point to another, when moving at the
 * minimum speed: that is 1deg / (CONFIG_MIN_SPEED (/s) *
 * CONFIG_MAX_DRIVE_<X> (rad) * 57 (deg/rad)) =~6s.
 *
 * Since the above CONFIG_MAX_DRIVE_<X> are underestimates, this should
 * be an overestimate of the minimum necessary to avoid a false
 * detection of the endstop.
 */
#define CONFIG_ENDSTOP_TIMEOUT 6000

/**
 * Home position.  This position is fully east, bearing about 26deg,
 * and 6deg above horizontal.
 * Fully west ('antihome') is about bearing 5deg.
 * The units here are radians in azimuth and altitude.
 *
 * See qp.ino:EnvironmentInformation::home_position_phi and friends,
 * and the 'O' calibration command.
 *
 * The azimuths must be in [0,2pi], and the altitude in [-pi,pi].
 */
#define CONFIG_HOME_POSITION_AZIMUTH 0.452
#define CONFIG_HOME_POSITION_ALTITUDE 0.105
#define CONFIG_ANTIHOME_POSITION_AZIMUTH 0.092

/**
 * How close may the telescope come to its endstop, in radians.
 * The value of 0.0087 rad is 0.5deg.
 * See TelescopeController::endstop_clearance
 */
#define CONFIG_ENDSTOP_CLEARANCE 0.0087

/**
 * Measurement error for drive positions (the clicks) (in rad).  An
 * estimate of the error in the actual positions of the reed switches,
 * or the error in the actual telescope pointing when it is at one of
 * the reed switches (ie, slop).  The precise value of this isn't
 * crucial, but it sets the scale for the sensitivity of the Kalman
 * filters.  This should be a comfortable over-estimate, as that keeps
 * the filter from being over-sensitive.
 *
 * The real value is probably something like 0.002 rad, or 0.1 degree.
 */
#define CONFIG_DRIVE_SLOP 0.004

/**
 * How close are we willing to go to the zenith?
 * Going closer than this could potentially, as things currently
 * stand, produce an error when converting to quaternions.  This delta
 * of 0.01 radians is about half a degree.  This figure is (again,
 * currently) a guess, and should probably be larger than the largest
 * likely error in an angular Kalman filter.
 */
//#define CONFIG_MAX_ALTITUDE (M_PI_2-0.01)

// maximum altitude is right over the top
// FIXME: document this better
#define CONFIG_MAX_ALTITUDE (M_PI-0.01)

/**
 * Process variance in the telescope position (in rad^2).  It's not
 * completely clear what this should be, but about 0.1deg (~0.002 rad)
 * seems about right
 */
#define CONFIG_ANGLE_PROCESS_VARIANCE 4e-6

/**
 * Process variance of the telescope velocity.  This is the square of
 * the fraction of the maximum velocity, which should be taken to be
 * the amount of variance in the actual vs commanded velocity.  Thus
 * 10% (giving a value here of 0.01) would indicate that the telescope
 * varies by +- 10% from its commanded velocity.  A value of 10% would
 * be very pessimistically, but it's OK to be at least slightly
 * pessimistic here.
 */
#define CONFIG_VELOCITY_FRACTIONAL_VARIANCE 0.001


/**
 * Maximum estimated drive velocity, in rad/s.
 * The telescope drives can nominally do a full rotation in 60--120s
 * (depending on voltage), which is 6--12 rad/min, or 0.1 to
 * 0.2 rad/sec.
 *
 * Note: 0.1rad/s = 0.1mrad/ms = 6 rad/min
 */
#define CONFIG_MAXIMUM_DRIVE_VELOCITY 0.2

/** The telescope advances half a degree per click in azimuth. */
#define RAD_PER_CLICK_AZ (0.5 * M_PI / 180.0)

/** the telescope advances half a degree per click in altitude. */
#define RAD_PER_CLICK_ALT (0.5 * M_PI / 180.0)

#endif /* CONFIG_QP_H_LOADED */
