#include "kalman.h"
#include "config-qp.h"

// For detailed notes, see doc/qp/kalman-filter.tex

#ifndef DEBUG
// The debug code below is intended only to be usable within PROVARINO
#define DEBUG 0
#endif

#include "buffered-serial.h"
#if PROVARINO
#define logprintf BufferedSerial::testenv_logprintf
#else
extern BufferedSerial bs;
#define logprintf bs.logprintf
#endif

#if DEBUG & PROVARINO
#include <iostream>

using std::endl;
#define COUT std::cout << __FILE__ << ':' << __LINE__ << ": "
#endif

/**
 * Create a one-dimensional Kalman filter for angular position.  The
 * filter manages estimates for the position and the angular reference
 * speed.  This speed is the speed in rad/s which corresponds to
 * relative speed 1.0.
 *
 * The filter is reasonably insensitive to the initial values chosen here;
 * indeed part of the point of it is to improve these and their
 * estimated errors.  The filter does not (currently) maintain an
 * estimate of the error `dz`.  Note that if this `dz` error is
 * estimated too small, the filter becomes rather unstable.
 *
 * The angle is not normalised to stay in any particular range.
 *
 * @param theta the initial estimate of the angle (rad)
 * @param maxthetadot the initial estimate of the reference speed
 *   (rad/s), corresponding to commanded speed 1.0
 * @param dz an estimate of the error in the actual positions of the
 *   reed switches (rad)
 * @param initial_time the start time of the filter (ms)
 */
Kalman::Kalman(double theta, double maxthetadot,
               double dz,
               ms_t initial_time)
    : theta_(theta),
      vmax_pos_(maxthetadot/1000.0), vmax_neg_(maxthetadot/1000.0),
      dtheta2_(CONFIG_ANGLE_PROCESS_VARIANCE),
      dvmax2_(1e-6*maxthetadot*maxthetadot*CONFIG_VELOCITY_FRACTIONAL_VARIANCE), // (maxthetadot/1000)^2 * var'c
      dz2_(dz*dz),
      p00_(dtheta2_), p01_(0), p10_(0), p11_(dvmax2_),
      t_(initial_time),
      logchar_('\0')
{
    // empty
}

/**
 * No-arg constructor.
 */
Kalman::Kalman()
    : theta_(0), vmax_pos_(0), vmax_neg_(0), dtheta2_(0), dvmax2_(0), dz2_(0),
      p00_(0), p01_(0), p10_(0), p11_(0), t_(0)
{
    // empty
}

/**
 * Set the current estimate of the angle.  This is not an update of
 * the position estimate, but an unequivocal reset of the value.
 *
 * @param ang the current angle
 */
void Kalman::reset_angle(double ang)
{
    theta_ = ang;
}

/**
 * Update the position.  The provided speed is taken to have been
 * constant over the interval from the last update to now.  This is
 * equivalent to the {@link click} method but with an infinite
 * measurement variance, and hence a trivial (effectively absent) corrector step.
 *
 * @param t the current time (ms)
 * @param speed the current speed (fractional speed)
 */
void Kalman::update(ms_t t, double speed)
{
    ms_t delta_t = t - t_;
    double omegat = speed * delta_t;

    if (speed > 0)
        theta_ += omegat * vmax_pos_;
    else
        theta_ += omegat * vmax_neg_;
    // vmax_ unchanged

    p00_ += omegat*(p01_ + p10_ + omegat*p11_) + dtheta2_;
    p01_ += omegat*p11_;
    p10_ += omegat*p11_;
    p11_ += dvmax2_;

    if (logchar_ != '\0') {
        logprintf(PSTR("%cu,%g,%g,%g,  " MS_FMT),
                  logchar_, theta_,
                  vmax_pos_, vmax_neg_,
                  t);
    }

    t_ = t;

    // The following happens too frequently for the debugging output to be useful
// #if DEBUG
//     logprintf(PSTR("#kalman update: t=%g theta=%g vmax=%g/%g p00=%g p01=%g p10=%g p11=%g"),
//               t/1000.0, theta_,
//               vmax_pos_, vmax_neg_,
//               p00_, p01_, p10_, p11_);
// #endif
}

/**
 * Update the position with a position measurement.
 * The provided speed is taken to have been constant over the interval
 * from the last update to now.
 *
 * @param t the current time (ms)
 * @param speed the current speed (fractional speed)
 * @param ref_angle the current actual angle (rad)
 * @see update(ms_t,double)
 */
void Kalman::click(ms_t t, double speed, double ref_angle)
{
    ms_t delta_t = t - t_;    // interval since last click

    double A = p00_ + dz2_;              // Kalman gain denominator
    double zctheta = ref_angle - theta_; // change in angle since last click
    double omegat = speed * delta_t;     // \omega\Delta t

    // update/predictor step
    if (speed > 0)
        theta_ += omegat * vmax_pos_;   // x_{t|t-1}
    else
        theta_ += omegat * vmax_neg_;
    // vmax_ is unchanged

    p00_ += omegat*(p01_ + p10_ + omegat*p11_) + dtheta2_;
    p01_ += omegat*p11_;
    p10_ += omegat*p11_;
    p11_ += dvmax2_;

    // corrector step
    theta_ += p00_ * zctheta /A;
    if (speed > 0) {
        double newval = vmax_pos_ + p10_ * zctheta /A;
        if (newval < 0) {
            logprintf(PSTR("!Kalman tried to flip sign"));
        } else {
            vmax_pos_ = newval;
        }
    } else {
        double newval = vmax_neg_ + p10_ * zctheta /A;
        if (newval < 0) {
            logprintf(PSTR("!Kalman tried to flip sign"));
        } else {
            vmax_neg_ = newval;
        }
    }

    float inc00 = p00_*p00_ /A;
    float inc01 = p00_*p01_ /A;
    float inc10 = p10_*p00_ /A;
    float inc11 = p10_*p01_ /A;

    p00_ -= inc00;
    p01_ -= inc01;
    p10_ -= inc10;
    p11_ -= inc11;

    if (logchar_ != '\0') {
        logprintf(PSTR("%cc,%g,%g,%g,  %g,   " MS_FMT "," MS_FMT ",  %g"),
                  logchar_, theta_, vmax_pos_, vmax_neg_, A,
                  t, delta_t, ref_angle);
    }

#if DEBUG
    logprintf(PSTR("#kalman click: t=%g angle=%g vmax=%g/%g"),
              t/1000.0, theta_,
              vmax_pos_, vmax_neg_);
    // logprintf(PSTR("#kalman click: t=%g theta=%g vmax=%g/%g p00=%g p01=%g p10=%g p11=%g a=%g"),
    //           t/1000.0, theta_,
    //           vmax_pos_, vmax_neg_,
    //           p00_, p01_, p10_, p11_, A);
#endif

    t_ = t;
}
