#ifndef KALMAN_H_LOADED
#define KALMAN_H_LOADED 1

#include "types.h"

/**
 * A one-dimensional Kalman filter, which has as its state variables
 * position, measured in radians from some implied reference point,
 * and a reference speed, in radians per unit time, which is the
 * actual maximum speed when the telescope motors are full on (ie,
 * corresponding to commanded speed 1.0).
 *
 * The 'speeds' provided to this class, in the range [-1.0,1.0], are
 * relative to this reference speed.
 */
class Kalman {
    // This implementation surreptitiously manages _two_ Kalman filters,
    // one for movement in the positive direction, and another for
    // movement in the negative direction (since the elevation drive moves
    // at different speeds going and and going down).  We update
    // either vmax_pos_ or vmax_neg_ in the two cases.
    // This should be transparent to the caller.

private:
    /** Position, in rad */
    double theta_;
    /** Reference speed (the speed in rad/millisecond at maximum +ve motor speed) */
    double vmax_pos_;
    /** Reference speed (the absolute speed in rad/ms at maximum -ve motor speed) */
    double vmax_neg_;
    /** Uncertainty in theta, in radians, squared */
    double dtheta2_;
    /** Uncertainty in vmax, squared (rad/ms)^2.  This characterises how much
        the actual driven maximum speed will vary in practice. */
    double dvmax2_;
    /** Uncertainty in click position, in units of radians^2 */
    double dz2_;

    // Internal covariance matrix
    double p00_, p01_, p10_, p11_;

    ms_t t_;                    // last update time

    char logchar_;              // if not '\0', then be chatty

public:
    Kalman();
    Kalman(double theta,        // initial position
           double maxthetadot,  // initial estimate of actual speed (rad/s) for commanded speed 255
           double dz,
           ms_t initial_time);
    // in both of the following, speed is given as a fraction of full speed
    void update(ms_t t, double speed);
    void click(ms_t t, double speed, double click_angle);

    void reset_angle(double ang);

    /** The current estimate of the angle */
    double angle(void) const { return theta_; }
    /** The current estimate of the reference/maximum speed, when driven in the positive direction */
    double fullspeed_pos(void) const { return vmax_pos_*1000.0; }
    /** The current estimate of the reference/maximum speed, when driven in the negative direction */
    double fullspeed_neg(void) const { return vmax_neg_*1000.0; }
    // are the following the best returns, here?
    /** An estimate of the error in angular position, in units of radians-squared */
    double angle_error(void) const { return p00_; }
    /** An estimate of the error in reference speed */
    double fullspeed_error(void) const { return p11_; }

    /**
     * Enable tracing of the behaviour of the Kalman filter.  After
     * this method is called with a character argument, updates to the
     * filter result in a tracing line written to the serial port,
     * which starts with that character.  The format of these is not
     * currently documented here: see the code for details.
     */
    void verbose(char verbose_char) { logchar_ = verbose_char; }
};


#endif /* KALMAN_H_LOADED */
