#include "position.h"
#include "quaternion.h"
#include "angles.h"
#include "exception.h"

#include <math.h>


/**
 * Create a new pair of angles.  The `lon` and `lat` attributes are a
 * pair of angles giving longitude and latitude, in radians.
 *
 * There is _limited_ support for converting between systems, the
 * set of possible 'conversions' is rather ad-hoc, and any
 * impermissible ones call `fatal()`.
 *
 * The `system` argument to this constructor indicates the coordinate
 * system in which the two angle arguments are expressed, and the
 * `into_system` argument indicates the system of the resulting
 * object.  If the `into_system` argument is not provided, then both
 * `TopoPolar` and `AzAlt` coordinates are turned into `TopoPolar`,
 * and both `HADec` and `EqPolar` arguments into `HADec`.  If the fourth
 * argument is present, then `TopoPolar` and `AzAlt` coordinates are
 * interconvertible, as are `HADec` and `EqPolar`.  For conversions
 * between those pairs, see {@link #to_system}.
 *
 * After these conversions, the `.lon` and `.lat` values are
 * initialised in their appropriately adjusted form.
 * Thus (writing degrees instead of radians, for convenience):
 * after Position(10, 30, AzAlt), the .lon is +170 and .lat is +30;
 * after Position(200, 30, AzAlt), the .lon is +340;
 * after Position(-10, 30), the .lon is +350;
 * after Position(10, 30, HADec), the .lon is +350 and .lat is +30.
 *
 * In all cases, the longitude is normalised into [0,2pi), and the
 * latitude is forced to be in [-pi/2,pi/2].
 *
 * FIXME: Tidyups/refactoring: the Position class is now over-complicated, with
 * far too much fiddly branching.  Or is it?  Things like TrackingSchedule
 * need specific documentation concerning which coordinate system the arguments
 * are in, with the result that rotation-by-colatitude happens in multiple
 * places (and I _always_ get the signs wrong first time).  I can't help
 * feeling these should be in a single place, but every time I conclude
 * that single place is 'in the Position class', I end up with a mess in here.
 *
 * @param lon the longitudinal angle in the specified system
 * @param lat the latitudinal angle in the specified system
 * @param system the system in which the above two arguments are expressed
 * @param into_system the system in which the resulting Position object holds its coordinates
 */
Position::Position(double lon, double lat, System system, System into_system)
    : obs_lat_(-999), obs_lmst_(-999)
{
    double pipi = 2*M_PI;

    //std::cout << "Position(" << lon << ',' << lat << "; " << system << ',' << into_system << std::endl;

    // Normalise the latitude:
    //
    // We expect the latitude to be in [-pi/2,+pi/2], but since the
    // telescope can physically flip over the top, this constructor
    // can on occasion be reasonably given a 'latitude' bigger than
    // +pi/2 (we _really_ shouldn't get latitudes below zero, much
    // less in [-pi/2,-pi], but let's not fail in that case).
    while (lat > M_PI)
        lat -= pipi;
    while (lat < -M_PI)
        lat += pipi;
    if (lat > M_PI_2) {
        lat = M_PI - lat;
        lon = M_PI + lon;
    }

    // normalise the longitude into [0,2pi]
    while (lon < 0.0)
        lon += pipi;
    while (lon >= pipi)
        lon -= pipi;

    bool ok_p = false;

    switch (into_system) {
      case _Natural:
        {
            switch (system) {
              case AzAlt:
                lon_ = M_PI - lon;
                if (lon_ < 0.0) {
                    lon_ += pipi;
                }
                lat_ = lat;
                system_ = TopoPolar;
                break;

              case HADec:
                lon_ = pipi - lon; // local hour angle is -longitude
                lat_ = lat;
                system_ = EqPolar;
                break;

              case Velocity:
                lon_ = lon;
                lat_ = lat;
                system_ = Velocity;
                break;

              case _Natural:
                lon_ = lon;    // no change
                lat_ = lat;
                system_ = TopoPolar;
                break;

              default:
                lon_ = lon;    // no change
                lat_ = lat;
                system_ = system;
                break;
            }
            ok_p = true;
        }
        break;

      case TopoPolar:
      case AzAlt:
        if (system == TopoPolar || system == AzAlt) {
            lon_ = (system == into_system ? lon : M_PI - lon);
            if (lon_ < 0.0) {
                lon_ += pipi;
            }
            lat_ = lat;
            system_ = into_system;
            ok_p = true;
        } else {
            ok_p = false;
        }
        break;

      case HADec:
      case EqPolar:
        if (system == HADec || system == EqPolar) {
            lon_ = (system == into_system ? lon : 2*M_PI - lon);
            lat_ = lat;
            system_ = into_system;
            ok_p = true;
        } else {
            ok_p = false;
        }
        break;


        switch (system) {
          case HADec:
            lon_ = 2*M_PI - lon;
            lat_ = lat;
            system_ = EqPolar;
            ok_p = true;
            break;
          case EqPolar:
            lon_ = lon;
            lat_ = lat;
            system_ = EqPolar;
            ok_p = true;
            break;
          default:
            ok_p = false;
            break;
        }
        break;

      case RADec:
      case Geog:
        if (system == into_system) {
            lon_ = lon;
            lat_ = lat;
            system_ = into_system;
            ok_p = true;
        } else {
            ok_p = false;
        }
        break;

      case Velocity:
        if (system == into_system) {
            lon_ = lon;
            lat_ = lat;
            system_ = into_system;
            ok_p = true;
        } else {
            ok_p = false;
        }
        break;

      default:
        ok_p = false;
    }

    if (! ok_p) {
        fatal(PSTR("Position: can't initialise system %d from %d"),
              into_system, system);
    }
}

Position::Position()
    : lon_(0), lat_(0), system_(TopoPolar), obs_lat_(-999)
{
    // empty: out-of-range obs_lat_ indicates sentinel
}

/**
 * Set the observatory position, for a subsequent call to {@link to_system}.
 * It is a fatal error if this is not set before a subsequent
 * conversion to AzAlt coordinates.
 *
 * The position must be in the `Geog` system.
 * @see Position::System
 */
void Position::set_observatory(const Position& obs)
{
    if (obs.system_ != Geog) {
        fatal(PSTR("Position::set_observatory must have a Geog position"));
    }
    obs_lat_ = obs.lat_;
}

/**
 * Set the reference time, for a subsequent call to {@link to_system}.
 * It is a fatal error if this is not set before a subsequent
 * conversion from RA-Dec coordinates.
 */
void Position::set_reftime(const AstroTime& t)
{
    obs_lmst_ = t.lmst();
}

/**
 * Return another Position, in the given system.
 * Depending on the conversion, it may be necessary to have called
 * {@link set_reftime} (for conversions involving the `RADec`
 * system) or {@link set_observatory} (for conversion to equatorial systems).
 * If such required information is needed but has not been set, this
 * results in a fatal error.
 */
Position Position::to_system(Position::System into_system) const
{
    Quaternion QJ = Quaternion(Quaternion::J);
    Position rval;

    //cout << "to_system: converting from " << *this << " to " << into_system << endl;

    if (system_ == into_system) {
        //rval = Position(lon_, lat_, into_system, system_);
        rval = *this;
    } else if (system_ == TopoPolar || system_ == AzAlt) {
        double polarlon = (system_ == TopoPolar ? lon_ : M_PI - lon_);
        switch (into_system) {
          case TopoPolar:
            rval = Position(polarlon, lat_, TopoPolar, TopoPolar);
            break;

          case AzAlt:
            rval = Position(M_PI-polarlon, lat_, AzAlt, AzAlt);
            break;

          case EqPolar:         // FALLTHROUGH
          case HADec:
            {
                if (obs_lat_ < -M_PI_2) {
                    fatal(PSTR("Can't convert horizontal->equatorial systems without observatory"));
                }
                Quaternion curr = Quaternion::vector(polarlon, lat_);
                Quaternion rot = QJ.rotation(M_PI_2 - obs_lat_);
                rval = curr.rotate(rot).coords(into_system, EqPolar);
            }
            break;

          default:
            fatal(PSTR("Can't convert TopoPolar/AzAlt into %d"), into_system);
        }
    } else if (system_ == EqPolar || system_ == HADec) {
        double polarlon = (system_ == EqPolar ? lon_ : 2*M_PI - lon_);
        switch (into_system) {
          case EqPolar:
            rval = Position(polarlon, lat_, EqPolar, EqPolar);
            break;
          case HADec:
            rval = Position(2*M_PI - polarlon, lat_, HADec, HADec);
            break;

          case TopoPolar:         // FALLTHROUGH
          case AzAlt:
            {
                if (obs_lat_ < -M_PI_2) {
                    fatal(PSTR("Can't convert equatorial->horizontal systems without observatory"));
                }
                Quaternion curr = Quaternion::vector(polarlon, lat_);
                Quaternion rot = QJ.rotation(-(M_PI_2 - obs_lat_));
                rval = curr.rotate(rot).coords(into_system, TopoPolar);
            }
            break;

          default:
            fatal(PSTR("Can't convert HADec/EqPolar into %d"), into_system);
        }
    } else if (system_ == RADec) {
        switch (into_system) {
          case RADec:
            rval = *this;
            break;

          case HADec:
          case EqPolar:
            if (obs_lmst_ < -M_PI_2) {
                fatal(PSTR("Can't convert RA-Dec to Equatorial without time information"));
            } else {
                Quaternion curr = Quaternion::vector(lon_, lat_);
                Quaternion rot = Quaternion(Quaternion::K).rotation(-obs_lmst_);
                rval = curr.rotate(rot).coords(into_system, EqPolar);
            }
            break;

          case TopoPolar:
          case AzAlt:
            if (obs_lmst_ < -M_PI_2) {
                fatal(PSTR("Can't convert RA-Dec to Azimuth without time information"));
            } else if (obs_lat_ < -M_PI_2) {
                fatal(PSTR("Can't convert RA-Dec to Azimuth without observatory information"));
            } else {
                double colatitude = M_PI_2 - obs_lat_;
                Quaternion curr = Quaternion::vector(lon_, lat_);
                Quaternion rot_lon(Quaternion(Quaternion::K).rotation(-obs_lmst_));
                Quaternion rot_lat(Quaternion(Quaternion::J).rotation(-colatitude));
                rval = curr.rotate(rot_lat * rot_lon).coords(into_system, TopoPolar);
            }
            break;

          default:
            fatal(PSTR("Can't convert RA-Dec to that system"));
            break;
        }
    } else {
        fatal(PSTR("Can't convert coordinates %d -> %d"), system_, into_system);
    }

    return rval;
}

#if PROVARINO
std::ostream& operator<<(std::ostream& strm, const Position& pos)
{
    const char* sysstring = "dummy";
    switch (pos.system()) {
      case Position::TopoPolar:
        sysstring = "TopoPolar";
        break;
      case Position::AzAlt:
        sysstring = "AzAlt";
        break;
      case Position::HADec:
        sysstring = "HADec";
        break;
      case Position::EqPolar:
        sysstring = "EqPolar";
        break;
      case Position::RADec:
        sysstring = "RADec";
        break;
      case Position::Geog:
        sysstring = "Geog";
        break;
      default:
        fatal(PSTR("Impossible pos.system: %d"), pos.system());
        break;
    }
    return strm << "[Position "
                << Angles::r2d(pos.lon()) << ',' << Angles::r2d(pos.lat())
                << ';' << sysstring << ']';
    // return strm << "[Position "
    //             << Angles::r2dms(pos.lon()) << ',' << Angles::r2dms(pos.lat())
    //             << ';' << sysstring << ']';
}
#endif
