#ifndef POSITION_H_LOADED
#define POSITION_H_LOADED 1

#if PROVARINO
#include <ostream>
#endif

#include "astrotime.h"

/**
 * The Position class allows us to represent positions on the
 * celestial sphere (or on Earth, at a push).  We're able to support a
 * subset of coordinate conversions.
 */
class Position {
 public:
    /**
     * The known coordinate systems
     *
     * enum         | explanation
     * -------------|------------
     * TopoPolar    | horizontal coordinate system: longitude is azimuthal coordinate of spherical polar coordinates (east of south, in range [0,2pi]); latitude is north of horizontal
     * AzAlt        | horizontal coordinate system: azimuth (azimuth is east of north, in range [0,2pi]) and altitude (north of horizontal)
     * HADec        | equatorial: LHA (west-positive) and Declination
     * EqPolar      | equatorial: same as HADec, but longitude is negative HA, thus the azimuthal coordinate of spherical polars, but in a right-handed equatorial frame with zero longitude on the meridian
     * RADec        | equatorial: Right Ascension and Declination
     * Geog         | geographic: latitude and east longitude
     * Velocity     | the coordinates are velocities in a right-handed system, in rad/sec
     *
     * (the `_Natural` and `_Null` systems are placemarkers, and should not be used
     * externally to this class implementation).
     */
    enum System {
        _Natural,
        TopoPolar,
        AzAlt,
        EqPolar,
        HADec,
        RADec,
        Geog,
        Velocity,
        _Null,
    };

 private:
    double lon_;
    double lat_;
    System system_;
    double obs_lat_;
    double obs_lmst_;

 public:

    Position();
    Position(double lon, double lat,
             System system,
             System into_system=_Natural);

    /**
     * The azimuthal coordinate.
     * This is east-positive, from the reference position, which has
     * zero degrees at south.
     */
    double lon(void) const { return lon_; }

    /** The latitudinal coordinate */
    double lat(void) const { return lat_; }

    /**
     * The coordinate system in which this position is expressed
     */
    System system(void) const { return system_; }

    /** Return a 'null' Position, which can be used as a sentinel value */
    static Position get_null(void) { return Position(0, 0, _Null); }
    /** Test if the Position is a 'null' one */
    bool is_null_p(void) const { return system_ == _Null; }

    void set_observatory(const Position&);
    void set_reftime(const AstroTime&);
    Position to_system(System into_system) const;
};


#if PROVARINO
std::ostream& operator<<(std::ostream& strm, const Position& pos);
#endif

#endif /* POSITION_H_LOADED */
