#ifndef QP_H_LOADED
#define QP_H_LOADED 1

/**
 * A class for controlling the telescope.  This exists to export a few
 * functions from the main qp.ino module.
 *
 * This is a pure virtual class, so that it can be subclassed differently
 * in actual and in test code.
 */
class TelescopeController
{
public:
    /**
     * Command the controller to slew directly upwards, at a fraction `v`
     * of the maximum/reference elevation speed.
     *
     * @param v the velocity relative to the maximum possible; set
     * negative to go down
     * @param arm_timeout if true (default false) the endstop-timeout function will
     * be called if the motion reaches that endstop
     */
    virtual void up(float v, bool arm_timeout=false) = 0;

    /**
     * Command the controller to slew directly eastwards, at a fraction `v`
     * of the maximum/reference azimuthal speed.
     *
     * @param v the velocity relative to the maximum possible; set
     * negative to go westwards
     * @param arm_timeout if true (default false) the endstop-timeout function will
     * be called if the motion reaches that endstop
     */
    virtual void east(float v, bool arm_timeout=false) = 0;

    /** Stop the telescope's motion. */
    virtual void stop(void) = 0;

    /**
     * The current azimuthal speed, as a fractional speed, in the range
     * [-1,1], with positive indicating that the telescope is moving east
     * (that is, clockwise around the pole, in the direction of _increasing_
     * phi, but _decreasing_ azimuth).
     */
    virtual double curr_az_speed(void) const = 0;

    /**
     * The current latitudinal speed, as a fractional speed, in the range
     * [-1,1], with positive indicating that the telescope is moving upwards.
     */
    virtual double curr_alt_speed(void) const = 0;

    /**
     * The current azimuthal angular position of the telescope.  This
     * is the phi of spherical polars, as radians east of south.
     * @return angular position, radians
     */
    virtual double phi(void) const = 0;
    /**
     * The current angular altitude of the telescope.  This
     * is the theta of (one convention for) spherical polars, as
     * radians zenithwards of the equator.
     * @return angular position, radians
     */
    virtual double theta(void) const = 0;

    /**
     * Return the current pointing of the state, as a vector in AzAlt coordinates.
     */
    Quaternion T(void) const;

    /**
     * Return the current position of the state, as a {@link Position} object.
     */
    Position position(Position::System);

    /**
     * Note that the telescope is now physically at the home
     * position.  The notion of angular position is reset as such.
     */
    virtual void reset_at_home(void) = 0;

    /**
     * Longitude of the eastmost endstop.
     * The 'eastmost endstop' is the one we hit after travelling east.
     * @return radians east of south, in [-pi,pi]
     */
    virtual float eastmost_phi(void) const = 0;

    /**
     * Longitude of the westmost endstop.
     * The 'westmost endstop' is the one we hit after travelling west.
     * @return radians east of south, in [-pi,pi]
     */
    virtual float westmost_phi(void) const = 0;

    /**
     * Altitude of the telescope's greatest depression
     * @return radians north of horizontal
     */
    virtual float lowest_altitude(void) const = 0;

    /**
     * How close may the telescope come to its endstop.  The distance here
     * is a distance in the coordinate perpendicular to the endstop; thus
     * a difference in longitude for the azimuthal endstops, and in
     * latitude for the depression endstop.
     * @return radians
     */
    virtual float endstop_clearance(void) const = 0;
};

TelescopeController* get_controller(void);

#if ! PROVARINO
// this function is provided by the qp.ino code, only for that case
void fault_exception(bool);
#endif


#endif /* QP_H_LOADED */
