// SRT control program
//
// Based on an earlier version by Graham Woan and Conner Gettings, 2014.

// #ifndef __STDC_LIMIT_MACROS
// #define __STDC_LIMIT_MACROS 1
// #endif
#include <stdint.h>             // for UINT32_MAX

#include <ctype.h>
#include <setjmp.h>

#include "types.h"
#include "activities.h"
#include "memory_free.h"

#include "util.h"

#include "astrotime.h"
#include "angles.h"
#include "qp.h"
#include "config-qp.h"
#include "exception.h"
#include "kalman.h"
#include "buffered-serial.h"

using Angles::r2d;

// pins

// fault pins: these indicate faults.
// https://www.pololu.com/product/757
//
// Flag State   | Fault Description | Disable Outputs  | Latched Until Reset
// FF1	FF2     |                   |                  |
// L    L       | No fault          | No               | No
// L    H       | Short Circuit	    | Yes              | Yes
// H    L       | Over Temperature  | No               | No
// H    H       | Under Voltage	    | Yes              | No
#define FF2_AZ 4                 // purple
#define FF1_AZ 5                 // blue
#define FF2_ALT 6                // purple
#define FF1_ALT 7                // blue

// bring RESET low to clear latched faults
#define RESET_AZ 22              // black
#define RESET_ALT 24             // white

// motor pins
#define PWM_AZ 8                 // green
#define PWM_ALT 10               // green

// ...and direction
#define DIR_AZ 9                 // yellow
#define DIR_ALT 11               // yellow

// ...and current sensors
// (I'm not positive which way round these are)
#define PWM_ALT_CURRENT A0
#define PWM_AZ_CURRENT A1

// Interrupt pins
#define IMPULSE_AZ 12     // orange
#define IMPULSE_ALT 13    // blue

// The endstop_timeout time is the time, in units of millis(), when we
// should regard the drive as having reached its end-stop.
// If we haven't received any position clicks by the time we reach
// time endstop_timeout, then we assume we've stopped because we're at
// the end stop.  Each click handler resets this timeout to a time
// CONFIG_ENDSTOP_TIMEOUT into the future.
ms_t endstop_timeout;

// Status cadence, in ms.  Write a status remark to the Serial line at this cadence.
#define STATUS_CADENCE_INIT 10000

// Our interface to the serial line
BufferedSerial bs;
#define logprintf bs.logprintf

// The time when we are next to poke an ongoing activity
ms_t activity_poke_time;

// The pointing state of the telescope
State telescope = State();

// Angles: all angles are in radians, unless otherwise noted.
//
// We can consume angles in RA/Dec where appropriate, but
// where we consume angles in a horizontal coordinate system, they are
// always in the AltAz system, where the azimuthal coordinate is in
// radians east of north, so increasing as we rotate anticlockwise
// about the pole.  We report angles in AltAz coordinates.
//
// Internally, however, we use angles in the 'opposite' system, where
// angles increase clockwise around the pole.  This is the same as
// polar coordinates, and we refer to these as 'polar angles' below.
// This is because quaternions can be more naturally described (or at
// least described with fewer confusing minus signs) in such angles.
// The zeros of the coordinate will vary; they will typically be
// labeled phi and theta.

// az_count is the azimuthal position in units of clicks, increasing
// eastwards; alt_count is the elevation in units of clicks,
// increasing towards the zenith.
//
// These are 'polar' coordinates, with the zeros at the home position.
// Note that this means that zero altitude will not be at the horizon, and
// (since the home position is at the limit of eastward travel)
// that all of the positions that the telescope can take will have
// negative click coordinates.
int az_count;
volatile ms_t az_last_click;
ms_t ignore_az_clicks_before;

// The elevation in units of clicks.
// Positive is zenith-wards from the home position.
// See above.
int alt_count;
volatile ms_t alt_last_click;
ms_t ignore_alt_clicks_before;

// heading_home_p is true if we're driving to the home position, and
// should reset az_count and alt_count when we get there
bool heading_home_p = false;

static bool queue_next_activity = false;
static ActivityQueue activity_queue;

/**
 * This struct holds a mixture of environmental and calibration information.
 */
struct EnvironmentInformation {

    /** The `prog_ref_time_ms` and `prog_ref_time_ut` objects
        represent the same time, and tie the two timescales together. */
    ms_t prog_ref_time_ms;

    /** @see prog_ref_time_ms */
    AstroTime prog_ref_time_ut;

    /** The actual latitude of the observatory, in radians north */
    double obs_lat;
    /** The actual longitude of the observatory, in radians EAST of Greenwich */
    double obs_lon;

    /**
     * The maximum azimuthal speed in rad/sec, corresponding to giving
     * the argument `1.0` to the driving function.  This is useful
     * _only_ for starting the State object from scratch, and the
     * values here will be quickly improved by the Kalman filter
     * within that State object.
     */
    double max_drive_az;
    /**
     * The maximum elevation speed.
     * See {@link max_drive_az}.
     */
    double max_drive_alt;

    /**
     * In contrast to `obs_lat` and `obs_lon`, `tel_{lat,lon}_eff`
     * represent the effective lat/long of the telescope, including
     * corrections for misalignments
     */
    double tel_lat_eff;
    /** @see tel_lat_eff */
    double tel_lon_eff;         // same units and origin as obs_lon

    /**
     * The longitude of the telescope when it is in the home position,
     * in 'polar' radians (ie, east of south, in [-pi,pi]).
     * This must be the eastmost position that the telescope can move to.
     */
    double home_position_phi;
    /**
     * The altitude of the telescope when it is in the home position,
     * in radians north of horizontal.
     */
    double home_position_theta;

    /**
     * The longitude (as 'polar' radians, east of south) of the
     * telescope when it is in the westmost 'antihome' position.  This
     * is in the range [-PI,PI].
     */
    double antihome_position_phi;
} env;

/**
 * The concrete telescope controller.
 */
class QPTelescopeController : public TelescopeController
{
private:
    // The power set on the PWM_AZ pin, as an integer from -255 to +255,
    // with positive making the drive move EAST
    int az_drive_commanded_;

    // The power set on the PWM_ALT pin, as an integer from -255 to +255,
    // with positive being north
    int alt_drive_commanded_;

    /**
     * The Kalman filter for the azimuthal coordinate.  This angle is
     * the phi of spherical polars, so zero to the south, increasing eastwards.
     *
     * This is the actual current pointing of the telescope.
     * This is always in the telescope's horizontal coordinate system,
     * with the i basis quaternion pointing south and the j quaternion pointing east.
     *
     * That is, the 'natural' coordinate system of the state is
     * {@link Position::TopoPolar}.
     */
    Kalman phi_;
    /**
     * The Kalman filter for the elevation coordinate.  This angle is
     * the theta of spherical polars (by one convention), so zero on the equator,
     * increasing towards the zenith.
     *
     * cf `phi`.
     */
    Kalman theta_;

public:
    QPTelescopeController();

    void up(float v, bool arm_timeout=false) OVERRIDE;
    void east(float v, bool arm_timeout=false) OVERRIDE;
    void stop(void) OVERRIDE;
    double curr_az_speed(void) const OVERRIDE;
    double curr_alt_speed(void) const OVERRIDE;

    double phi(void) const { return phi_.angle() };
    double theta(void) const { return theta_.angle() };
    void reset_at_home(void);

    /** The current commanded azimuthal speed, as an integer.
        This is the speed sent to the motor, in the range [-255, +255]. */
    int curr_az_speed_int(void) const { return az_drive_commanded_; };
    /** The current commanded altitude speed, as an integer.
        This is the speed sent to the motor, in the range [-255, +255]. */
    int curr_alt_speed_int(void) const { return alt_drive_commanded_; };

    float eastmost_phi(void) const OVERRIDE;
    float westmost_phi(void) const OVERRIDE;
    float lowest_altitude(void) const OVERRIDE;
    float endstop_clearance(void) const OVERRIDE;
};

QPTelescopeController controller;
TelescopeController* get_controller(void)
{
    return &controller;
}

QPTelescopeController::QPTelescopeController(void)
    : phi_(eastmost_phi(),
           env.max_drive_az,
           CONFIG_DRIVE_SLOP,
           millis()),
      theta_(lowest_altitude(),
             env.max_drive_alt,
             CONFIG_DRIVE_SLOP,
             millis())
{
    // phi_.verbose('A');
    // theta_.verbose('E');
}

void QPTelescopeController::up(float v, bool arm_timeout)
{
    // Move the drive upwards at speed v, which is in the range [-1, +1].
    // Positive speeds indicate upward, and negative downwards, and are as
    // a fraction of maximum speed in that direction.
    //
    // The motor drive truth table is as follows:
    //
    // PWM	DIR	OUTA	OUTB	Operation
    // H	L	L	H	Forward
    // H	H	H	L	Backward
    // L	X	L	L	Brake
    //
    // However note that because of the way the Arduino is connected
    // to the drive, through an inverting level shifter, setting the
    // PWM pin _low_ is results in maximum drive velocity, and PWM
    // _high_ is minimum.
    //
    // For similar reasons, it turns out that setting DIR_ALT HIGH
    // drives towards the zenith.
    byte dir;
    bool drive_positive_p;

    ms_t curr_time = millis();

    theta_.update(curr_time, curr_alt_speed());

    if (arm_timeout) {
        endstop_timeout = curr_time + CONFIG_ENDSTOP_TIMEOUT;
    }

    if (v > 0.0) {
        drive_positive_p = true;
        dir = LOW;
    } else {
        drive_positive_p = false;
        dir = HIGH;
        v = -v;
    }
    int speed = round(v * 255);
    if (speed > 255) {
        speed = 255;
    }
    alt_drive_commanded_ = (drive_positive_p ? +1 : -1) * speed;

    digitalWrite(DIR_ALT, dir);
    analogWrite(PWM_ALT, 255 - speed);
    //logprintf(PSTR("#up@%g"), (drive_positive_p ? v : -v));
}

void QPTelescopeController::east(float v, bool arm_timeout)
{
    // Move the drive eastwards at speed v.  See controller.up()
    //
    // DIR_AZ set to HIGH drives towards the east
    byte dir;
    bool drive_positive_p;

    ms_t curr_time = millis();

    phi_.update(curr_time, curr_az_speed());

    if (arm_timeout) {
        endstop_timeout = curr_time + CONFIG_ENDSTOP_TIMEOUT;
    }

    if (v > 0.0) {
        drive_positive_p = true;
        dir = HIGH;
    } else {
        drive_positive_p = false;
        dir = LOW;
        v = -v;
    }
    int speed = round(v * 255);
    if (speed > 255) {
        speed = 255;
    }
    az_drive_commanded_ = (drive_positive_p ? +1 : -1) * speed;

    digitalWrite(DIR_AZ, dir);
    analogWrite(PWM_AZ, 255 - speed);
    //logprintf(PSTR("#east@%g"), (drive_positive_p ? v : -v));
}

void QPTelescopeController::stop(void)
{
    endstop_timeout = TIME_NEVER;

    up(0);
    east(0);
}

void QPTelescopeController::reset_at_home(void)
{
    phi_.reset_angle(eastmost_phi());
    theta_.reset_angle(lowest_altitude());
}

double QPTelescopeController::curr_az_speed(void) const
{
    return (az_drive_commanded_ / 255.0);
}

double QPTelescopeController::curr_alt_speed(void) const
{
    return (alt_drive_commanded_ / 255.0);
}

float QPTelescopeController::eastmost_phi(void) const
{
    return env.home_position_phi;
}

float QPTelescopeController::westmost_phi(void) const
{
    return env.antihome_position_phi;
}

float QPTelescopeController::lowest_altitude(void) const
{
    return env.home_position_theta;
}

float QPTelescopeController::endstop_clearance(void) const
{
    return CONFIG_ENDSTOP_CLEARANCE;
}

// It's OK to call millis() inside an ISR, since it only reads a time register.
void click_az_isr(void)
{
    ms_t t = millis();
    if (t > ignore_az_clicks_before) {
        az_last_click = t;
        if (controller.curr_az_speed_int() > 0)
            az_count++;
        else
            az_count--;
        // debounce by ignoring clicks for the next 10ms
        ignore_az_clicks_before = t + 10;
    }
}

void click_alt_isr(void)
{
    ms_t t = millis();
    if (t > ignore_alt_clicks_before) {
        alt_last_click = t;
        if (controller.curr_alt_speed_int() > 0)
            alt_count++;
        else
            alt_count--;
        ignore_alt_clicks_before = t + 10; // see click_az_isr
    }
}

static void move_to_home()
{
    // the home position is fully east, and fully depressed
    logprintf(PSTR("#moving to home..."));
    az_last_click = 0;
    alt_last_click = 0;
    endstop_timeout = millis() + CONFIG_ENDSTOP_TIMEOUT;

    heading_home_p = true;

    // Move down and east at full speed on both motors
    controller.up(-1);
    controller.east(+1);
}

static void reset_at_home()
{
    az_count = 0;
    alt_count = 0;
    Quaternion home_position = Quaternion(Position(env.home_position_phi,
                                                   env.home_position_theta,
                                                   Position::TopoPolar));
    Position drive_default_velocity = Position(env.max_drive_az,
                                               env.max_drive_alt,
                                               Position::Velocity);

    activity_queue.cancel();
    Activity::delete_all_allocations();

    telescope = State(home_position,
                      drive_default_velocity,
                      NULL,      // no schedule
                      millis()); // current time

    const char* p = Quaternion::coords_str(telescope.T(), Position::AzAlt);
    logprintf(PSTR("#home: pos=(%g, %g)polar (%s)"),
              telescope.kalman_phi()->angle(),
              telescope.kalman_theta()->angle(),
              p);
    heading_home_p = false;
}

// The bits in fault_flags are as follows
//
// bit 7 : fault detected; set by ISR
//     6 : fault reported; system stopped
//     5 : recoverable error
//     4 : internal assertion violation or exception
//     3 : FF1_ALT
//     2 : FF2_ALT
//     1 : FF1_AZ
//     0 : FF2_AZ
// all bits set: set up the fault system
#define FAULT_DETECTED  1<<7
#define FAULT_STOPPED  1<<6
#define FAULT_RECOVERABLE 1<<5
#define FAULT_EXCEPTION 1<<4
// the FAULT_INITIALISE value is tested for equality; it should not
// have the FAULT_STOPPED bit set
#define FAULT_INITIALISE (0xff ^ FAULT_STOPPED)
// fault_flags == 0 is the normal no-fault case
volatile byte fault_flags = FAULT_INITIALISE;
jmp_buf fault_jmp_buf;
// The flags in FAULT_GENUINE are the actual faults.  If the
// fault_flags variable does not have one of these set after
// fault_read(), then there is no fault.  These should include
// FAULT_RECOVERABLE.
#define FAULT_GENUINE 0x3f
void fault_isr(void)
{
    fault_flags |= FAULT_DETECTED;
    // turn off both motors
    // (recall that setting these HIGH turns them off -- see the
    // discussion for controller.up)
    // analogWrite(PWM_ALT, 255);
    // analogWrite(PWM_AZ,  255);
}

// Called by the 'exception' code, to get back here in a hurry, and
// declare a fault.
void fault_exception(bool fatal)
{
    if (fatal) {
        fault_flags |= FAULT_EXCEPTION;
        fault_isr();
    } else {
        fault_flags |= FAULT_RECOVERABLE;
    }
    // note that we are relying on the fault_jmp_buf being set up here
    // -- that should be guaranteed since we can't get here before the
    // first time round the loop(), which is when that's initialised
    longjmp(fault_jmp_buf, 1);
}

// Read the four fault flags from the controller, and return them ORed
// together.  If this function returns 0, there is no fault.
byte fault_read(void)
{
    byte b = 0;

    if (digitalRead(FF1_ALT) == HIGH)
        b |= 1<<3;
    if (digitalRead(FF2_ALT) == HIGH)
        b |= 1<<2;
    if (digitalRead(FF1_AZ) == HIGH)
        b |= 1<<1;
    if (digitalRead(FF2_AZ) == HIGH)
        b |= 1<<0;

    return b;
}

// status display
static ms_t status_cadence = STATUS_CADENCE_INIT;
static ms_t status_time = status_cadence;
// These are documented in handle_commands below (search for "'s'")
#define STATUS_TIME					1UL<<0
#define STATUS_AZIMUTH_CLICKS				1UL<<1
#define STATUS_ALTITUDE_CLICKS				1UL<<2
#define STATUS_AZIMUTH_RAD				1UL<<3
#define STATUS_ALTITUDE_RAD				1UL<<4
#define STATUS_AZIMUTH_ERROR				1UL<<5
#define STATUS_ALTITUDE_ERROR				1UL<<6
#define STATUS_COMMANDED_AZ_SPEED			1UL<<7
#define STATUS_COMMANDED_ALT_SPEED			1UL<<8
#define STATUS_CURRENT_AZ				1UL<<9
#define STATUS_CURRENT_ALT				1UL<<10
#define STATUS_WORKLOAD_TOTAL_CYCLES			1UL<<11
#define STATUS_WORKLOAD_WORK_CYCLE			1UL<<12
#define STATUS_WORKLOAD_WORK_FRACTION			1UL<<13
#define STATUS_WORKLOAD_SUCCESSIVE_ACTIVE_CYCLES	1UL<<14
#define STATUS_WORKLOAD_DISCARDED_OUTPUT		1UL<<15
#define STATUS_ACTIVITY_QUEUE_SIZE			1UL<<16
#define STATUS_CURRENT_ACTIVITY				1UL<<17
#define STATUS_ALL					0xffff
unsigned long status_fields = STATUS_TIME | STATUS_AZIMUTH_RAD | STATUS_ALTITUDE_RAD | STATUS_WORKLOAD_SUCCESSIVE_ACTIVE_CYCLES | STATUS_WORKLOAD_DISCARDED_OUTPUT;

void handle_command(InputItem* cmd, int cmdlen)
{
    // The following lines with '////' are magic comments that the
    // makefile extracts to copy in to the documentation.
    // So don't 'fix' them.

    switch (cmd[0].v.c) {
      case 'c':
        //// - `c`
        ////
        ////    Starts a calibration run.  The result is stored
        ////    internally.
        ////
        ////    _Reply_: The result is written back to the serial line
        ////    in the form `>c nnn nnn`.
        ////
        //// - `c float float ...`
        ////
        ////    Allows the caller to specify pre-known calibration
        ////    values.  The line that should be supplied here is the
        ////    line that was written back after a previous `c` call.
        ////
        ////    _Reply_: `>c nnn nnn`, echoing the arguments.
        ////
        if (cmdlen == 1) {
            // Calibration.  Move around, logging click times.
            Calibration* cal = new Calibration(&controller, &telescope);
            if (queue_next_activity) {
                activity_queue.append(cal);
                queue_next_activity = false;
            } else {
                activity_queue.replace(cal);
                activity_poke_time = cal->start(millis());
            }

        } else if (cmdlen == 3) {
            // this should be [ 'c', max_drive_az, max_drive_alt ]
            env.max_drive_az = cmd[1].v.f;
            env.max_drive_alt = cmd[2].v.f;
            logprintf(PSTR(">c %G %G"), env.max_drive_az, env.max_drive_alt);

        } else {
            logprintf(PSTR("!Bad command: c [<float> <float>]"));
        }

        break;

      case 'g':
        //// - `g ...`
        ////
        ////   Go to positions indicated by coordinates, or one of a
        ////   number of 'well-known' places, as indicated below.
        ////
        ////   If the telescope believes itself to be at one of the endstops,
        ////   then it will print `>g E ` _az alt_, where the
        ////   _az_ and _alt_ are the coordinates of the current
        ////   position in _clicks_ (these coordinates may not be
        ////   terribly useful).  Detecting the endstop is unreliable
        ////   -- it depends on not detecting any click activity for a
        ////   few seconds, when it expects some -- so this signal
        ////   should be treated with some scepticism.  If it _has_
        ////   reached the endstop, then the telescope's notion of its
        ////   position has probably become unreliable.
        ////
        {
            if (cmdlen < 2) {
                logprintf(PSTR("!Malformed g command: no arguments"));

            } else if (cmd[1].type == ITEM_CH) {
                switch (cmd[1].v.c) {
                  case 'h':
                    //// - `gh az alt`
                    ////
                    ////    Drive to the given horizontal coordinates,
                    ////    and stop.  Both arguments are floats, in
                    ////    radians, giving azimuth (east from north)
                    ////    and altitude (north from horizontal).
                    ////
                    ////    _Reply_: When the telescope arrives at its
                    ////    destination, it prints the response
                    ////    `>g A float float` to the serial port,
                    ////    reporting the final azimuth and altitude in radians.
                    ////
                    if (! (cmdlen == 4
                           && cmd[2].type == ITEM_FLOAT
                           && cmd[3].type == ITEM_FLOAT)) {
                        logprintf(PSTR("!Bad command: gh <float> <float>"));
                    } else {
                        double az = cmd[2].v.f;
                        double alt = cmd[3].v.f;

                        const char* tp = Quaternion::coords_str(telescope.T(),
                                                                Position::AzAlt,
                                                                false);
                        logprintf(PSTR("#gh: telescope at %s; drive to (az, alt)=(%g,%g)"),
                                  tp, az, alt);

                        DriveAzAlt* drive = new DriveAzAlt(&controller,
                                                           &telescope,
                                                           az, alt);
                        if (drive->is_valid()) {
                            // all OK
                            if (queue_next_activity) {
                                activity_queue.append(drive);
                                queue_next_activity = false;
                            } else {
                                activity_queue.replace(drive);
                                activity_poke_time = drive->start(millis());
                            }
                        } else {
                            delete drive;
                            // do nothing -- this has already been reported by
                            // the DriveAzAlt constructor
                        }
                    }
                    break;

                  case 'e':
                    //// - `ge ra dec`
                    ////
                    ////   Drive to the given equatorial coordinates
                    ////   (RA and Dec), and stop.
                    ////   Both coordinates are floats, in radians.
                    ////   Neither this control system, nor the telescope, are
                    ////   accurate enough that we have to worry about
                    ////   the epoch in which the coordinates are given.
                    ////
                    ////   _Reply_: When the telescope arrives at its
                    ////   destination, it prints the response
                    ////   `>g A float float` to the serial port,
                    ////   reporting the final azimuth and altitude in radians.
                    ////
                    if (! (cmdlen == 4
                           && cmd[2].type == ITEM_FLOAT
                           && cmd[3].type == ITEM_FLOAT)) {
                        logprintf(PSTR("!bad command: ge <float> <float>"));
                    } else {
                        double ra = cmd[2].v.f;
                        double dec = cmd[3].v.f;

                        Position pointing = Position(ra, dec, Position::RADec);
                        Position pointing_h = pointing.to_system(Position::AzAlt);

                        const char* tp = Quaternion::coords_str(telescope.T(), Position::AzAlt);
                        logprintf(PSTR("#ge: telescope at %s; drive to RA %g dec %gdeg, which is (az,alt)=(%g, %g) deg"),
                                  tp,
                                  r2d(ra), r2d(dec),
                                  r2d(pointing_h.lon()),
                                  r2d(pointing_h.lat()));

                        DriveAzAlt* drive = new DriveAzAlt(&controller,
                                                           &telescope,
                                                           pointing_h.lon(),
                                                           pointing_h.lat());
                        if (drive->is_valid()) {
                            if (queue_next_activity) {
                                activity_queue.append(drive);
                                queue_next_activity = false;
                            } else {
                                activity_queue.replace(drive);
                                activity_poke_time = drive->start(millis());
                            }
                        } else {
                            delete drive;
                        }
                    }
                    break;

                  case 'H':
                    //// - `gH`
                    ////
                    ////    Go to the 'home' position.  This position
                    ////    is defined as the place where the
                    ////    telescope ends up if it is driven all the
                    ////    way to its easternmost end-stop, and all
                    ////    the way down.  The actual azimuth and
                    ////    altitude of this point is calibrated
                    ////    using the `O` command.
                    ////
                    ////    This is effectively a ‘reset’ command for
                    ////    the system's notion of position.
                    ////
                    ////    After the telescope arrives at its home
                    ////    position, it updates its internal notion
                    ////    of the telescope's position with the
                    ////    calibrated coordinates of the home
                    ////    position as provided by the `O` command.
                    ////
                    ////    The telescope's reported position should
                    ////    not be regarded as reliable, between issuing this
                    ////    command and the telescope arriving at the
                    ////    home position and resetting.
                    ////
                    ////    _Reply_: `>g E nnn nnn` reporting that the
                    ////    telescope has reached its endstops, and...
                    ////    FIXME: this would more sensibly report
                    ////    angles, if those are available.
                    ////
                    move_to_home();
                    break;

                  case 'U':
                    //// - `gU`
                    ////
                    ////    Drive the telescope upwards, until the
                    ////    system reaches one of the end-stops, until
                    ////    it receives one of the other similar
                    ////    movement commands, or it receives the stop
                    ////    command `x`.
                    ////
                    ////    **This command, and the corresponding
                    ////    down/east/west commands, should generally
                    ////    _not_ be used for moving the
                    ////    telescope, since they are intended to be
                    ////    low-level commands.**  Instead, use the
                    ////    `nu/d/e/w` commands to nudge the telescope
                    ////    by small amounts.
                    ////
                    ////    Nonetheless, the telescope should retain
                    ////    an accurate notion of its position.  If
                    ////    for some reason it does not, it can be
                    ////    reset by driving it home with `gH`.
                    ////
                    ////    If you use the `gU` command to drive the
                    ////    telescope over the zenith, it will produce
                    ////    an error, and the control software will
                    ////    probably have to be restarted.
                    ////
                    ////    These commands may be removed in future
                    ////    revisions of the software.
                    ////
                    ////    _Reply_: None.
                    ////
                    telescope.step(millis()); // update pointing before change of speed
                    controller.up(+1, true);
                    break;
                  case 'D':
                    //// - `gD`
                    ////
                    ////    Drive the telescope downwards.
                    ////
                    telescope.step(millis());
                    controller.up(-1, true);
                    break;
                  case 'E':
                    //// - `gE`
                    ////
                    ////    Drive the telescope eastwards.
                    ////
                    telescope.step(millis());
                    controller.east(+1, true);
                    break;
                  case 'W':
                    //// - `gW`
                    ////
                    ////    Drive the telescope westwards.
                    ////
                    telescope.step(millis());
                    controller.east(-1, true);
                    break;

                  default:
                    logprintf(PSTR("!Unrecognised position character: %c"), cmd[1].v.c);
                    break;
                }

            } else {
                logprintf(PSTR("!Malformed g command: first argument must be a character"));
            }
        }
        break;

      case 'n':
        {
            if (cmdlen < 2) {
                logprintf(PSTR("!malformed n command: no arguments"));

            } else if (cmdlen > 3) {
                logprintf(PSTR("!malformed n command: too many arguments"));

            } else if (cmdlen == 3 && cmd[2].type != ITEM_FLOAT) {
                logprintf(PSTR("!malformed n command: n<char> <float>"));

            } else if (cmd[1].type == ITEM_CH) {
                Position current_position = telescope.position(Position::AzAlt);
                // double alt = telescope.kalman_theta()->angle();
                // double az = telescope.kalman_phi()->angle();
                double az = current_position.lon();
                double alt = current_position.lat();
                double delta = 0.01; // rad

                if (cmdlen == 3) {
                    delta = double(cmd[2].v.f);
                    if (delta <= 0.0) {
                        logprintf(PSTR("!malformed n command: argument must be positive"));
                        delta = 0.0; // inhibits movement
                    }
                }

                switch (cmd[1].v.c) {
                  case 'u':
                    //// - `nu [float]`
                    ////
                    ////    Nudge the telescope up by a small amount,
                    ////    ‘delta’, given in radians.  This command
                    ////    is equivalent to a `gh` command moving the
                    ////    telescope to a position ‘delta’ higher in
                    ////    altitude.  If ‘delta’ is not provided, the
                    ////    default is 0.01 rad (about 0.5 deg; yes,
                    ////    giving an argument in degrees would be
                    ////    more convenient, but it seems best to be
                    ////    consistent in the argument type).
                    ////
                    ////    _Reply_: `g A float float` reports that
                    ////    the telescope has arrived, and notes its new current coordinates.
                    ////
                    alt += delta;
                    break;

                  case 'd':
                    //// - `nd [float]`
                    ////
                    ////    As with `nu`, but nudging the telescope down.
                    ////
                    alt -= delta;
                    break;

                  case 'e':
                    //// - `ne [float]`
                    ////
                    ////    As with `nu`, but nudging the telescope east.
                    ////
                    az -= delta;
                    break;

                  case 'w':
                    //// - `nw [float]`
                    ////
                    ////    As with `nu`, but nudging the telescope west.
                    ////
                    az += delta;
                    break;

                  default:
                    logprintf(PSTR("!malformed n command: nu/nd/ne/nw [<float>]"));
                    delta = 0.0; // inhibit actual movement
                    break;
                }

                if (delta > 0.0) {
                    DriveAzAlt* drive = new DriveAzAlt(&controller,
                                                       &telescope,
                                                       az, alt);
                    if (drive->is_valid()) {
                        if (queue_next_activity) {
                            activity_queue.append(drive);
                            queue_next_activity = false;
                        } else {
                            activity_queue.replace(drive);
                            activity_poke_time = drive->start(millis());
                        }
                    } else {
                        delete drive;
                    }
                }
            } else {
                logprintf(PSTR("!malformed n command: n<char> <float>"));
            }
        }
        break;

      case 'O':
        //// - `O lat lon dlat dlon az alt span`
        ////
        ////    Gives the location of the telescope/observatory and the pointing
        ////    of the 'home position'.  The first two
        ////    arguments are respectively observatory latitude and
        ////    longitude, and the second two are deltas to that, to
        ////    give the telescope effective latitude and longitude,
        ////    taking into account misalignments of the telescope's
        ////    axes.  The fifth and sixth are the azimuth and altitude of
        ////    the telescope's home position; this home position is
        ////    the eastmost and most depressed position.  The seventh
        ////    argument is the (positive) total angle between the two
        ////    azimuthal end-stops; thus the westmost endstop has
        ////    azimuth equal to the home azimuth (argument 5) plus
        ////    the span (argument 7), normalised to [0,2pi).
        ////
        ////    All arguments are floats, given in radians.  The
        ////    longitude is east-longitude, in [0, 2pi), and the
        ////    telescope effective position is taken to be
        ////    `(lat+dlat, lon+dlon)`.  The azimuth is in radians east of north,
        ////    and the altitude in radians zenithwards from the
        ////    horizontal.
        ////
        ////    _Reply_: none.
        ////
        //// - `O`
        ////
        ////    Queries the orientation coordinates.
        ////
        ////    _Reply_: Returns `>O float float float float float float float`,
        ////    giving the current status of these values in the same
        ////    form as the seven-argument command.
        ////
        if (cmdlen == 1) {
            float span = env.home_position_phi - env.antihome_position_phi;
            while (span < 0) {
                // ensure span is positive
                span += 2*M_PI;
            }
            logprintf(PSTR(">O %G %G %G %G %G %G %G"),
                      env.obs_lat, env.obs_lon,
                      env.tel_lat_eff - env.obs_lat, env.tel_lon_eff - env.obs_lon,
                      M_PI - env.home_position_phi, env.home_position_theta,
                      span);
        } else if (cmdlen != 8) {
            logprintf(PSTR("!Wrong number of arguments (%d) to 'O' command; should be 7"),
                      cmdlen-1);
        } else if (! (   cmd[1].type == ITEM_FLOAT
                      && cmd[2].type == ITEM_FLOAT
                      && cmd[3].type == ITEM_FLOAT
                      && cmd[4].type == ITEM_FLOAT
                      && cmd[5].type == ITEM_FLOAT
                      && cmd[6].type == ITEM_FLOAT
                      && cmd[7].type == ITEM_FLOAT)) {
            logprintf(PSTR("!Invalid O arguments -- wrong types"));
        } else {
            env.obs_lat = cmd[1].v.f;
            env.obs_lon = cmd[2].v.f;
            env.tel_lat_eff = env.obs_lat + cmd[3].v.f;
            env.tel_lat_eff = env.obs_lon + cmd[4].v.f;
            env.home_position_phi = M_PI - cmd[5].v.f;
            env.home_position_theta = cmd[6].v.f;
            env.antihome_position_phi = env.home_position_phi - cmd[7].v.f;
            while (env.antihome_position_phi < -M_PI) {
                // bring into range [-PI,PI]
                env.antihome_position_phi += 2*M_PI;
            }
            env.prog_ref_time_ut = env.prog_ref_time_ut.move(env.obs_lon);

            logprintf(PSTR("#observatory at (%g, %g) deg, telescope at (%g,%g) deg, "
                           "home at (%g,%g)AzAlt, "
                           "time MJD %d + %gUT"),
                      r2d(env.obs_lat), r2d(env.obs_lon),
                      r2d(env.tel_lat_eff), r2d(env.tel_lon_eff),
                      r2d(M_PI - env.home_position_phi), r2d(env.home_position_theta),
                      env.prog_ref_time_ut.mjd(), env.prog_ref_time_ut.ut1());
        }
        break;

      case 'q':
        //// - `q`
        ////
        ////    When an activity, such as a slew or tracking, is
        ////    commanded, it immediately replaces any current activity.
        ////    If the command `q` is given, then the next activity to
        ////    be commanded will _not_ replace the current one, but
        ////    will be queued up behind it, and replace it when that
        ////    activity is finished.  The queue has a fixed maximum
        ////    size, and if too many items are queued, the system
        ////    will return an error (though this isn't fatal).
        ////
        ////    _Reply_: none.
        ////
        queue_next_activity = true;
        break;

      case 'S':
        //// - `S`
        ////
        ////     Return status information, describing the
        ////     controller's view of the world.
        ////
        ////     This is intended to be for overall status and
        ////     calibration information.  Routine status information
        ////     should be read from the periodic status output in the
        ////     `s,...` line.
        ////
        ////     _Reply_: returns a string in the form `>S [key=val]...`.
        ////     The `Taz` position is radians east of north.
        ////
        {
            Position p = telescope.T().coords(Position::AzAlt);
            char buf[30];
            env.prog_ref_time_ut.str(buf, 30);
            logprintf(PSTR(">S reftime_ms=" MS_FMT
                           " reftime_mjd=%d reftime_ut=%g"
                           " reftime=%s"
                           " obs_lat=%g obs_lon=%g tel_lat=%g tel_lon=%g"
                           " Taz=%g Talt=%g"
                           " max_activity=%d"),
                      env.prog_ref_time_ms,
                      env.prog_ref_time_ut.mjd(), env.prog_ref_time_ut.ut1(),
                      buf,
                      env.obs_lat, env.obs_lon, env.tel_lat_eff, env.tel_lon_eff,
                      p.lon(), p.lat(),
                      Activity::max_activities());
        }
        break;

      case 's':
        //// - `s int`
        ////
        ////    Change the status cadence to the given number of
        ////    milliseconds.  If this is changed to zero, then status
        ////    messages are switched off.
        ////
        ////    _Reply_: none.
        ////
        //// - `s` _char_...
        ////
        ////    Adjust the status fields which are shown.  The fields
        ////    are selected for printing by a single character in the given
        ////    list.  The fields are always printed in the same
        ////    order shown below, thus `s za` and `s az` both display the azimuth
        ////    followed by the altitude (and nothing else).  By
        ////    default, only a selection of fields are shown, which
        ////    are equivalent to the command `s tzaCd`.
        ////
        ////    The fields are as given below.
        ////
        ////    _Reply_: none.
        ////
        if (cmdlen == 1) {
            logprintf(PSTR("!no arguments provided to 's' command"));
        } else if (cmd[1].type == ITEM_INT) {
            if (cmd[1].v.i == 0) {
                status_cadence = TIME_NEVER;
            } else {
                status_cadence = cmd[1].v.i;
            }
        } else {
            status_fields = 0;
            for (int i=1; i<cmdlen; i++) {
                if (cmd[i].type == ITEM_CH) {
                    switch (cmd[i].v.c) {
                        // NOTE: the cases below, or at least their
                        // documentation, must be in the same order as
                        // the various tests in the (reftime >
                        // status_time) branch in the main loop.
                      case 't':
                        ////    + t: time in ms (int)
                        status_fields |= STATUS_TIME;
                        break;
                      case 'Z':
                        ////    + Z: click-azimuth (relative to the reference
                        ////         position, increasing as the
                        ////         position moves westwards -- that
                        ////         is, in the same sense as
                        ////         increasing azimuth; in clicks, int)
                        status_fields |= STATUS_AZIMUTH_CLICKS;
                        break;
                      case 'A':
                        ////    + A: click-altitude (upwards of the
                        ////         reference position; in clicks, int)
                        status_fields |= STATUS_ALTITUDE_CLICKS;
                        break;
                      case 'z':
                        ////    + z: azimuth (east of north) in radians (float)
                        status_fields |= STATUS_AZIMUTH_RAD;
                        break;
                      case 'a':
                        ////    + a: altitude (north of horizon) in radians (float)
                        status_fields |= STATUS_ALTITUDE_RAD;
                        break;
                      case 'e':
                        ////    + e: estimated error in azimuthal
                        ////      position reported above, in units of radians
                        status_fields |= STATUS_AZIMUTH_ERROR;
                        break;
                      case 'E':
                        ////    + E: estimated error in altitudinal
                        ////      position reported above, in units of radians
                        status_fields |= STATUS_ALTITUDE_ERROR;
                        break;
                      case 's':
                        ////    + s: commanded az drive speed (integer in [-255,255])
                        status_fields |= STATUS_COMMANDED_AZ_SPEED;
                        break;
                      case 'S':
                        ////    + S: commanded alt drive speed (integer in [-255,255])
                        status_fields |= STATUS_COMMANDED_ALT_SPEED;
                        break;
                      case 'i':
                        ////    + i: current draw for the azimuth motor (units???, float)
                        status_fields |= STATUS_CURRENT_AZ;
                        break;
                      case 'I':
                        ////    + I: current draw for the altitude motor (units???, float)
                        status_fields |= STATUS_CURRENT_ALT;
                        break;
                      case 'c':
                        ////    + c: workload: number of cycles since last status (int)
                        status_fields |= STATUS_WORKLOAD_TOTAL_CYCLES;
                        break;
                      case 'w':
                        ////    + w: workload: fraction of work cycles
                        ////      (ie, count of non-idle runs through main loop,
                        ////      as a fraction of 'c'; float)
                        status_fields |= STATUS_WORKLOAD_WORK_CYCLE;
                        break;
                      case 'W':
                        ////    + W: workload: fraction of time working
                        ////      (ie, the total time spent in work
                        ////      cycles, as a fraction of the status cadence; float)
                        status_fields |= STATUS_WORKLOAD_WORK_FRACTION;
                        break;
                      case 'C':
                        ////    + C: workload: number of successive
                        ////      active cycles (ie, the number of times
                        ////      within the last status cadence that an
                        ////      'active' cycle was immediately
                        ////      followed by another one, indicating
                        ////      that events may be piling up; int)
                        status_fields |= STATUS_WORKLOAD_SUCCESSIVE_ACTIVE_CYCLES;
                        break;
                      case 'd':
                        ////    + d: workload: number of bytes discarded from full output buffer
                        status_fields |= STATUS_WORKLOAD_DISCARDED_OUTPUT;
                        break;
                      case 'q':
                        ////    + q: activities queued: number of
                        ////      activities in the queue, including the
                        ////      current one (int)
                        status_fields |= STATUS_ACTIVITY_QUEUE_SIZE;
                        break;
                      case 'b':
                        ////    + b: current activity/busy status.
                        ////      Can be 'n', indicating that there is no current activity,
                        ////      'y' indicating that there is an ongoing activity,
                        ////      or 'c' indicating that the current activity has been completed.
                        status_fields |= STATUS_CURRENT_ACTIVITY;
                        break;
                      default:
                        logprintf(PSTR("!unexpected status character %c"), cmd[i].v.c);
                    }
                }
            }
            logprintf(PSTR("#status_fields=%x"), status_fields);
        }
        break;

      case 'T':
        //// - `T year mon day  hour min sec`
        ////
        ////    Set the current date-time.
        ////    The first five are integers, the last
        ////    is a float.  The time is in UT.
        ////
        ////    _Reply_: none.
        ////
        if (! (cmdlen == 7
               && cmd[1].type == ITEM_INT
               && cmd[2].type == ITEM_INT
               && cmd[3].type == ITEM_INT
               && cmd[4].type == ITEM_INT
               && cmd[5].type == ITEM_INT
               && cmd[6].type == ITEM_FLOAT)) {
            logprintf(PSTR("!Wrong number (%d) or type of arguments to 'T' command; should be int,int,int,int,int,float"),
                      cmdlen-1);

        } else {
            env.prog_ref_time_ms = millis();

            int y, m, d, H, M;
            double S;
            y = cmd[1].v.i;
            m = cmd[2].v.i;
            d = cmd[3].v.i;
            H = cmd[4].v.i;
            M = cmd[5].v.i;
            S = cmd[6].v.f;

            if (y < 0 || m < 0 || d < 0 || H < 0 || M < 0 || S < 0) {
                logprintf(PSTR("!Invalid y,m,d,h,m,s -- must be positive"));
            } else {
                env.prog_ref_time_ut = AstroTime(y, m, d, H, M, S, env.obs_lon);
                logprintf(PSTR("#MJD %d + %gUT (local sidereal time %grad, %gdegE)"),
                          env.prog_ref_time_ut.mjd(),
                          env.prog_ref_time_ut.ut1(),
                          env.prog_ref_time_ut.lmst(),
                          env.obs_lon);
            }
        }
        break;

      case 't':
        if (cmdlen >= 2 && cmd[1].type == ITEM_CH) {
            switch (cmd[1].v.c) {
              case 's':
                //// - `ts [speed]`
                ////
                ////   **DO NOT USE**  This command does not currently work.
                ////
                ////   Track a sidereal rotation.  If no `speed` is
                ////   provided, we rotate about the
                ////   celestial pole at one sidereal day per solar
                ////   day.  If a `speed` is provided, it is a float,
                ////   in rad/s clockwise about the celestial pole
                ////   (sidereal rotation is 15'/min x 1.00274, or 7.2e-5 rad/s).
                ////
                if (cmdlen == 2
                    || (cmdlen == 3 && cmd[2].type == ITEM_FLOAT)) {
                    Position obs = Position(env.obs_lon, env.obs_lat, Position::Geog);
                    double speed = (cmdlen == 2 ? 0.0 : cmd[2].v.f);

                    TrackSidereal* track = new TrackSidereal(&controller,
                                                             &telescope,
                                                             obs,
                                                             speed);
                    if (queue_next_activity) {
                        activity_queue.append(track);
                        queue_next_activity = false;
                    } else {
                        activity_queue.replace(track);
                        activity_poke_time = track->start(millis());
                    }
                } else {
                    logprintf(PSTR("!Wrong number or type or arguments for 'ts'"));
                }
                break;

              case 'a':
                //// - `ta speed`
                ////
                ////   Track in azimuth at the given speed, provided
                ////   in rad/s, with positive meaning tracking in
                ////   increasing azimuth; thus westwards.
                ////
                ////   NOTE: this function is under-tested, and it
                ////   should probably not be used at present.
                ////
                if (cmdlen == 3 && cmd[2].type == ITEM_FLOAT) {
                    double azimuth_speed = cmd[2].v.f; // rad/s

                    TrackAzimuth* track = new TrackAzimuth(&controller,
                                                           &telescope,
                                                           azimuth_speed);

                    if (queue_next_activity) {
                        activity_queue.append(track);
                        queue_next_activity = false;
                    } else {
                        activity_queue.replace(track);
                        activity_poke_time = track->start(millis());
                    }
                } else {
                    logprintf(PSTR("!Wrong number or type of arguments for 'ta'"));
                }
                break;

              default:
                logprintf(PSTR("!Unexpected type argument to 't' command"));
                break;
            }
        } else {
            logprintf(PSTR("!Wrong number of arguments to 't' command (should be one character arg)"));
        }
        break;

      case 'v':
        //// - `v`
        ////
        ////    Display the version of the telescope control
        ////    software.
        ////
        ////    _Reply_: The response is of the form `>v integer
        ////    string string`.  The integer is the version encoded as
        ////    major*10000 + minor*100 + release (so version 1.2.34
        ////    would turn into 10234.  The first string indicates the
        ////    release status, and will be `RELEASE` if it's a formal
        ////    versioned release.  The second string is the
        ////    repository revision hash.
        ////
        logprintf(PSTR(">v %s"), Util::version_string());
        break;

      case 'x':
        //// - `x`
        ////
        ////    Stop the telescope moving.  This cancels the current
        ////    activity, and de-queues any queued activity.
        ////
        ////    _Reply_: none.
        ////
        {
            // If there's a current activity, let it update itself
            Activity* curr = activity_queue.current();
            if (curr != NULL) {
                curr->poke(millis()); // updates the state, as well
            }

            activity_queue.cancel();

            // stop the telescope -- this is redundant if we were
            // mid-activity, but necessary if what we're stopping is
            // g{U,D,W,E}.
            controller.stop();

            Position p = telescope.T().coords(Position::AzAlt);
            logprintf(PSTR("#stopped at (az,alt)=%g,%g"),
                      p.lon(), p.lat());

            // reset the endstop_timeout: we're not going to get any
            // further clicks in the immediate future
            endstop_timeout = TIME_NEVER;
            activity_poke_time = TIME_NEVER;
        }
        break;

      case 'X':
        //// - `X`
        ////
        ////    Stow the telescope in the pointing-upright position.
        ////    This cancels any other current and queued activities.
        ////
        ////    _Reply_: `g A float float` reports that it has
        ////    arrived, and notes the new current coordinates.
        ////
        {
            float current_az = telescope.kalman_phi()->angle();

            DriveAzAlt* drive = new DriveAzAlt(&controller,
                                               &telescope,
                                               M_PI - current_az,
                                               M_PI_2);

            if (drive->is_valid()) {
                activity_queue.replace(drive);
                activity_poke_time = drive->start(millis());
            } else {
                delete drive;
            }
        }
        break;

      default:
        logprintf(PSTR("!Unrecognised command: %c"), cmd[0].v.c);
        break;
    }
}

void setup()
{
    // input pins
    pinMode(FF1_AZ,		INPUT);
    pinMode(FF2_AZ,		INPUT);
    pinMode(FF1_ALT,		INPUT);
    pinMode(FF2_ALT,		INPUT);
    pinMode(IMPULSE_AZ,		INPUT);
    pinMode(IMPULSE_ALT,	INPUT);

    // output pins; initialise to sensible values.
    pinMode(RESET_AZ,		OUTPUT);
    digitalWrite(RESET_AZ,	HIGH);
    pinMode(RESET_ALT,		OUTPUT);
    digitalWrite(RESET_ALT,	HIGH);

    // Note: the DIR and PWM pins are inverted before they reach the
    // controller board
    pinMode(PWM_AZ,		OUTPUT);
    pinMode(PWM_ALT,		OUTPUT);
    pinMode(DIR_AZ,		OUTPUT);
    pinMode(DIR_ALT,		OUTPUT);
    digitalWrite(PWM_AZ,	HIGH);
    digitalWrite(PWM_ALT,	HIGH);
    digitalWrite(DIR_AZ,	HIGH);
    digitalWrite(DIR_ALT,	HIGH);

    pinMode(PWM_ALT_CURRENT,	INPUT);
    pinMode(PWM_AZ_CURRENT,	INPUT);

    // interrupts for clickers and faults
    // FIXME: find out the distinction between this approach and using ISR
    attachInterrupt(IMPULSE_AZ, click_az_isr, FALLING);
    attachInterrupt(IMPULSE_ALT, click_alt_isr, FALLING);

    // XXX do not ensable these faults: we get spurious
    // over-temperature and under-voltage faults -- I should fix this
    // by 'debouncing' the detection of this
    // attachInterrupt(FF1_ALT, fault_isr, RISING);
    // attachInterrupt(FF2_ALT, fault_isr, RISING);
    // attachInterrupt(FF1_AZ,  fault_isr, RISING);
    // attachInterrupt(FF2_AZ,  fault_isr, RISING);

    az_last_click = 0;
    alt_last_click = 0;

    endstop_timeout = TIME_NEVER;
    activity_poke_time = TIME_NEVER;
    ignore_az_clicks_before = 0;
    ignore_alt_clicks_before = 0;

    // Initialise approximate (at least, sane) values for the calibration
    // parameters, so that things don't break if they're not set.
    env.obs_lat = CONFIG_OBS_LAT_ACTUAL;
    env.obs_lon = CONFIG_OBS_LON_ACTUAL;
    env.tel_lat_eff = CONFIG_OBS_LAT_EFFECTIVE;
    env.tel_lon_eff = CONFIG_OBS_LON_EFFECTIVE;
    env.max_drive_az = CONFIG_MAX_DRIVE_AZ;
    env.max_drive_alt = CONFIG_MAX_DRIVE_ALT;
    env.home_position_phi = M_PI - CONFIG_HOME_POSITION_AZIMUTH;
    env.home_position_theta = CONFIG_HOME_POSITION_ALTITUDE;
    env.antihome_position_phi = M_PI - CONFIG_ANTIHOME_POSITION_AZIMUTH;

    bs.init();

    move_to_home();
}

void loop()
{
    ms_t reftime = millis();

    InputItem* cmd;
    int cmdlen;

    // gather statistics on the processor workload
    static unsigned int workload_cycles = 0;
    static unsigned int workload_idle_cycles = 0;
    static unsigned int workload_backup_cycles = 0;
    static ms_t workload_start = 0;
    static ms_t workload_working = 0;
    static bool workload_active_last_time_p = false;

    bool active_p = true;

    ////
    //// Status output
    //// -------------
    ////
    //// The CSV status outputs from the running program are as follows (note
    //// that these are still rather unstable.
    ////

#if 0
    if ((activity_queue.current() != NULL && activity_queue.current()->is_complete())
        || (activity_queue.current() != NULL && reftime >= activity_poke_time)
        || (reftime > endstop_timeout)
        || (az_last_click != 0)
        || (alt_last_click != 0)
        || Serial.available() > 1) {
        logprintf(PSTR("#reftime " MS_FMT "%|"), reftime);
        if (activity_queue.current() != NULL) {
            logprintf(PSTR(" current: complete=%c  poke-time=" MS_FMT "%|"),
                      activity_queue.current()->is_complete() ? 'y' : 'n',
                      activity_poke_time);
        }
        if (endstop_timeout != TIME_NEVER) {
            logprintf(PSTR(" timeout@" MS_FMT "%|"), endstop_timeout);
        }
        if (Serial.available() > 1) {
            logprintf(PSTR(" available%|"));
        }
        logprintf(PSTR(" az_click@" MS_FMT ", alt_click@" MS_FMT),
                  az_last_click, alt_last_click);
    }
#endif


    if (fault_flags & FAULT_STOPPED) {
        // we are in a fault state, but we've already reported the status: do nothing
        static ms_t next_report = 0;
        if (reftime > next_report) {
            logprintf(PSTR("stopped"));
            next_report = reftime + 10000;
        }

    } else if (fault_flags) {
        if (fault_flags == FAULT_INITIALISE) {
            // first time
            if (setjmp(fault_jmp_buf) == 0) {
                // set up OK
                fault_flags = 0;
            }
            // We're here either because we've just set up the
            // jmp_buf, or because there's been an error and we've
            // jumped here from fault_exception.  We're now in the
            // correct place -- inside the fault branch of the loop()
            // function -- to deal with the latter case.
        }

        // Delay briefly before reading the fault flags, in order to
        // ignore transient faults (typically undervoltage faults when
        // the motors start moving).  It is tempting to be smarter
        // about scheduling this, but we should resist that (a)
        // because if this is a genuine fault we should be very
        // cautious about doing anything at all, and (b) because we
        // don't want to be smart here, and risk missing things.
        delay(10);

        fault_flags |= fault_read();
        // logprintf(PSTR("#fault? %x (real? %c)"),
        //           fault_flags, (fault_flags & FAULT_GENUINE ? 'y' : 'n'));

        if (fault_flags & FAULT_GENUINE) {
            // the drive should have been turned off in the fault_isr
            // routine, but just in case we've got here by another route...
            controller.up(0);
            controller.east(0);

            activity_queue.cancel();
            Activity::delete_all_allocations();

            if (fault_flags == FAULT_RECOVERABLE) {
                // This is a recoverable error: stop the drives, but
                // don't stop the system.  Test this for equality
                // (that is, this must be the _only_ bit set),
                // rather than for just this bit set, because we
                // wouldn't want to go down this path if there are,
                // coincidentally, other faults present.
                logprintf(PSTR("!recoverable error: motors stopped, but system continuing"));
                fault_flags = FAULT_INITIALISE;
            } else {
                // report what's happened
                logprintf(PSTR("!!! fault %x; system stopped"), fault_flags);

                // It's not clear what to do next.
                //
                // We've turned the drives off, but not completely cut the
                // power to the system.  Note that the short circuit fault is
                // latched, but we include here no facility for resetting
                // this.

                // Don't come here again
                fault_flags |= FAULT_STOPPED;
            }
        } else {
            fault_flags = 0;
        }

    } else if (az_last_click != 0) {
        // click in azimuth

        // It's important that the checks for az_last_click and
        // alt_last_click come early: if we do a step() in the State
        // (via activity_queue.current()->poke()) and later do a click
        // with an _earlier_ time, the Kalman filters get _very_ confused.

        // Note the time of the click, for reporting below.  We reset
        // az_last_click as quickly as possible: if we waited to the
        // end of this block to do so, there's a danger that the ISR
        // setting it would fire between now and then, which would
        // result in us missing it.  These clicks are a least hundreds
        // of ms apart, so the risk is small, but it seems best to be
        // over-careful.
        ms_t clicktime = az_last_click;
        az_last_click = 0;

        double phi   = az_count*RAD_PER_CLICK_AZ + env.home_position_phi;
        telescope.az_click(clicktime, controller.curr_az_speed(), phi);
        if (activity_queue.current() != NULL) {
            activity_queue.current()->az_click(clicktime, phi);
        }

        //// - `az,int,int,float,float,int`
        ////
        ////    An azimuth click, which is produced each time the
        ////    telescope drive produces a 'click' when it passes a
        ////    reed switch.  These happen every half degree.  The columns are:
        ////    + current time in ms
        ////    + the time that the click was registered
        ////    + azimuth (east of north) in radians
        ////    + altitude (north of horizon) in radians
        ////    + azimuth (relative to home position, in the same
        ////      sense as increasing azimuth) in clicks
        ////
        ////    The azimuth and altitude should be accurate to less
        ////    than half a degree, and are obtained as offsets from
        ////    the calibrated coordinates of the home position.  The
        ////    azimuth in clicks is a running count of clicks, and so
        ////    should differ by either +1 or -1 from the previously
        ////    reported value.
        ////
        Kalman* Klon = telescope.kalman_phi();
        Kalman* Kalt = telescope.kalman_theta();
        logprintf(PSTR("az," MS_FMT "," MS_FMT ",%g,%g,%d"),
                  reftime,
                  clicktime,
                  M_PI - Klon->angle(),
                  Kalt->angle(),
                  -az_count);   // report in the same sense as real azimuth
        endstop_timeout = reftime + CONFIG_ENDSTOP_TIMEOUT;

    } else if (alt_last_click != 0) {
        // click in altitude

        ms_t clicktime = alt_last_click; // see above
        alt_last_click = 0;

        double theta = alt_count*RAD_PER_CLICK_ALT + env.home_position_theta;
        telescope.alt_click(clicktime, controller.curr_alt_speed(), theta);
        if (activity_queue.current() != NULL) {
            activity_queue.current()->alt_click(clicktime, theta);
        }

        //// - `alt,int,int,float,float,int`
        ////
        ////    An altitude click, which is produced each time the
        ////    telescope drive produces a 'click' when it passes a
        ////    reed switch.  These happen every half degree.  The
        ////    columns are as above, except that the final column is
        ////    altitude (north of home position) in clicks.
        ////
        Kalman* Klon = telescope.kalman_phi();
        Kalman* Kalt = telescope.kalman_theta();
        logprintf(PSTR("alt," MS_FMT "," MS_FMT ",%g,%g,%d"),
                  reftime,
                  clicktime,
                  M_PI - Klon->angle(),
                  Kalt->angle(),
                  alt_count);
        endstop_timeout = reftime + CONFIG_ENDSTOP_TIMEOUT;

    } else if ((cmd = bs.read_items_get(&cmdlen)) != NULL) {
        handle_command(cmd, cmdlen);
        bs.read_items_reset();

    } else if (activity_queue.current() != NULL
               && activity_queue.current()->is_complete()) {
        Activity* curr = activity_queue.current();

        // there's no updating to do,
        // since the activity decided it was complete, and stopped

        // dynamic_cast<Calibration*> is disabled on Arduino
        // (-fno-rtti), so use this technique instead
        Activity::ActivityTypes doing = curr->type();

        if (doing == Activity::CalibrationActivity) {
            Calibration* cal = static_cast<Calibration*>(curr);
            env.max_drive_az = cal->clicks_per_sec_az() * RAD_PER_CLICK_AZ;
            env.max_drive_alt = cal->clicks_per_sec_alt() * RAD_PER_CLICK_ALT;
            az_count = alt_count = 0;
            logprintf(PSTR(">c %g %g"), env.max_drive_az, env.max_drive_alt);
            // we are (incidentally) at the home position
            reset_at_home();
        } else if (doing == Activity::DriveAzAltActivity) {
            telescope.step(millis());
            controller.stop();

            const char* p = Quaternion::coords_str(telescope.T(), Position::AzAlt);
            logprintf(PSTR("#arrived at %s"), p);
            logprintf(PSTR(">g A %g %g"),
                      M_PI - telescope.kalman_phi()->angle(),
                      telescope.kalman_theta()->angle());
        }

        curr = activity_queue.next();
        if (curr == NULL) {
            activity_poke_time = TIME_NEVER;
            endstop_timeout = TIME_NEVER;
        } else {
            activity_poke_time = curr->start(millis());
        }
        // logprintf(PSTR("#activity: more? %c, next@" MS_FMT),
        //           (curr == NULL ? 'n' : 'y'), activity_poke_time);

    } else if (activity_queue.current() != NULL && reftime >= activity_poke_time) {
        activity_poke_time = activity_queue.current()->poke(millis());

    } else if (reftime > endstop_timeout) {
        //logprintf(PSTR("#at endstop: (%d, %d)"), az_count, alt_count);
        logprintf(PSTR(">g E %d %d"), az_count, alt_count);
        // If we have in fact reached the endpoint, then there is
        // little point in poking the activity, to update its
        // position: since we've hit the endstop, we no longer have a
        // reliable notion of position (and we document this).
        // However we should update it nonetheless, (a) in case this
        // is a false detection, and (b) to be consistent with the
        // requirement that we call State::step whenever we change the speeds.
        telescope.step(reftime);
        controller.stop();
        endstop_timeout = TIME_NEVER; // don't come here again
        if (activity_queue.current() != NULL) {
            activity_queue.current()->at_endstop(reftime, az_count, alt_count);
        }
        if (heading_home_p) {
            reset_at_home();
        }

    } else if (Serial.available() > 0 && bs.read_items_read_char() > 0) {
        // do nothing: the relevant work was done in bs.read_items_read_char();
        // but deem this to be an active cycle

    } else if (reftime > status_time) {
        //// - `s`, flags, _various_...
        ////
        ////    Status: the status line is produced at a regular
        ////    interval, and is intended to contain information about
        ////    the current status of the telescope and its
        ////    movements.  This is distinct from the output of the
        ////    `S` command, which includes some of this information,
        ////    but is intended to be for overall status and
        ////    calibration information.
        ////
        ////    The output `flags` is a hex value indicating what
        ////    status fields are being output in the following columns.
        ////
        ////    The comma-separated columns are as documented for the
        ////    `s` command above, and they appear in the order listed there.
        ////
        logprintf(PSTR("s,%x%|"), status_fields);
        if (status_fields & STATUS_TIME) {
            logprintf(PSTR("," MS_FMT "%|"), reftime);
        }
        if (status_fields & STATUS_AZIMUTH_CLICKS) {
            // We have to negate this, in order to show the
            // click-azimuth increasing westwards (like real azimuth),
            // as documented
            logprintf(PSTR(",%d%|"), -az_count);
        }
        if (status_fields & STATUS_ALTITUDE_CLICKS) {
            logprintf(PSTR(",%d%|"), alt_count);
        }
        if (status_fields & STATUS_AZIMUTH_RAD) {
            Kalman* K = telescope.kalman_phi();
            logprintf(PSTR(",%g%|"), M_PI - K->angle());
        }
        if (status_fields & STATUS_ALTITUDE_RAD) {
            Kalman* K = telescope.kalman_theta();
            logprintf(PSTR(",%g%|"), K->angle());
        }
        if (status_fields & STATUS_AZIMUTH_ERROR) {
            Kalman* K = telescope.kalman_phi();
            logprintf(PSTR(",%g%|"), sqrt(K->angle_error()));
        }
        if (status_fields & STATUS_ALTITUDE_ERROR) {
            Kalman* K = telescope.kalman_theta();
            logprintf(PSTR(",%g%|"), sqrt(K->angle_error()));
        }
        if (status_fields & STATUS_COMMANDED_AZ_SPEED) {
            logprintf(PSTR(",%d%|"), controller.curr_az_speed_int());
        }
        if (status_fields & STATUS_COMMANDED_ALT_SPEED) {
            logprintf(PSTR(",%d%|"), controller.curr_alt_speed_int());
        }
        if (status_fields & STATUS_CURRENT_AZ) {
            // I don't currently know what units these measurements are in
            int pwm_az_current = analogRead(PWM_AZ_CURRENT);
            logprintf(PSTR(",%g%|"), pwm_az_current*1.0);
        }
        if (status_fields & STATUS_CURRENT_ALT) {
            // I don't currently know what units these measurements are in
            int pwm_alt_current = analogRead(PWM_ALT_CURRENT);
            logprintf(PSTR(",%g%|"), pwm_alt_current*1.0);
        }
        if (status_fields & STATUS_WORKLOAD_TOTAL_CYCLES) {
            logprintf(PSTR(",%d%|"), workload_cycles);
        }
        if (status_fields & STATUS_WORKLOAD_WORK_CYCLE) {
            logprintf(PSTR(",%g%|"), 1.0 - workload_idle_cycles*1.0/workload_cycles);
        }
        if (status_fields & STATUS_WORKLOAD_WORK_FRACTION) {
            logprintf(PSTR(",%g%|"), workload_working*1.0/(millis() - workload_start));
        }
        if (status_fields & STATUS_WORKLOAD_SUCCESSIVE_ACTIVE_CYCLES) {
            logprintf(PSTR(",%d%|"), workload_backup_cycles);
        }
        if (status_fields & STATUS_WORKLOAD_DISCARDED_OUTPUT) {
            logprintf(PSTR(",%d%|"), bs.count_discarded());
        }
        if (status_fields & STATUS_ACTIVITY_QUEUE_SIZE) {
            logprintf(PSTR(",%d%|"), Activity::n_activities());
        }
        if (status_fields & STATUS_CURRENT_ACTIVITY) {
            char st;
            if (activity_queue.current() == NULL) {
                st = 'n';
            } else if (activity_queue.current()->is_complete()) {
                st = 'c';
            } else {
                st = 'y';
            }
            logprintf(PSTR(",%c%|"), st);
        }
        logprintf(PSTR(""));

        workload_cycles = 0;
        workload_idle_cycles = 0;
        workload_backup_cycles = 0;
        workload_start = reftime;
        workload_working = 0;

        active_p = false;       // count this as a non-working cycle

        status_time += status_cadence;
    } else {
        active_p = false;
    }

    if (active_p) {
        workload_working += millis() - reftime;
        if (workload_active_last_time_p) {
            // active for two consecutive cycles
            workload_backup_cycles++;
        }
    } else {
        workload_idle_cycles++;
    }
    workload_active_last_time_p = active_p;
    workload_cycles++;
}

// The following is how reading from the serial should be done, as it's interrupt driven.
// For some reason which I can't pin down, however, this appears to
// fail in some circumstances, and the main loop appears to completely
// ignore serial input.  Wassup?!
// void serialEvent()
// {
//     bs.read_items_read_char();
// }
