#include <math.h>

#include "quaternion.h"
#include "exception.h"
#include "angles.h"

#if PROVARINO
// if this is non-zero, we output quaternions, in operator<< as
// angles; if zero, we output them as a list of components
// (this only matters in debugging/test code)
#define FORMAT_WITH_ANGLES 1
#endif

/**
 * The default constructor creates a K quaternion.
 * Creating a Unit quaternion might be more intuitive, but since this
 * isn't a vector, it has the potential to cause lots of trouble if
 * used accidentally.
 */
Quaternion::Quaternion()
{
    q[0] = q[1] = q[2] = 0;
    q[3] = 1;
}

/**
 * The constructor creates a quaternion from four coefficients of the
 * basis quaternions.  This is the only constructor that can create a
 * non-unit quaternion.
 */
Quaternion::Quaternion(const double q0, const double q1, const double q2, const double q3)
{
    q[0] = q0;
    q[1] = q1;
    q[2] = q2;
    q[3] = q3;
}

/**
 * Produce one of the basis quaternions.
 */
Quaternion::Quaternion(const basis_quaternion b)
{
    q[0] = q[1] = q[2] = q[3] = 0;

    switch (b) {
      case Unit:
        q[0] = 1;
        break;
      case I:
        q[1] = 1;
        break;
      case J:
        q[2] = 1;
        break;
      case K:
        q[3] = 1;
        break;
    }
}

/**
 * Create a quaternion vector which points in the direction of the given position.
 */
Quaternion::Quaternion(const Position& p)
{
    Position p_topo = p.to_system(Position::TopoPolar); // usually a no-op
    const double lon = p_topo.lon();
    const double colatitude = M_PI_2 - p.lat();
    const double sincolat = sin(colatitude);

    q[0] = 0;
    q[1] = sincolat * cos(lon);
    q[2] = sincolat * sin(lon);
    q[3] = cos(colatitude);
}

/**
 * Produce a new quaternion which is the negative of this one.
 */
Quaternion Quaternion::operator-() const
{
    return Quaternion(-q[0], -q[1], -q[2], -q[3]);
}

/**
 * Produces the sum of two quaternions.
 */
Quaternion Quaternion::operator+(const Quaternion& q2) const
{
    return Quaternion(q[0] + q2.q[0],
                      q[1] + q2.q[1],
                      q[2] + q2.q[2],
                      q[3] + q2.q[3]);
}

/**
 * Produces the difference of two quaternions.
 */
Quaternion Quaternion::operator-(const Quaternion& q2) const
{
    return Quaternion(q[0] - q2.q[0],
                      q[1] - q2.q[1],
                      q[2] - q2.q[2],
                      q[3] - q2.q[3]);
}

/**
 * Performs the quaternion product of two quaternions.
 */
Quaternion Quaternion::operator*(const Quaternion& qmul) const
{
    const double* p = qmul.q;

    // The following requires 16 multiplications and 12 additions
    return Quaternion(q[0]*p[0] - q[1]*p[1] - q[2]*p[2] - q[3]*p[3],
                      q[0]*p[1] + q[1]*p[0] + q[2]*p[3] - q[3]*p[2],
                      q[0]*p[2] - q[1]*p[3] + q[2]*p[0] + q[3]*p[1],
                      q[0]*p[3] + q[1]*p[2] - q[2]*p[1] + q[3]*p[0]);
}

/**
 * 'Rotate' the quaternion.  Given a quaternion, r, and taking the
 * current quaternion to be A, return `r*A*r.conj`.  This requires 23
 * multiplications and 24 additions, as opposed to 32 multiplications
 * and 24 additions when done the 'long' way.  This should also be
 * more accurate.
 *
 * Note that this is different from the {@link rotation} method,
 * which constructs a quaternion which generates a rotation.
 *
 * See _quaternion-pointing_ notes, at note `rotation`.
 *
 * @param rq the quaternion which generates the rotation, which is
 * assumed to be a unit quaternion
 */
Quaternion Quaternion::rotate(const Quaternion& rq) const
{
    const double* r = rq.q;

    // the following requires 23 multiplications and 24 additions
    double r00 = r[0]*r[0];
    double r01 = r[0]*r[1];
    double r02 = r[0]*r[2];
    double r03 = r[0]*r[3];
    double r11 = r[1]*r[1];
    double r12 = r[1]*r[2];
    double r13 = r[1]*r[3];
    double r22 = r[2]*r[2];
    double r23 = r[2]*r[3];
    double r33 = r[3]*r[3];

    return Quaternion(q[0] * (r00 + r11 + r22 + r33),
                      q[1] * (r00 + r11 - r22 - r33) + 2*( q[3]*(r02 + r13)
                                                          +q[2]*(r12 - r03)),
                      q[2] * (r00 - r11 + r22 - r33) + 2*( q[1]*(r03 + r12)
                                                          +q[3]*(r23 - r01)),
                      q[3] * (r00 - r11 - r22 + r33) + 2*( q[2]*(r01 + r23)
                                                          +q[1]*(r13 - r02)));
}

/**
 * Reflect the quaternion through the line generated by the argument.
 * This is equivalent to a rotation of pi radians around the argument,
 * which is assumed to be a unit vector.  The result of this method
 * should be equivalent to the application of Quaternion::rotate,
 * but should be faster because of the assumption that the argument is a vector.
 *
 * @param rq the quaternion which generates the rotation, which is
 * assumed to be a unit vector
 */
Quaternion Quaternion::reflect(const Quaternion& rq) const
{
    const double* r = rq.q;

    double r11 = r[1]*r[1];
    double r12 = r[1]*r[2];
    double r13 = r[1]*r[3];
    double r22 = r[2]*r[2];
    double r23 = r[2]*r[3];
    double r33 = r[3]*r[3];

    return Quaternion(q[0],
                      q[1] * (+r11 - r22 - r33) + 2*(q[3]*r13 + q[2]*r12),
                      q[2] * (-r11 + r22 - r33) + 2*(q[1]*r12 + q[3]*r23),
                      q[3] * (-r11 - r22 + r33) + 2*(q[2]*r23 + q[1]*r13));
}

/**
 * Return quaternion scaled by real factor.
 */
Quaternion Quaternion::scale(double factor) const
{
    return Quaternion(q[0]*factor,
                      q[1]*factor,
                      q[2]*factor,
                      q[3]*factor);
}

/**
 * Return the vector product of this quaternion with the argument.
 * We assume, but don't check, that the both the current and the
 * argument quaternion are in Vec(H).
 *
 * @param arg the quaternion to be crossed with this one (on the right)
 * @param norm if true, the result is normalised to unit length
 */
Quaternion Quaternion::cross(const Quaternion& arg, bool norm) const
{
    const double* p = arg.q;
    double a1 = q[2]*p[3] - q[3]*p[2];
    double a2 = q[3]*p[1] - q[1]*p[3];
    double a3 = q[1]*p[2] - q[2]*p[1];

    if (norm) {
        double len = sqrt(a1*a1 + a2*a2 + a3*a3);
        a1 /= len;
        a2 /= len;
        a3 /= len;
    }

    return Quaternion(0, a1, a2, a3);
}

/**
 * Return scalar product of two quaternions in Vec(H).
 * We assume, but don't check, that the both the current and the
 * argument quaternion are in Vec(H).
 */
double Quaternion::dot(const Quaternion& arg) const
{
    const double* p = arg.q;

    return q[1]*p[1] + q[2]*p[2] + q[3]*p[3];
}

/**
 * Returns a new quaternion which is the conjugate of this one
 */
Quaternion Quaternion::conj(void) const
{
    return Quaternion(q[0], -q[1], -q[2], -q[3]);
}

/**
 * The components of the quaternion, as an array of doubles, in radians
 */
const double* Quaternion::cpts(void) const
{
    return q;
}

/**
 * Test whether a quaternion is a unit quaternion.
 * That is, test whether it is (1, 0, 0, 0).
 * We sometimes use this quaternion, generated by {@link Quaternion(const basis_quaternion)},
 * as a sentinel value.
 */
bool Quaternion::is_unit_p(void)
{
    // what we actually test is whether component 0 is exactly 1.0
    return (q[0] == 1.0);
}

/**
 * Obtains the direction of a quaternion, by decomposing it into polar
 * angles.  The coordinates are returned as part of a {@link Position}
 * object, so the specific result coordinate system may be chosen
 * using the arguments.
 *
 * @param system the system in which the result angles are to be expressed
 * @param from_system the system in which the quaternion's rectangular
 * components should be taken to be expressed
 * @see Position::Position(double,double,Position::System,Position::System)
 */
Position Quaternion::coords(Position::System system, Position::System from_system) const
{
    // The quaternion is Q=L (i q_1 + j q_2 + k q_3), where L is most
    // typically 1, for a unit quaternion.  We can represent
    // this by polar angles (a, d), with co-latitude cd=pi/2 - d, so that
    //
    //   Q = L (i cos a.sin cd + j sin a.sin cd + k cos cd)
    //
    // Thus tan a = q_2/q_1.  If L=1, then cd = acos(q_3).
    //
    // If L != 1, however, then we can calculate
    //
    //   (q_1^2 + q_2^2) / q_3^2 = sin^2 cd / cos^2cd = tan^2 cd
    double longitude = atan2(q[2], q[1]);
    while (longitude < 0) {
        longitude += 2*M_PI;
    }

    // The following version only works if the quaternion is unit
    //double latitude = M_PI_2 - acos(q[3]);
    // The following version works in all cases.
    double sincd = sqrt(q[1]*q[1] + q[2]*q[2]);
    double latitude = M_PI_2 - atan2(sincd, q[3]);

    return Position(longitude, latitude, from_system, system);
}

/**
 * Returns a pointer to a static buffer containing a formatted
 * representation of the quaternion.
 *
 * @param q the quaternion to be formatted
 * @param system the system to be used for the formatted output
 * @param sexagesimal_p if true, output as sexagesimal, if false, in radians
 * @return a pointer to a common static buffer
 */
const char* Quaternion::coords_str(const Quaternion& q,
                                   Position::System system,
                                   bool sexagesimal_p)
{
#define BUFLEN 40
    static char buf[BUFLEN];
    Position pos = q.coords(system);

        char* p;
    if (sexagesimal_p) {
        p = Angles::r2dms_b(buf, BUFLEN, pos.lon());
    } else {
        p = buf;
        snprintf(p, BUFLEN, "%g", pos.lon());
    }
    while (*p != '\0') p++;

    if (system == Position::TopoPolar) {
        strcpy(p, "E,");
    } else if (system == Position::AzAlt) {
        strcpy(p, " az,");
    } else {
        strcpy(p, " lon,");
    }
    while (*p != '\0') p++;

    size_t nleft = BUFLEN-(p-buf);
    if (sexagesimal_p) {
        Angles::r2dms_b(p, nleft, pos.lat());
    } else {
        snprintf(p, nleft, "%g", pos.lat());
    }
    while (*p != '\0') p++;

    nleft = BUFLEN-(p-buf);
    if (system == Position::TopoPolar) {
        strncpy(p, "N", nleft);
    } else if (system == Position::AzAlt) {
        strncpy(p, " alt", nleft);
    } else {
        strncpy(p, " lat", nleft);
    }

    return buf;

#undef BUFLEN
}

#if PROVARINO
/**
 * Prints a quaternion as a formatted string to the output stream.
 */
std::ostream& operator<<(std::ostream& strm, const Quaternion& quat)
{
#if FORMAT_WITH_ANGLES
    char buf[50];

    using Angles::r2d;

    const double* q = quat.cpts();
    if (q[0] > 1e-7) {
        // this ought to be a rotation generator, cos(a/2)+u*sin(a/2)
        double hangle = acos(q[0]);
        double lon = atan2(q[2], q[1]);
        double lat = M_PI_2 - acos(q[3]/sin(hangle));
        snprintf(buf, 50, "[ax=%.3g%+.3g ang=%.3g]", r2d(lon), r2d(lat), r2d(hangle*2));
    } else {
        Position p = quat.coords();
        snprintf(buf, 50, "[%g%+g]", Angles::r2d(p.lon()), Angles::r2d(p.lat()));
    }
    return strm << buf;
#else
    const double* q = quat.cpts();
    return strm << '[' << q[0] << ',' << q[1] << ',' << q[2] << ',' << q[3] << ']';
#endif
}
#endif

/**
 * Creates a new vector quaternion which is pointing in the direction
 * of the given (east-)longitude and latitude.  Both of the angles are
 * given in 'polar' radians, thus east of south.
 */
Quaternion Quaternion::vector(const double longitude, double latitude)
{
    // if (latitude < -M_PI_2 || latitude > M_PI_2) {
    //     fatal(PSTR("Range exception in quaternion::vector: %s=%g"), "lat", latitude);
    // }
    Position p = Position(longitude, latitude, Position::TopoPolar);
    return Quaternion(p);
}

/**
 * Produce a quaternion which specifies a rotation about the current
 * quaternion, by the given angle.
 */
Quaternion Quaternion::rotation(const double angle) const
{
    const double mag = sqrt(q[1]*q[1] + q[2]*q[2] + q[3]*q[3]);
    const double sin_half_angle = sin(angle/2.0);

    return Quaternion(cos(angle/2.0),
                      q[1]/mag * sin_half_angle,
                      q[2]/mag * sin_half_angle,
                      q[3]/mag * sin_half_angle);
}

#if 0
/**
 * Obtain a unit quaternion which represents a rotation about an axis
 * with the given longitude and latitude, about a clockwise angle.
 */
Quaternion Quaternion::rotation(const double longitude,
                                const double latitude,
                                const double angle)
{
    if (latitude < -M_PI_2 || latitude > M_PI_2) {
        fatal(PSTR("Range exception in %s: %s=%g"), "quaternion::rotation", "lat", latitude);
    }
    const double colatitude = M_PI_2 - latitude;
    const double sin_half_angle = sin(angle/2.0);
    const double sincolat = sin(colatitude);

    return Quaternion(cos(angle/2.0),
                      sincolat * cos(longitude) * sin_half_angle,
                      sincolat * sin(longitude) * sin_half_angle,
                      cos(colatitude) * sin_half_angle);
}
#endif
