/* * @file quaternion.h */

#ifndef QUATERNION_H_LOADED
#define QUATERNION_H_LOADED 1

#if PROVARINO
#include <ostream>
#endif

#include "position.h"

/**
 * The Quaternion class encapsulates quaternions (quelle surprise!),
 * and provides a small set of operations for manipulating them.
 *
 * The directions in this module are described either in terms of the
 * basis quaternions 1, i, j, k, or in terms of longitude [0..2pi] and
 * latitude [-pi/2..+pi/2], where the direction i is (long,lat) = (0,0),
 * j is (pi/2, 0) and k is (0, pi/2).  The longitude coordinate
 * therefore increases from i to j (ie, moving eastwards).
 *
 * Although it doesn't much matter to this class, when viewed as a
 * topocentric coordinate system, i should be taken as pointing
 * south, and j east.
 *
 * The arithmetic operations in this class return new constant
 * objects, rather than mutating existing ones.
 */
class Quaternion {
 private:
    // I'd like to make this const, so that there's no
    // possibility of changing the quaternion once it's created,
    // but I then need to assign the values inside the constructor,
    // and that's not possible in C++ (ie, C++ isn't like Java, where
    // the const/final status kicks in at the end of the constructor).
    //
    // It's possible to have quite a long argument with the compiler
    // about this, and get drawn into a maze of twisty passages, all
    // alike, which have scrawled on the walls remarks about
    // non-default copy assignment operators, and copy-and-swap
    // constructors, but it doesn't really seem worth the effort right now.
    double q[4];

 public:
    Quaternion();               // default constructor
    Quaternion(const double q0, const double q1, const double q2, const double q3);
    /**
     * The set of basis quaternions
     * @see Quaternion(const basis_quaternion)
     */
    enum basis_quaternion { Unit, ///< The unit quaternion (sometimes used as a sentinel value; see is_unit_p())
                            I,    ///< the basis quaternion in the i direction
                            J,    ///< the basis quaternion in the j direction
                            K     ///< the basis quaternion in the k direction
    };
    Quaternion(const basis_quaternion);
    Quaternion(const Position&);

    Quaternion operator-() const;                  // unary minus
    Quaternion operator+(const Quaternion&) const; // addition
    Quaternion operator-(const Quaternion&) const; // subtraction
    Quaternion operator*(const Quaternion&) const; // multiplication
    Quaternion scale(double factor) const;         // scale by real factor
    Quaternion rotate(const Quaternion&) const;    // A -> r*A*r.conj
    Quaternion reflect(const Quaternion&) const;   // A -> r*A*r.conj, but argument assumed to be vector

    Quaternion conj(void) const;
    const double* cpts(void) const;
    Position coords(Position::System system=Position::TopoPolar,
                    Position::System from_system=Position::TopoPolar) const;

    Quaternion cross(const Quaternion&, bool norm=false) const; // vector product of quaternions in Vec(H)
    double dot(const Quaternion&) const;       // scalar product of quaternions in Vec(H)

    Quaternion rotation(const double angle) const;

    bool is_unit_p(void);

    static Quaternion vector(const double longitude, double latitude);
    // static Quaternion rotation(const double longitude,
    //                            const double colatitude,
    //                            const double angle);
    static const char* coords_str(const Quaternion&,
                                  Position::System system=Position::TopoPolar,
                                  bool sexagesimal_p=true);
};

#if PROVARINO
std::ostream& operator<<(std::ostream &strm, const Quaternion& a);
#endif

#endif /* QUATERNION_H_LOADED */
