#ifndef ROTATION_H_LOADED
#define ROTATION_H_LOADED 1

// An object to determine some properties of a planned rotation, such
// as whether and when it will intersect with a plane

#include "quaternion.h"

/**
 * Encapsulates a rotation of a point about an axis.
 * It mostly exists to support the two `intersection` methods.
 */
class Rotation {
private:
    Quaternion A;  ///< the point which is rotated
    Quaternion u;  ///< the rotation axis
    mutable Quaternion Au;      ///< A reflected through the line u
    mutable Quaternion uxA;     ///< the vector (u x A)

public:
    /** Return the start point of the rotation */
    const Quaternion& start(void) const { return A; };
    /** Return the axis of rotation */
    const Quaternion& axis(void) const { return u; };

    /**
     * Create a rotation.
     *
     * @param start the point which is rotated
     * @param axis the axis about which the point is rotated
     */
    Rotation(const Quaternion start, const Quaternion axis);

    /**
     * Determine whether a point will be rotated across a meridian.
     * The point `start` is rotated around an axis `axis`.  At what point does
     * it cross a plane which is perpendicular to `normal`?  We are interested in
     * the meridian which is the intersection of this plane and the
     * half-great-circle obtained by rotating `pole` by pi clockwise
     * about `normal`.
     *
     * Returns tan(theta/2), where theta is the angle by which `start` must
     * be rotated to intersect the meridian.  Negative values should
     * be interpreted as theta/2 being in the second quadrant rather
     * than the fourth (that is, theta is in the third or fourth quadrants).
     *
     * See _quaternion-pointing_ notes, section `meridians`.
     *
     * @return tan(theta/2), as noted above; or return 0.0 if the
     * rotation would not intersect the meridian at all
     */
    double intersection(const Quaternion& normal, const Quaternion& pole) const;

    /**
     * Determine whether a point will be rotated across a plane
     * bounded by a small circle.
     * The point `start` is rotated around an axis `axis`.  At what point does
     * it cross a plane which is perpendicular to `normal` and which cuts
     * the vector `normal` at a distance `h` from the origin?
     *
     * Returns tan(theta/2), where theta is the angle by which A must
     * be rotated to intersect the plane.  Negative values should
     * be interpreted as theta/2 being in the second quadrant rather
     * than the fourth (that is, theta is in the third or fourth
     * quadrants).  We return the value of tan(theta/2) corresponding
     * to the smaller value of theta (that is, the first intersection
     * of the rotation with the plane).
     *
     * See _quaternion-pointing_ notes, section `meridians`.
     *
     * @return tan(theta/2), as noted above; or return 0.0 if the
     * rotation would not intersect the meridian at all
     */
    double intersection(const Quaternion& normal, const float h) const;
};


#endif /* ROTATION_H_LOADED */
