#include "state.h"
#include "angles.h"
#include "math.h"
#include "exception.h"
#include "rotation.h"
#include "buffered-serial.h"
#include "qp.h"
#include "config-qp.h"

#if PROVARINO
#include <iostream>

using std::endl;
#define COUT std::cout << __FILE__ << ':' << __LINE__ << ": "
// DEBUG=1 is only expected to work for this case
#ifndef DEBUG
#define DEBUG 0
#endif
#define logprintf BufferedSerial::testenv_logprintf

#else
// non-provarino case
extern BufferedSerial bs;
#define logprintf bs.logprintf
#define DEBUG 0

#endif

Schedule::Schedule()
    : phidot_(0), thetadot_(0)
{
    // empty
}

/** Construct an instance of this schedule */
SlewSchedule::SlewSchedule(const Position& target)
    : full_speed_time_(TIME_NEVER),
      route_planned_p_(false)
{
    Position target_polar = target.to_system(Position::TopoPolar);
    target_phi_ = target_polar.lon();
    target_theta_ = target_polar.lat();

    schedule_ok_p_ = true;

    TelescopeController* controller = get_controller();

    if (target_theta_ > CONFIG_MAX_ALTITUDE) {
        schedule_ok_p_ = false;

    } else if (target_theta_ < controller->lowest_altitude()) {
        schedule_ok_p_ = false;

    } else {
        // Calculate whether this target is within the 'empty sector'
        // where the telescope can't go
        // (ie, the sector between the eastmost and westmost azimuth stops).
        // Yes, there may well be a more elegant way of determining this,
        // but this is confusing enough, without being clever.

        // rot_lon and rot_es are the target and eastmost stop, in a frame
        // rotated so that the westmost stop has longitude pi
        float rot_lon = target_phi_ + M_PI - controller->westmost_phi();
        float rot_es = controller->eastmost_phi() + M_PI - controller->westmost_phi();
        // logprintf(PSTR("# target: %g + pi - %g = %g"),
        //           target_phi_, controller->westmost_phi(), rot_lon);
        // logprintf(PSTR("# east  : %g + pi - %g = %g"),
        //           controller->eastmost_phi(), controller->westmost_phi(),
        //           rot_es);
        // make sure both are in [-pi,pi]
        if (rot_lon < -M_PI) rot_lon += 2*M_PI;
        if (rot_lon > M_PI) rot_lon -= 2*M_PI;
        if (rot_es < -M_PI) rot_es += 2*M_PI;
        if (rot_es > M_PI) rot_es -= 2*M_PI;
        if (rot_lon > rot_es) {
            // we'd be aiming for the empty sector -- no good
            schedule_ok_p_ = false;
        }
    }

    logprintf(PSTR("#Schedule: target=%s or (%g, %g)topo"),
              Quaternion::coords_str(target),
              target_phi_, target_theta_);
}

double SlewSchedule::calculate_angdot(double remaining_angle, ms_t current_time)
{
    // FIXME: this is a VERY crude stopping criterion
    double arrived_angle = 0.004; // 0.004 rad ~= 0.25deg

    // We want to slew at maximum speed, therefore set the
    // notional angdot (in fractional speed) to be greater than
    // one.
    //
    // angdot is absolute speed, rad/s
    double angdot = copysign(CONFIG_MAXIMUM_DRIVE_VELOCITY, remaining_angle);

    //angdot = CONFIG_MAXIMUM_DRIVE_VELOCITY;

    // We want to accelerate the telescope gracefully, and bring
    // it to a graceful stop.
    double slowdown_range = 0.1; // 0.1 radians ~= 5deg

    double mod_ang = fabs(remaining_angle);

    if (mod_ang < arrived_angle) {
        angdot = 0;

    } else if (mod_ang < slowdown_range) {
        // Graceful stop:
        //
        // We're now close (~5 degrees) to the slew target.
        // Avoid making this ratio zero, since
        // (a) the quantisation of the drive speeds (to fractions
        // n/255) could mean that we round to zero, and (b) the
        // PWM output is potentially slightly non-linear at low
        // values.
        double ratio = fmax(pow(mod_ang/slowdown_range, 2.0), CONFIG_MIN_SPEED);
        //double ratio = fmax(mod_ang/slowdown_range, 0.1);
        //logprintf(PSTR("#slowing: %g * %g"), ratio, angdot);
        angdot *= ratio;

    } else if (current_time < full_speed_time_) {
        // Graceful start:
        //
        // We're less than one accel_time from starting the movement
        ms_t accel_time = 5000; // ms
        if (full_speed_time_ == TIME_NEVER) {
            // first time: set the target time to be 5s into the future
            full_speed_time_ = current_time + accel_time;
            if (full_speed_time_ == TIME_NEVER) {
                // freaky coincidence; FIXME: what about wrapping round?
                full_speed_time_--;
            }
        }
        // avoid making this ratio zero, as above
        double ratio = fmax(pow(1.0 - ((full_speed_time_ - current_time)*1.0)/accel_time, 2.0),
                            0.1);
        //logprintf(PSTR("#accelerating: %g * %g"), ratio, angdot);

        angdot *= ratio;
    }

    // logprintf(PSTR("#calc: %g, %g%s"),
    //           remaining_angle,
    //           angdot,
    //           (mod_ang < slowdown_range ? "  slow" : ""));
    // COUT << "calculate: ang=" << mod_ang << '(' << arrived_angle
    //      << "), " << (mod_ang < slowdown_range ? "slow, " : "")
    //      << "angdot=" << angdot << endl;

    return angdot;
}

/**
 * Take a step on the slewing path.
 * Phi is the longitudinal coordinate, theta is the altitude (so
 * increasing towards the zenith).  Note that, like the outputs from
 * the Kalman filters, these arguments should _not_ be normalised into [-pi,pi].
 */
ms_t SlewSchedule::step(double current_phi, double current_theta, ms_t current_time)
{
    if (! route_planned_p_) {
        logprintf(PSTR("Sch: target is (%g, %g)%|"), target_phi_, target_theta_);

        TelescopeController* controller = get_controller();
        // We have to plan a route which avoids the empty sector.
        // Let c be the current phi, and t be the target
        // phi (this first time, these will both be in [-pi,pi]),
        // and w be the longitude of the westmost stop.
        // Define c' = c + (pi - w) and t' = t + (pi - w).
        // Then the route is straightforward in these coordinates,
        // since the empty sector now goes between longitude +pi and
        // some longitude a little less than that (to the eastmost azimuth
        // stop): we simply have to increase the current phi
        // by (t'-c') (which may of course be negative).  Thus reset the
        // target_phi_ to be current_phi + (t'-c').
        float rot = M_PI - controller->westmost_phi(); // rot in [0,2pi]
        float cp = current_phi + rot;
        if (cp > M_PI) cp -= 2*M_PI;
        float tp = target_phi_ + rot;
        if (tp > M_PI) tp -= 2*M_PI;

        target_phi_ = current_phi + (tp - cp);

        // logprintf(PSTR(" cp=%g  tp=%g  newlon=%g"), cp, tp, target_phi_);

        route_planned_p_ = true;
    }

    double phi_to_go = target_phi_ - current_phi;
    double theta_to_go = target_theta_ - current_theta;

    phidot_ = calculate_angdot(phi_to_go, current_time);
    thetadot_ = calculate_angdot(theta_to_go, current_time);

    // logprintf(PSTR("# c=(%g,%g)  d=(%g,%g)  v=(%g,%g)"),
    //           current_phi, current_theta,
    //           phi_to_go, theta_to_go,
    //           phidot_, thetadot_);

    // Suggest returning in the time, in ms, it takes to rotate 10 mrad
    // (a rotation in 24hrs is 0.000072 rad/s, or a little under 5mrad/min)
    double angdot;
    double angle;
    double phidotabs = fabs(phidot_);
    double thetadotabs = fabs(thetadot_);
    if (phidotabs > thetadotabs) {
        angdot = phidotabs;
        angle = fabs(phi_to_go);
    } else {
        angdot = thetadotabs;
        angle = fabs(theta_to_go);
    }

    ms_t returned_time;

    if (angdot == 0) {
        returned_time = 0;
    } else {
        returned_time = (ms_t)round(fmin(10.0/angdot, // time/ms to rotate 10mrad
                                         (angle/2)/(angdot/1000))); // time to rotate half of remainder
    }

#if DEBUG
    COUT << "SlewSchedule::step(" << current_time
         << "): phi=" << current_phi
         << " + " << phidot_
         << "/s, theta=" << current_theta
         << " + " << thetadot_
         << "/s, angle="
         << angle
         << endl;
#endif

    return returned_time;
}

/**
 * Initialise a schedule which tracks a point in a rotation about a
 * given axis.  The result is a rotation expressed in horizontal coordinates.
 * If, however, the `observatory` position is given as the null
 * position (see Position::get_null()), then we take the axis to be in horizontal coordinates.
 *
 * @param axis the axis to be tracked about, in equatorial coordinates
 * @param hdot the angular speed to track at, in radians/sec clockwise
 *   about this axis
 * @param observatory the location of the observatory
 */
TrackingSchedule::TrackingSchedule(const Quaternion& axis,
                                   double hdot,
                                   Position observatory)
    : initial_step_(TIME_NEVER),
      maximum_angle_(0.0),
      hdot_(hdot/1000)
{
    // initial_ is not meaningfully initialised, but
    // initial_step_==TIME_NEVER is used as a sentinel value

    if (observatory.is_null_p()) {
        axis_ = axis;
    } else {
        double colatitude = (M_PI_2 - observatory.lat());
        Quaternion toobservatory = Quaternion(Quaternion::J).rotation(-colatitude);
        axis_ = axis.rotate(toobservatory);
    }

#if DEBUG
    COUT << "TrackingSchedule: ax=" << axis_ << ", speed=" << hdot_*1000 << "rad/s" << endl;
#endif
}

ms_t TrackingSchedule::step(double current_phi, double current_theta, ms_t t)
{
    // Suggest returning in the time, in ms, that it takes to rotate about 1.5 clicks
    // (the clicks are 8.7mrad apart),
    // but no less than 1s (because we don't want a too-fast flurry of corrections)
    ms_t step_time = round(fmax(1.0 * RAD_PER_CLICK_AZ / hdot_, 1000.0));

    Quaternion actual = Quaternion::vector(current_phi, current_theta);

    if (initial_step_ == TIME_NEVER) {
        // this is the first step
        initial_ = actual;
        initial_step_ = t;

    } else {
        // where ought we to be?
        float ought_angle = hdot_ * (t - initial_step_); // rotation from initial_

        if (maximum_angle_ != 0.0 && fabs(ought_angle) > maximum_angle_) {
            // there is a maximum rotation angle,
            // and we've reached it,
            // so stop
#if DEBUG
            COUT << "TrackingSchedlule: reached maximum angle " << maximum_angle_ << endl;
#endif
            step_time = 0.0;

        } else {
            // For the calculations below, see quaternion-pointing.pdf
            // notes, Sects. 3.2 and 3.3.2.
            //
            // 'ought' is where we ought to be at this point (T in the
            // notes); we are actually at 'actual', and want to
            // correct this hopefully small error at the same time.
            //
            // 'axis' is the rotation axis (u in the notes).
            Quaternion ought = initial_.rotate(axis_.rotation(ought_angle));

            // We now want to combine the rotation about axis_ at speed
            // hdot with a 'correction' rotation which brings 'actual' to
            // 'ought'.
            //
            // To apply this correction, we need only subtract this
            // correction from 'axis' below (I think -- I haven't fully
            // convinced myself of this).
            Quaternion axo = actual.cross(ought);
            double r0 = sqrt((1 + actual.dot(ought))/2);
            double rho = 1/(2.0 * r0);
            Quaternion correction = axo.scale(rho) + Quaternion(Quaternion::Unit).scale(r0);

            // Tuh is $(u \cross T) \dot h$ (in eq (16a))
            Quaternion Tuh = (axis_ - correction).scale(1000*hdot_).cross(ought);

            Quaternion QK = Quaternion(Quaternion::K);
            Quaternion kxT = QK.cross(ought);
            double modkxT = kxT.dot(kxT); // |k x T|^2

            phidot_ = Tuh.dot(kxT)/modkxT;
            double kdotT = QK.dot(ought);
            // the 'extra' minus sign (compared with the notes) here
            // means that thetadot_ is positive for rotations towards the zenith
            thetadot_ = - Tuh.dot(QK) * sqrt(modkxT) / (kdotT*kdotT - 1);

#if PROVARINO
#if DEBUG
            COUT << "TrackingSchedule: t=" << t
                 << ", actual=" << actual
                 << ", ought=" << ought
                 << ", max=" << maximum_angle_
                 << " at (" << phidot_ << ',' << thetadot_ << ")rad/s"
                 << endl;
#endif
#else
            logprintf(PSTR("#TS: speed (%g,%g)rad/s (vs %g), step=%dms"),
                      phidot_, thetadot_,
                      hdot_*1000.0, // hdot_ is rad/ms
                      step_time);
#endif
        }
    }

    return step_time;
}

/**
 * Returns the rotation speed of the schedule, in rad/s
 */
double TrackingSchedule::speed(void)
{
    return hdot_*1000.0;
}

/**
 * Establish the bounds of the permitted movement of the telescope.
 */
float TrackingSchedule::limit_tracking(Quaternion& actual, const TelescopeController* tc)
{
    Rotation r = Rotation(actual, (hdot_ > 0 ? axis_ : -axis_));
    // Construct a vector pointing normal to the plane which lies
    // vertically through the eastmost and westmost azimuth lines.
    // In both cases, the boundary of the sector is formed by the
    // half-meridian drawn out by rotation the north pole
    // clockwise about this vector.
    //
    // Hmm: can't help feeling that this is really Doing It The Hard Way.
    Quaternion eastmost_endstop = Quaternion(Position(tc->eastmost_phi(),
                                                      0,
                                                      Position::TopoPolar));
    Quaternion westmost_endstop = Quaternion(Position(tc->westmost_phi(),
                                                      0,
                                                      Position::TopoPolar));
    // Calculate the height of the greatest depression of the
    // telescope, above the plane through the origin, as a
    // proportion of the radius.
    float height_of_plane = sin(tc->lowest_altitude());

    Quaternion pole = Quaternion(Quaternion::K);
    float tan1 = r.intersection(eastmost_endstop, pole);
    float tan2 = r.intersection(westmost_endstop, pole);
    float tan3 = r.intersection(pole, height_of_plane);

    // tan<N> is 0.0 if the rotation doesn't intersect the specified plane.
    // Each of the tan<N> is in [0,2pi)
    float max_tan = tan1;
    if (tan2 != 0.0 && (max_tan == 0.0 || tan2 < max_tan)) max_tan = tan2;
    if (tan3 != 0.0 && (max_tan == 0.0 || tan3 < max_tan)) max_tan = tan3;

    // max_tan is tan(maximum_angle_ / 2), with negative values of max_tan
    // corresponding to max_ang in 3rd or 4th quadrants
    maximum_angle_ = 2*atan(max_tan);                 // in [-pi,pi)
    if (maximum_angle_ < 0) maximum_angle_ += 2*M_PI; // in [0,2pi)
    maximum_angle_ -= tc->endstop_clearance();        // clear of the endstop
    logprintf(PSTR("#TS: maximum angle=%grad"), maximum_angle_);

#if DEBUG
    COUT << "TrackingSchedule: limits " << eastmost_endstop
         << " / " << westmost_endstop
         << " --> " << tan1 << ',' << tan2 << ',' << tan3
         << " => max_tan " << max_tan
         << " => maximum angle " << maximum_angle_ << endl;
#endif

    return maximum_angle_;
}

/**
 * Initialise the state.
 *
 * @param schedule the schedule which this state is to follow
 * @param current_time the current `millis()` time
 * @param observatory a `System::Geog` location for the observatory,
 *   or NULL if the position is to be taken to be in AzAlt coordinates
 */
State::State(Schedule* schedule,
             const ms_t current_time,
             const Position* observatory)
    : schedule(schedule)
{
    // if (initial_max_speed.system() != Position::Velocity) {
    //     fatal("Bad call to State -- speeds should be of type Velocity, not %d",
    //           initial_max_speed.system());
    // }

    if (observatory == NULL) {
        to_observatory = Quaternion(Quaternion::Unit);
    } else {
        double colatitude = M_PI_2 - observatory->lat();
        to_observatory = Quaternion(Quaternion::J).rotation(-colatitude);
    }

    // Quaternion T_actual;

    // if (observatory == NULL) {
    //     T_actual = initial;
    // } else {
    //     double colatitude = M_PI_2 - observatory->lat();
    //     Quaternion to_observatory = Quaternion(Quaternion::J).rotation(-colatitude);
    //     T_actual = initial.rotate(to_observatory);
    // }

    // Position initial_pos = T_actual.coords();

    // phi = Kalman(initial_pos.lon(),
    //              initial_max_speed.lon(),
    //              CONFIG_DRIVE_SLOP,
    //              current_time);
    // //phi.verbose('A');
    // theta = Kalman(initial_pos.lat(),
    //                initial_max_speed.lat(),
    //                CONFIG_DRIVE_SLOP,
    //                current_time);
    // //theta.verbose('E');
}

/**
 * No-arg constructor.  This isn't able to do anything useful, and is
 * only for the case where we have no option but to initialise a
 * static State object without complete information.
 */
State::State()
    : schedule(NULL)
{
    // empty
}

/**
 * Sets the schedule which the state is to follow.
 * The state retains a pointer to this schedule, but does not take a copy of it.
 */
void State::set_schedule(Schedule* sch)
{
    schedule = sch;
}
/**
 * Return the current pointing of the state, as a vector in AzAlt coordinates.
 */
// Quaternion State::T(void) const
// {
//     return Quaternion::vector(phi.angle(), theta.angle());
// }

/**
 * Return the current position of the state, as a {@link Position} object.
 */
// Position State::position(Position::System target_system)
// {
//     return Position(phi.angle(), theta.angle(),
//                     Position::TopoPolar,
//                     target_system);
// }

#if PROVARINO
/**
 * Print out the state position as AzAlt.
 * This is for debugging or logging purposes.
 */
std::ostream& operator<<(std::ostream& strm, const State& st)
{
    Quaternion T = st.T();
    // p.lon() is azimuth (north=0, east=90), p.lat() is altitude

    return strm << "[State: " << T << "]";
}
std::ostream& operator<<(std::ostream& strm, const Schedule& s)
{
    return s.tostream(strm);
}
/**
 * Print out the schedule plan.
 * This is for debugging or logging purposes.
 */
std::ostream& SlewSchedule::tostream(std::ostream& strm) const
{
    return strm << "[SlewSchedule: to target " << target_phi_ << ',' << target_theta_ << "]";
}
std::ostream& TrackingSchedule::tostream(std::ostream& strm) const
{
    return strm << "[TrackingSchedule: about " << axis_ << " at " << hdot_*1000 << " rad/s]";
}
#endif

/**
 * Report an azimuthal click.
 * The caller is responsible for keeping track of the number of clicks received,
 * and thus being able to calculate the actual position of the telescope at this point.
 * @param t the time the click was registered
 * @param current_speed the current commanded speed of the telescope, as a fraction of full speed
 * @param angle the angular position of the telescope at the point this click was received
 */
void State::az_click(ms_t t, double current_speed, double angle)
{
    phi.click(t, current_speed, angle);
}

/**
 * Report a click in altitude.
 * The caller is responsible for keeping track of the number of clicks received,
 * and thus being able to calculate the actual position of the telescope at this point.
 * @param t the time the click was registered
 * @param current_speed the current commanded speed of the telescope, as a fraction of full speed
 * @param angle the angular position of the telescope at the point this click was received
 */
void State::alt_click(ms_t t, double current_speed, double angle)
{
    theta.click(t, current_speed, angle);
}

/**
 * Retrospectively update the state by incrementally rotating the pointing.
 * This updates the `phidot` and `thetadot` values.  This steps the
 * state by the rotation corresponding to the controller's actual phi and
 * theta speeds, which are presumed to have been applied for the given
 * number of ms.
 *
 * Because this updates the state's notion of its position, this
 * method should be called before any change in commanded speed. XXX NO
 *
 * @param actual_reftime_ms the current reference time, in ms.
 *
 * @return (new) step time in millisecond, or zero to indicate that
 *   we've arrived at a destination, or should otherwise stop
 */
ms_t State::step(ms_t actual_reftime_ms)
{
    TelescopeController* controller = get_controller();

    // We update the Kalman filters with _fractional_ speeds -- it's
    // the filters' responsibility to work out what the maximum speed
    // actually is
    // phi.update(actual_reftime_ms, controller->curr_az_speed());
    // theta.update(actual_reftime_ms, controller->curr_alt_speed());

    ms_t rval = 0;
    if (schedule != NULL) {
        // The filters return angles in radians, and they are deliberately
        // NOT normalised to [0,2pi]
        rval = schedule->step(phi.angle(), theta.angle(), actual_reftime_ms);
#if DEBUG
        COUT << "State::step: step at " << actual_reftime_ms << ", stepsize=" << rval
             << " to (" << phi.angle() << ',' << theta.angle() << ") -> "
             << T()
             << endl;
#endif

        if (rval == 0) {
            schedule = NULL;
        }
    }

    return rval;
}

/**
 * Return the suggested rate of change in longitude, in units of fractional speed.
 * This is a suggestion to the controller, made by this state's internal
 * schedule, of what commanded (fractional) speed would keep the
 * telescope working on that schedule.
 * Longitude increases east of south.
 */
double State::suggested_phidot_frac(void)
{
    Kalman* k = kalman_phi();
    double v = (schedule == NULL ? 0.0 : schedule->phidot());

    if (v > 0)
        v /= k->fullspeed_pos();
    else
        v /= k->fullspeed_neg();

    return v;
}

/**
 * Return the suggested rate of change in altitude, in units of fractional speed.
 * Positive moves towards the zenith.
 * @see suggested_phidot_frac
 */
double State::suggested_thetadot_frac(void)
{
    Kalman* k = kalman_theta();
    double v = (schedule == NULL ? 0.0 : schedule->thetadot());

    if (v > 0)
        v /= k->fullspeed_pos();
    else
        v /= k->fullspeed_neg();

    return v;
}
