#ifndef STATE_H_LOADED
#define STATE_H_LOADED 1

#include "quaternion.h"
#include "kalman.h"
#include "qp.h"
#include "types.h"

// Angular speeds in this the Schedule subclasses are in rad/s; speeds
// in the State class are either in rad/s or in 'fractional' speeds,
// which are speeds expressed as a fraction of the drive's maximum
// speed.  This maximum speed is discovered (starting with an initial
// estimate) by the Kalman filters within the State object.

class State;

/**
 * The Schedule class represents a planned rotation of the telescope.
 * This can be either a slew from one position directly to another
 * (which will typically happen by a great-circle route) or it will
 * generate the rotations to track an object across the celestial
 * sphere (typically along a small circle).
 */
class Schedule {
 protected:
    /** Suggested rate of change of phi (azimuthal coordinate, rad/s) */
    double phidot_;
    /** Suggested rate of change of theta (altitude coordinate, rad/s).
        Thus this is positive for rotations towards from the zenith.  */
    double thetadot_;

    Schedule();

 public:
    /**
     * Update the schedule with the current position of the system.
     * Step the schedule to the given time
     * @param current_phi the current pointing of the system (azimuth, rad)
     * @param current_theta the current pointing of the system (altitude, rad)
     * @param t the current time
     * @return a new suggested step time, or zero to indicate that
     *   we've arrived at a destination, or should otherwise stop
     */
    virtual ms_t step(double current_phi, double current_theta, ms_t t) = 0;

    /**
     * Obtain the suggested azimuthal angular speed.
     * The speed is positive for clockwise rotations about the zenith.
     * @return angular speed; units of rad/s.
     */
    double phidot(void) const { return phidot_; }

    /**
     * Obtain the suggested altitudinal angular speed.
     * The speed is positive for rotations towards the zenith.
     * @return angular speed; units of rad/s.
     */
    double thetadot(void) const { return thetadot_; }

#if PROVARINO
    virtual std::ostream& tostream(std::ostream& strm) const = 0;
#endif
};

/**
 * Provides the schedule for a slew to a particular point on the celestial sphere.
 * This is done by calculating a target phi and theta, and moving
 * directly to those angles by independent motions in the two axes.
 * This is in contrast to the initial implementation, which planned a
 * great-circle route and followed that.
 */
class SlewSchedule : public Schedule
{
private:
    float target_phi_;          // the longitude we're aiming for
                                // (+ve eastwards, rad, may be outside of [0,2pi])
    float target_theta_;        // altitude (rad, less than pi/2)
    ms_t full_speed_time_;      // time we aim to be at full speed
    bool route_planned_p_;      // have we worked out the route?
    bool schedule_ok_p_;        // is this a feasible schedule?

    double calculate_angdot(double remaining_angle, ms_t current_time);

 public:
    SlewSchedule(const Position& target);
    ms_t step(double current_phi, double current_theta, ms_t t);
    /** Is this a feasible schedule?  Return false if it would end up
        in the empty sector. */
    bool schedule_ok() const { return schedule_ok_p_; };

#if PROVARINO
    std::ostream& tostream(std::ostream& strm) const;
#endif
};

/**
 * Provides the schedule for tracking a point across the celestial
 * sphere, typically at a speed of one sidereal day per solar day.
 */
class TrackingSchedule : public Schedule {
 private:
    // initial_ and initial_step_ are 'almost' const, but they're not
    // initialised until the first call to step(), so can't be
    // labelled const
    Quaternion initial_;       // where the rotation is from
    ms_t initial_step_;        // the time of the first step
    double maximum_angle_;     // the maximum angle we may move before encountering a stop

    Quaternion axis_;          // the axis being rotated about
    double hdot_;              // the intended rotation speed in rad/ms

 public:
    TrackingSchedule(const Quaternion& axis,
                     double hdot,
                     Position observatory);

    ms_t step(double current_phi, double current_theta, ms_t t);
    double speed(void);

    float limit_tracking(Quaternion& actual, const TelescopeController*);

#if PROVARINO
    std::ostream& tostream(std::ostream& strm) const;
#endif
};

/**
 * The State class holds the state of the system while it is slewing
 * or tracking.
 */
class State {
private:
    /**
     * Because Schedule is polymorphic, we can't take a copy of it,
     * but have to point to an external instance.
     */
    Schedule* schedule;
    // schedule is not const, because we will occasionally
    // (in set_schedule()) update a current state with a new schedule/goal

    Quaternion to_observatory;

public:
    State(// const Quaternion& initial,
          // const Position& initial_max_speed,
          Schedule* schedule,
          const ms_t current_time,
          const Position* obs=NULL);

    State();

    ms_t step(ms_t t);
    void alt_click(ms_t t, double current_speed, double angle);
    void az_click(ms_t t, double current_speed, double angle);
    double suggested_phidot_frac(void);
    double suggested_thetadot_frac(void);

    void set_schedule(Schedule*);

    /**
     * The Kalman filter which manages the longitudinal coordinate.
     * This is the angle east of south, in radians.
     * It will generally be more convenient to obtain the telescope's
     * current position using the {@link position} method.
     */
    //Kalman* kalman_phi() { return &phi; }
    /**
     * The Kalman filter which manages the elevation coordinate.
     * This is the angle north of the horizontal, in radians.
     */
    //Kalman* kalman_theta() { return &theta; }

    // Quaternion T(void) const;
    // Position position(Position::System);
};

#if PROVARINO
std::ostream& operator<<(std::ostream& strm, const State& s);
std::ostream& operator<<(std::ostream& strm, const Schedule& s);
#endif
#endif /* STATE_H_LOADED */
